# PACKAGE_VERSION from setup.py generated by MakePy.ps1
PACKAGE_VERSION = '0.1.51'


# TODO Move get_package_version() to python-sdk (generated by MakePy.ps1)
# TODO Shall we make it method in all of our classes?
def get_package_version() -> str:
    return PACKAGE_VERSION
