# Generated by generate_table_columns.py from MySQL system tables. unlike table_definition which generated from our applicative table in the database (should be generated by Sql2Code in the future

table_columns = {
    "innodb_table_stats": (
        "clustered_index_size",
        "database_name",
        "last_update",
        "n_rows",
        "sum_of_other_index_sizes",
        "table_name"
    ),
    "innodb_index_stats": (
        "database_name",
        "index_name",
        "last_update",
        "sample_size",
        "stat_description",
        "stat_name",
        "stat_value",
        "table_name"
    ),
    "general_log": (
        "argument",
        "command_type",
        "event_time",
        "server_id",
        "thread_id",
        "user_host"
    ),
    "slow_log": (
        "db",
        "insert_id",
        "last_insert_id",
        "lock_time",
        "query_time",
        "rows_examined",
        "rows_sent",
        "server_id",
        "sql_text",
        "start_time",
        "thread_id",
        "user_host"
    ),
    "version": (
        "mysql_version",
        "sys_version"
    ),
    "innodb_buffer_stats_by_schema": (
        "allocated",
        "data",
        "object_schema",
        "pages",
        "pages_hashed",
        "pages_old",
        "rows_cached"
    ),
    "x$innodb_buffer_stats_by_schema": (
        "allocated",
        "data",
        "object_schema",
        "pages",
        "pages_hashed",
        "pages_old",
        "rows_cached"
    ),
    "innodb_buffer_stats_by_table": (
        "allocated",
        "data",
        "object_name",
        "object_schema",
        "pages",
        "pages_hashed",
        "pages_old",
        "rows_cached"
    ),
    "x$innodb_buffer_stats_by_table": (
        "allocated",
        "data",
        "object_name",
        "object_schema",
        "pages",
        "pages_hashed",
        "pages_old",
        "rows_cached"
    ),
    "schema_object_overview": (
        "count",
        "db",
        "object_type"
    ),
    "schema_auto_increment_columns": (
        "auto_increment",
        "auto_increment_ratio",
        "column_name",
        "column_type",
        "data_type",
        "is_signed",
        "is_unsigned",
        "max_value",
        "table_name",
        "table_schema"
    ),
    "x$schema_flattened_keys": (
        "index_columns",
        "index_name",
        "non_unique",
        "subpart_exists",
        "table_name",
        "table_schema"
    ),
    "schema_redundant_indexes": (
        "dominant_index_columns",
        "dominant_index_name",
        "dominant_index_non_unique",
        "redundant_index_columns",
        "redundant_index_name",
        "redundant_index_non_unique",
        "sql_drop_index",
        "subpart_exists",
        "table_name",
        "table_schema"
    ),
    "ps_check_lost_instrumentation": (
        "variable_name",
        "variable_value"
    ),
    "latest_file_io": (
        "file",
        "latency",
        "operation",
        "requested",
        "thread"
    ),
    "x$latest_file_io": (
        "file",
        "latency",
        "operation",
        "requested",
        "thread"
    ),
    "io_by_thread_by_latency": (
        "avg_latency",
        "max_latency",
        "min_latency",
        "processlist_id",
        "thread_id",
        "total",
        "total_latency",
        "user"
    ),
    "x$io_by_thread_by_latency": (
        "avg_latency",
        "max_latency",
        "min_latency",
        "processlist_id",
        "thread_id",
        "total",
        "total_latency",
        "user"
    ),
    "io_global_by_file_by_bytes": (
        "avg_read",
        "avg_write",
        "count_read",
        "count_write",
        "file",
        "total",
        "total_read",
        "total_written",
        "write_pct"
    ),
    "x$io_global_by_file_by_bytes": (
        "avg_read",
        "avg_write",
        "count_read",
        "count_write",
        "file",
        "total",
        "total_read",
        "total_written",
        "write_pct"
    ),
    "io_global_by_file_by_latency": (
        "count_misc",
        "count_read",
        "count_write",
        "file",
        "misc_latency",
        "read_latency",
        "total",
        "total_latency",
        "write_latency"
    ),
    "x$io_global_by_file_by_latency": (
        "count_misc",
        "count_read",
        "count_write",
        "file",
        "misc_latency",
        "read_latency",
        "total",
        "total_latency",
        "write_latency"
    ),
    "io_global_by_wait_by_bytes": (
        "avg_latency",
        "avg_read",
        "avg_written",
        "count_read",
        "count_write",
        "event_name",
        "max_latency",
        "min_latency",
        "total",
        "total_latency",
        "total_read",
        "total_requested",
        "total_written"
    ),
    "x$io_global_by_wait_by_bytes": (
        "avg_latency",
        "avg_read",
        "avg_written",
        "count_read",
        "count_write",
        "event_name",
        "max_latency",
        "min_latency",
        "total",
        "total_latency",
        "total_read",
        "total_requested",
        "total_written"
    ),
    "io_global_by_wait_by_latency": (
        "avg_latency",
        "avg_read",
        "avg_written",
        "count_read",
        "count_write",
        "event_name",
        "max_latency",
        "misc_latency",
        "read_latency",
        "total",
        "total_latency",
        "total_read",
        "total_written",
        "write_latency"
    ),
    "x$io_global_by_wait_by_latency": (
        "avg_latency",
        "avg_read",
        "avg_written",
        "count_read",
        "count_write",
        "event_name",
        "max_latency",
        "misc_latency",
        "read_latency",
        "total",
        "total_latency",
        "total_read",
        "total_written",
        "write_latency"
    ),
    "innodb_lock_waits": (
        "blocking_lock_id",
        "blocking_lock_mode",
        "blocking_pid",
        "blocking_query",
        "blocking_trx_age",
        "blocking_trx_id",
        "blocking_trx_rows_locked",
        "blocking_trx_rows_modified",
        "blocking_trx_started",
        "locked_index",
        "locked_table",
        "locked_table_name",
        "locked_table_partition",
        "locked_table_schema",
        "locked_table_subpartition",
        "locked_type",
        "sql_kill_blocking_connection",
        "sql_kill_blocking_query",
        "wait_age",
        "wait_age_secs",
        "wait_started",
        "waiting_lock_id",
        "waiting_lock_mode",
        "waiting_pid",
        "waiting_query",
        "waiting_trx_age",
        "waiting_trx_id",
        "waiting_trx_rows_locked",
        "waiting_trx_rows_modified",
        "waiting_trx_started"
    ),
    "x$innodb_lock_waits": (
        "blocking_lock_id",
        "blocking_lock_mode",
        "blocking_pid",
        "blocking_query",
        "blocking_trx_age",
        "blocking_trx_id",
        "blocking_trx_rows_locked",
        "blocking_trx_rows_modified",
        "blocking_trx_started",
        "locked_index",
        "locked_table",
        "locked_table_name",
        "locked_table_partition",
        "locked_table_schema",
        "locked_table_subpartition",
        "locked_type",
        "sql_kill_blocking_connection",
        "sql_kill_blocking_query",
        "wait_age",
        "wait_age_secs",
        "wait_started",
        "waiting_lock_id",
        "waiting_lock_mode",
        "waiting_pid",
        "waiting_query",
        "waiting_trx_age",
        "waiting_trx_id",
        "waiting_trx_rows_locked",
        "waiting_trx_rows_modified",
        "waiting_trx_started"
    ),
    "memory_by_user_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "total_allocated",
        "user"
    ),
    "x$memory_by_user_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "total_allocated",
        "user"
    ),
    "memory_by_host_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "host",
        "total_allocated"
    ),
    "x$memory_by_host_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "host",
        "total_allocated"
    ),
    "memory_by_thread_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "thread_id",
        "total_allocated",
        "user"
    ),
    "x$memory_by_thread_by_current_bytes": (
        "current_allocated",
        "current_avg_alloc",
        "current_count_used",
        "current_max_alloc",
        "thread_id",
        "total_allocated",
        "user"
    ),
    "memory_global_by_current_bytes": (
        "current_alloc",
        "current_avg_alloc",
        "current_count",
        "event_name",
        "high_alloc",
        "high_avg_alloc",
        "high_count"
    ),
    "x$memory_global_by_current_bytes": (
        "current_alloc",
        "current_avg_alloc",
        "current_count",
        "event_name",
        "high_alloc",
        "high_avg_alloc",
        "high_count"
    ),
    "memory_global_total": (
        "total_allocated"
    ),
    "x$memory_global_total": (
        "total_allocated"
    ),
    "schema_index_statistics": (
        "delete_latency",
        "index_name",
        "insert_latency",
        "rows_deleted",
        "rows_inserted",
        "rows_selected",
        "rows_updated",
        "select_latency",
        "table_name",
        "table_schema",
        "update_latency"
    ),
    "x$schema_index_statistics": (
        "delete_latency",
        "index_name",
        "insert_latency",
        "rows_deleted",
        "rows_inserted",
        "rows_selected",
        "rows_updated",
        "select_latency",
        "table_name",
        "table_schema",
        "update_latency"
    ),
    "x$ps_schema_table_statistics_io": (
        "count_misc",
        "count_read",
        "count_write",
        "sum_number_of_bytes_read",
        "sum_number_of_bytes_write",
        "sum_timer_misc",
        "sum_timer_read",
        "sum_timer_write",
        "table_name",
        "table_schema"
    ),
    "schema_table_statistics": (
        "delete_latency",
        "fetch_latency",
        "insert_latency",
        "io_misc_latency",
        "io_misc_requests",
        "io_read",
        "io_read_latency",
        "io_read_requests",
        "io_write",
        "io_write_latency",
        "io_write_requests",
        "rows_deleted",
        "rows_fetched",
        "rows_inserted",
        "rows_updated",
        "table_name",
        "table_schema",
        "total_latency",
        "update_latency"
    ),
    "x$schema_table_statistics": (
        "delete_latency",
        "fetch_latency",
        "insert_latency",
        "io_misc_latency",
        "io_misc_requests",
        "io_read",
        "io_read_latency",
        "io_read_requests",
        "io_write",
        "io_write_latency",
        "io_write_requests",
        "rows_deleted",
        "rows_fetched",
        "rows_inserted",
        "rows_updated",
        "table_name",
        "table_schema",
        "total_latency",
        "update_latency"
    ),
    "schema_table_statistics_with_buffer": (
        "delete_latency",
        "fetch_latency",
        "innodb_buffer_allocated",
        "innodb_buffer_data",
        "innodb_buffer_free",
        "innodb_buffer_pages",
        "innodb_buffer_pages_hashed",
        "innodb_buffer_pages_old",
        "innodb_buffer_rows_cached",
        "insert_latency",
        "io_misc_latency",
        "io_misc_requests",
        "io_read",
        "io_read_latency",
        "io_read_requests",
        "io_write",
        "io_write_latency",
        "io_write_requests",
        "rows_deleted",
        "rows_fetched",
        "rows_inserted",
        "rows_updated",
        "table_name",
        "table_schema",
        "update_latency"
    ),
    "x$schema_table_statistics_with_buffer": (
        "delete_latency",
        "fetch_latency",
        "innodb_buffer_allocated",
        "innodb_buffer_data",
        "innodb_buffer_free",
        "innodb_buffer_pages",
        "innodb_buffer_pages_hashed",
        "innodb_buffer_pages_old",
        "innodb_buffer_rows_cached",
        "insert_latency",
        "io_misc_latency",
        "io_misc_requests",
        "io_read",
        "io_read_latency",
        "io_read_requests",
        "io_write",
        "io_write_latency",
        "io_write_requests",
        "rows_deleted",
        "rows_fetched",
        "rows_inserted",
        "rows_updated",
        "table_name",
        "table_schema",
        "update_latency"
    ),
    "schema_tables_with_full_table_scans": (
        "latency",
        "object_name",
        "object_schema",
        "rows_full_scanned"
    ),
    "x$schema_tables_with_full_table_scans": (
        "latency",
        "object_name",
        "object_schema",
        "rows_full_scanned"
    ),
    "schema_unused_indexes": (
        "index_name",
        "object_name",
        "object_schema"
    ),
    "schema_table_lock_waits": (
        "blocking_account",
        "blocking_lock_duration",
        "blocking_lock_type",
        "blocking_pid",
        "blocking_thread_id",
        "object_name",
        "object_schema",
        "sql_kill_blocking_connection",
        "sql_kill_blocking_query",
        "waiting_account",
        "waiting_lock_duration",
        "waiting_lock_type",
        "waiting_pid",
        "waiting_query",
        "waiting_query_rows_affected",
        "waiting_query_rows_examined",
        "waiting_query_secs",
        "waiting_thread_id"
    ),
    "x$schema_table_lock_waits": (
        "blocking_account",
        "blocking_lock_duration",
        "blocking_lock_type",
        "blocking_pid",
        "blocking_thread_id",
        "object_name",
        "object_schema",
        "sql_kill_blocking_connection",
        "sql_kill_blocking_query",
        "waiting_account",
        "waiting_lock_duration",
        "waiting_lock_type",
        "waiting_pid",
        "waiting_query",
        "waiting_query_rows_affected",
        "waiting_query_rows_examined",
        "waiting_query_secs",
        "waiting_thread_id"
    ),
    "statement_analysis": (
        "avg_latency",
        "cpu_latency",
        "db",
        "digest",
        "err_count",
        "exec_count",
        "first_seen",
        "full_scan",
        "last_seen",
        "lock_latency",
        "max_controlled_memory",
        "max_latency",
        "max_total_memory",
        "query",
        "rows_affected",
        "rows_affected_avg",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "rows_sorted",
        "sort_merge_passes",
        "tmp_disk_tables",
        "tmp_tables",
        "total_latency",
        "warn_count"
    ),
    "x$statement_analysis": (
        "avg_latency",
        "cpu_latency",
        "db",
        "digest",
        "err_count",
        "exec_count",
        "exec_secondary_count",
        "first_seen",
        "full_scan",
        "last_seen",
        "lock_latency",
        "max_controlled_memory",
        "max_latency",
        "max_total_memory",
        "query",
        "rows_affected",
        "rows_affected_avg",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "rows_sorted",
        "sort_merge_passes",
        "tmp_disk_tables",
        "tmp_tables",
        "total_latency",
        "warn_count"
    ),
    "statements_with_errors_or_warnings": (
        "db",
        "digest",
        "error_pct",
        "errors",
        "exec_count",
        "first_seen",
        "last_seen",
        "query",
        "warning_pct",
        "warnings"
    ),
    "x$statements_with_errors_or_warnings": (
        "db",
        "digest",
        "error_pct",
        "errors",
        "exec_count",
        "first_seen",
        "last_seen",
        "query",
        "warning_pct",
        "warnings"
    ),
    "statements_with_full_table_scans": (
        "db",
        "digest",
        "exec_count",
        "first_seen",
        "last_seen",
        "no_good_index_used_count",
        "no_index_used_count",
        "no_index_used_pct",
        "query",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "total_latency"
    ),
    "x$statements_with_full_table_scans": (
        "db",
        "digest",
        "exec_count",
        "first_seen",
        "last_seen",
        "no_good_index_used_count",
        "no_index_used_count",
        "no_index_used_pct",
        "query",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "total_latency"
    ),
    "x$ps_digest_avg_latency_distribution": (
        "avg_us",
        "cnt"
    ),
    "x$ps_digest_95th_percentile_by_avg_us": (
        "avg_us",
        "percentile"
    ),
    "statements_with_runtimes_in_95th_percentile": (
        "avg_latency",
        "db",
        "digest",
        "err_count",
        "exec_count",
        "first_seen",
        "full_scan",
        "last_seen",
        "max_latency",
        "query",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "total_latency",
        "warn_count"
    ),
    "x$statements_with_runtimes_in_95th_percentile": (
        "avg_latency",
        "db",
        "digest",
        "err_count",
        "exec_count",
        "first_seen",
        "full_scan",
        "last_seen",
        "max_latency",
        "query",
        "rows_examined",
        "rows_examined_avg",
        "rows_sent",
        "rows_sent_avg",
        "total_latency",
        "warn_count"
    ),
    "statements_with_sorting": (
        "avg_rows_sorted",
        "avg_sort_merges",
        "db",
        "digest",
        "exec_count",
        "first_seen",
        "last_seen",
        "query",
        "rows_sorted",
        "sort_merge_passes",
        "sort_using_range",
        "sorts_using_scans",
        "total_latency"
    ),
    "x$statements_with_sorting": (
        "avg_rows_sorted",
        "avg_sort_merges",
        "db",
        "digest",
        "exec_count",
        "first_seen",
        "last_seen",
        "query",
        "rows_sorted",
        "sort_merge_passes",
        "sort_using_range",
        "sorts_using_scans",
        "total_latency"
    ),
    "statements_with_temp_tables": (
        "avg_tmp_tables_per_query",
        "db",
        "digest",
        "disk_tmp_tables",
        "exec_count",
        "first_seen",
        "last_seen",
        "memory_tmp_tables",
        "query",
        "tmp_tables_to_disk_pct",
        "total_latency"
    ),
    "x$statements_with_temp_tables": (
        "avg_tmp_tables_per_query",
        "db",
        "digest",
        "disk_tmp_tables",
        "exec_count",
        "first_seen",
        "last_seen",
        "memory_tmp_tables",
        "query",
        "tmp_tables_to_disk_pct",
        "total_latency"
    ),
    "user_summary_by_file_io_type": (
        "event_name",
        "latency",
        "max_latency",
        "total",
        "user"
    ),
    "x$user_summary_by_file_io_type": (
        "event_name",
        "latency",
        "max_latency",
        "total",
        "user"
    ),
    "user_summary_by_file_io": (
        "io_latency",
        "ios",
        "user"
    ),
    "x$user_summary_by_file_io": (
        "io_latency",
        "ios",
        "user"
    ),
    "user_summary_by_statement_type": (
        "cpu_latency",
        "full_scans",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "statement",
        "total",
        "total_latency",
        "user"
    ),
    "x$user_summary_by_statement_type": (
        "cpu_latency",
        "full_scans",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "statement",
        "total",
        "total_latency",
        "user"
    ),
    "user_summary_by_statement_latency": (
        "cpu_latency",
        "full_scans",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "total",
        "total_latency",
        "user"
    ),
    "x$user_summary_by_statement_latency": (
        "cpu_latency",
        "full_scans",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "total",
        "total_latency",
        "user"
    ),
    "user_summary_by_stages": (
        "avg_latency",
        "event_name",
        "total",
        "total_latency",
        "user"
    ),
    "x$user_summary_by_stages": (
        "avg_latency",
        "event_name",
        "total",
        "total_latency",
        "user"
    ),
    "user_summary": (
        "current_connections",
        "current_memory",
        "file_io_latency",
        "file_ios",
        "statement_avg_latency",
        "statement_latency",
        "statements",
        "table_scans",
        "total_connections",
        "total_memory_allocated",
        "unique_hosts",
        "user"
    ),
    "x$user_summary": (
        "current_connections",
        "current_memory",
        "file_io_latency",
        "file_ios",
        "statement_avg_latency",
        "statement_latency",
        "statements",
        "table_scans",
        "total_connections",
        "total_memory_allocated",
        "unique_hosts",
        "user"
    ),
    "host_summary_by_file_io_type": (
        "event_name",
        "host",
        "max_latency",
        "total",
        "total_latency"
    ),
    "x$host_summary_by_file_io_type": (
        "event_name",
        "host",
        "max_latency",
        "total",
        "total_latency"
    ),
    "host_summary_by_file_io": (
        "host",
        "io_latency",
        "ios"
    ),
    "x$host_summary_by_file_io": (
        "host",
        "io_latency",
        "ios"
    ),
    "host_summary_by_statement_type": (
        "cpu_latency",
        "full_scans",
        "host",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "statement",
        "total",
        "total_latency"
    ),
    "x$host_summary_by_statement_type": (
        "cpu_latency",
        "full_scans",
        "host",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "statement",
        "total",
        "total_latency"
    ),
    "host_summary_by_statement_latency": (
        "cpu_latency",
        "full_scans",
        "host",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "total",
        "total_latency"
    ),
    "x$host_summary_by_statement_latency": (
        "cpu_latency",
        "full_scans",
        "host",
        "lock_latency",
        "max_latency",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "total",
        "total_latency"
    ),
    "host_summary_by_stages": (
        "avg_latency",
        "event_name",
        "host",
        "total",
        "total_latency"
    ),
    "x$host_summary_by_stages": (
        "avg_latency",
        "event_name",
        "host",
        "total",
        "total_latency"
    ),
    "host_summary": (
        "current_connections",
        "current_memory",
        "file_io_latency",
        "file_ios",
        "host",
        "statement_avg_latency",
        "statement_latency",
        "statements",
        "table_scans",
        "total_connections",
        "total_memory_allocated",
        "unique_users"
    ),
    "x$host_summary": (
        "current_connections",
        "current_memory",
        "file_io_latency",
        "file_ios",
        "host",
        "statement_avg_latency",
        "statement_latency",
        "statements",
        "table_scans",
        "total_connections",
        "total_memory_allocated",
        "unique_users"
    ),
    "wait_classes_global_by_avg_latency": (
        "avg_latency",
        "event_class",
        "max_latency",
        "min_latency",
        "total",
        "total_latency"
    ),
    "x$wait_classes_global_by_avg_latency": (
        "avg_latency",
        "event_class",
        "max_latency",
        "min_latency",
        "total",
        "total_latency"
    ),
    "wait_classes_global_by_latency": (
        "avg_latency",
        "event_class",
        "max_latency",
        "min_latency",
        "total",
        "total_latency"
    ),
    "x$wait_classes_global_by_latency": (
        "avg_latency",
        "event_class",
        "max_latency",
        "min_latency",
        "total",
        "total_latency"
    ),
    "waits_by_user_by_latency": (
        "avg_latency",
        "event",
        "max_latency",
        "total",
        "total_latency",
        "user"
    ),
    "x$waits_by_user_by_latency": (
        "avg_latency",
        "event",
        "max_latency",
        "total",
        "total_latency",
        "user"
    ),
    "waits_by_host_by_latency": (
        "avg_latency",
        "event",
        "host",
        "max_latency",
        "total",
        "total_latency"
    ),
    "x$waits_by_host_by_latency": (
        "avg_latency",
        "event",
        "host",
        "max_latency",
        "total",
        "total_latency"
    ),
    "waits_global_by_latency": (
        "avg_latency",
        "events",
        "max_latency",
        "total",
        "total_latency"
    ),
    "x$waits_global_by_latency": (
        "avg_latency",
        "events",
        "max_latency",
        "total",
        "total_latency"
    ),
    "metrics": (
        "Enabled",
        "Type",
        "Variable_name",
        "Variable_value"
    ),
    "processlist": (
        "command",
        "conn_id",
        "cpu_latency",
        "current_memory",
        "current_statement",
        "db",
        "execution_engine",
        "full_scan",
        "last_statement",
        "last_statement_latency",
        "last_wait",
        "last_wait_latency",
        "lock_latency",
        "pid",
        "program_name",
        "progress",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "source",
        "state",
        "statement_latency",
        "thd_id",
        "time",
        "tmp_disk_tables",
        "tmp_tables",
        "trx_autocommit",
        "trx_latency",
        "trx_state",
        "user",
        "COMMAND",
        "DB",
        "EXECUTION_ENGINE",
        "HOST",
        "ID",
        "INFO",
        "STATE",
        "TIME",
        "USER"
    ),
    "x$processlist": (
        "command",
        "conn_id",
        "cpu_latency",
        "current_memory",
        "current_statement",
        "db",
        "execution_engine",
        "full_scan",
        "last_statement",
        "last_statement_latency",
        "last_wait",
        "last_wait_latency",
        "lock_latency",
        "pid",
        "program_name",
        "progress",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "source",
        "state",
        "statement_latency",
        "thd_id",
        "time",
        "tmp_disk_tables",
        "tmp_tables",
        "trx_autocommit",
        "trx_latency",
        "trx_state",
        "user"
    ),
    "session": (
        "command",
        "conn_id",
        "cpu_latency",
        "current_memory",
        "current_statement",
        "db",
        "execution_engine",
        "full_scan",
        "last_statement",
        "last_statement_latency",
        "last_wait",
        "last_wait_latency",
        "lock_latency",
        "pid",
        "program_name",
        "progress",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "source",
        "state",
        "statement_latency",
        "thd_id",
        "time",
        "tmp_disk_tables",
        "tmp_tables",
        "trx_autocommit",
        "trx_latency",
        "trx_state",
        "user"
    ),
    "x$session": (
        "command",
        "conn_id",
        "cpu_latency",
        "current_memory",
        "current_statement",
        "db",
        "execution_engine",
        "full_scan",
        "last_statement",
        "last_statement_latency",
        "last_wait",
        "last_wait_latency",
        "lock_latency",
        "pid",
        "program_name",
        "progress",
        "rows_affected",
        "rows_examined",
        "rows_sent",
        "source",
        "state",
        "statement_latency",
        "thd_id",
        "time",
        "tmp_disk_tables",
        "tmp_tables",
        "trx_autocommit",
        "trx_latency",
        "trx_state",
        "user"
    ),
    "session_ssl_status": (
        "ssl_cipher",
        "ssl_sessions_reused",
        "ssl_version",
        "thread_id"
    ),
    "rds_global_status_history": (
        "collection_end",
        "collection_start",
        "variable_delta",
        "variable_name",
        "variable_value"
    ),
    "rds_global_status_history_old": (
        "collection_end",
        "collection_start",
        "variable_delta",
        "variable_name",
        "variable_value"
    ),
    "general_log_template": (
        "argument",
        "command_type",
        "event_time",
        "server_id",
        "thread_id",
        "user_host"
    ),
    "slow_log_template": (
        "db",
        "insert_id",
        "last_insert_id",
        "lock_time",
        "query_time",
        "rows_examined",
        "rows_sent",
        "server_id",
        "sql_text",
        "start_time",
        "thread_id",
        "user_host"
    ),
    "rds_heartbeat2": (
        "id",
        "value"
    ),
    "rds_sysinfo": (
        "name",
        "value"
    ),
    "aurora_s3_load_history": (
        "bytes_loaded",
        "file_name",
        "load_prefix",
        "load_timestamp",
        "version_number"
    ),
    "general_log_backup": (
        "argument",
        "command_type",
        "event_time",
        "server_id",
        "thread_id",
        "user_host"
    ),
    "slow_log_backup": (
        "db",
        "insert_id",
        "last_insert_id",
        "lock_time",
        "query_time",
        "rows_examined",
        "rows_sent",
        "server_id",
        "sql_text",
        "start_time",
        "thread_id",
        "user_host"
    ),
    "workflow_state_old": (
        "id",
        "profile_id",
        "state_id",
        "timestamp"
    ),
    "user_status_table": (
        "id"
    ),
    "user_status_ml_table": (
        "id",
        "lang_code",
        "title",
        "user_status_id"
    ),
    "user_status_history_table": (
        "end_timestamp",
        "id",
        "start_timestamp",
        "user_status_id"
    ),
    "circlez_team_members": (
        "Email",
        "First Name",
        "Gender",
        "Last Name",
        "Number",
        "Roles",
        "Skills",
        "Team(s)",
        "Title"
    ),
    "profile_ml_view": (
        "lang_code",
        "profile_id",
        "title"
    ),
    "environment": (
        "id",
        "name"
    ),
    "deployment_status": (
        "id",
        "title_en"
    ),
    "deployment": (
        "branch",
        "deployment_status_id",
        "end_datetime",
        "environment_id",
        "id",
        "reason",
        "requestor_profile_id",
        "start_datetime"
    ),
    "user_quota": (
        "activity_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "hard_quota",
        "soft_quota",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id"
    ),
    "dialog_workflow_state_history_table_old": (
        "id",
        "profile_id",
        "state_id",
        "timestamp",
        "variable_id",
        "variable_value"
    ),
    "age_media": (
        "age",
        "filename",
        "id",
        "media"
    ),
    "presence": (
        "id"
    ),
    "presence_history": (
        "end_timestamp",
        "id",
        "presence_id",
        "start_timestamp",
        "user_id"
    ),
    "presence_ml": (
        "id",
        "lang_code",
        "presence_id",
        "title"
    ),
    "group_en_approved_main_title_view": (
        "group_id",
        "is_interest",
        "parent_group_id",
        "seq",
        "title"
    ),
    "table_deployments": (
        "environment_id",
        "id",
        "table_id",
        "timestamp",
        "version"
    ),
    "trigger_event": (
        "event_name",
        "id"
    ),
    "checklist_table": (
        "actual_timestamp",
        "checklist_template_id",
        "checklist_template_seq",
        "due_timestamp",
        "id",
        "owner_profile_id",
        "subject_profile_id",
        "task",
        "triger_event_id"
    ),
    "activity_trigger_event": (
        "activity_id",
        "id",
        "trigger_event"
    ),
    "user_list_table": (
        "id"
    ),
    "user_list_old": (
        "id"
    ),
    "user_list_member_table": (
        "id",
        "user_id",
        "user_list_id"
    ),
    "environment_en_view": (
        "environment_id",
        "name"
    ),
    "table_deployment": (
        "environment_id",
        "id",
        "schema",
        "table_id",
        "table_name",
        "version"
    ),
    "table_list_member": (
        "id",
        "table_id",
        "table_list_id"
    ),
    "checklists_templates_table": (
        "checklist_id",
        "checklist_name",
        "tasks"
    ),
    "table_old": (
        "entity_id",
        "id",
        "schema",
        "table_name"
    ),
    "logger_table_old": (
        "activity",
        "activity_id",
        "component_id",
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "id",
        "ipv4",
        "ipv6",
        "location_id",
        "payload",
        "profile_id",
        "severity_id",
        "start_datetime",
        "state_id",
        "timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "identity_table": (
        "email_address_id",
        "facebook_id",
        "id",
        "phone_id",
        "profile_id"
    ),
    "activity_list_member_table": (
        "activity_id",
        "activity_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "os": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "os_name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "mobile_device": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "os_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "version"
    ),
    "mobile_account": (
        "created_timestamp",
        "created_user_id",
        "device_id",
        "end_timestamp",
        "id",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "mobile_account_profile": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "mobile_account_id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "severity_table_old": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "id",
        "start_datetime",
        "updated_timestamp",
        "updated_user_id"
    ),
    "severity_table_m_old": (
        "description",
        "description_approved",
        "id",
        "lang_code",
        "severity_id"
    ),
    "os_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_desktop",
        "is_mobile_phone",
        "is_tv",
        "os_name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_group_location_occurrence_profile_ml_table": (
        "activity_location_occurrence_profile_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "lang_code",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "vacancy_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "vacancy_ml_en_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_name",
        "vacancy_name_approved"
    ),
    "employer_profile_vacancy_view": (
        "created_timestamp",
        "created_user_id",
        "employer_profile_id",
        "employer_profile_vacancy_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "employer_profile_profile_view": (
        "created_timestamp",
        "created_user_id",
        "employer_profile_id",
        "employer_profile_profile_id",
        "end_timestamp",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "service_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_provider_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_provider_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "service_provider_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_receiver_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_receiver_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "service_receiver_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_offer_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "service_offer_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_offer_location_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "extra_price",
        "id",
        "location_list_id",
        "service_offer_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_review_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "order_id",
        "reaction_id",
        "reaction_value",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_review_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "service_review_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_type_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "profile_media_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "location_field_ml_title_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "group_category_ml_title_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "group_category_ml_description_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "location_field_ml_field_name_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "person_ml_first_name_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "person_ml_last_name_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "contact_nickname_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "contact_website1_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "contact_website2_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "contact_website3_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "criteria_vacancy_view": (
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "criteria_vacancy_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "organization_approved_en_main_view": (
        "description",
        "organization_id",
        "parent_organization_id"
    ),
    "severity_view": (
        "created timestamp",
        "created user id",
        "severity id",
        "updated timestamp",
        "updated user id"
    ),
    "severity_ml_en_view": (
        "description",
        "language code",
        "severity id",
        "severity_ml_id"
    ),
    "computer_vision_storage_view": (
        "angry",
        "black",
        "computer_vision_storage_id",
        "coordinate",
        "created_timestamp",
        "disgust",
        "dominant_emotion",
        "dominant_gender",
        "end_timestamp",
        "fear",
        "gender_id",
        "gender_man",
        "gender_women",
        "happy",
        "json",
        "latino_hispanic",
        "location_id",
        "max_age",
        "middle eastern",
        "min_age",
        "profile_id",
        "region",
        "sad",
        "start_timestamp",
        "storage_id",
        "surprise",
        "time_end",
        "time_start",
        "updated_timestamp",
        "white",
        "x1",
        "x2",
        "y1",
        "y2"
    ),
    "text_category_table": (
        "category",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_ml_en_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "proxy_profile": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_email_table_old": (
        "created_timestamp",
        "created_user_id",
        "id",
        "label_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_table_old": (
        "approved",
        "approved_timestmp",
        "cira",
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "id",
        "profile_id",
        "start_datetime",
        "telephone",
        "telephone_label_id",
        "updated_timestamp",
        "updated_user_id",
        "verification_expiration",
        "verification_pin"
    ),
    "proxy_group_list_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "checklist_group_table": (
        "checklist_id",
        "group_id",
        "id"
    ),
    "product_deal_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "is_title_approved",
        "lang_code",
        "product_deal_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "status_table": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "id",
        "start_datetime",
        "updated_timestamp",
        "updated_user_id"
    ),
    "status_table_ml": (
        "description",
        "id",
        "lang_code",
        "status_id"
    ),
    "advertisement_profile": (
        "advertisment_id",
        "duration_ms",
        "id",
        "is_clicked",
        "profile_id"
    ),
    "advertisement_group_list_table": (
        "advertisement_id",
        "group_list_id",
        "id"
    ),
    "logger_group_profile_view": (
        "activity",
        "activity_id",
        "client_ip_v4",
        "client_ip_v6",
        "component name",
        "component_id",
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "error_stack",
        "group_id",
        "location_id",
        "logger_id",
        "message",
        "payload",
        "profile_id",
        "programming_language",
        "relationship_type_id",
        "server_ip_v4",
        "server_ip_v6",
        "severity_id",
        "start_datetime",
        "state_id",
        "status_id",
        "timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "logger_severity_level_table": (
        "component_id",
        "data_store",
        "environment",
        "environment_id",
        "id",
        "maximal_severity_level",
        "region",
        "region_id"
    ),
    "campaign_group_table": (
        "advertisement_id",
        "group_id",
        "id"
    ),
    "campaign_group_list_table": (
        "advertisement_id",
        "group_list_id",
        "id"
    ),
    "resource_ml_table": (
        "id",
        "language",
        "name",
        "resource_id"
    ),
    "resource_type_ml_table": (
        "id",
        "language",
        "name",
        "resource_type_id"
    ),
    "resource_type_table": (
        "activity_type",
        "id"
    ),
    "resource_instance_table": (
        "google_calendar",
        "id",
        "resource_id",
        "seq"
    ),
    "resource_instance_ml_table": (
        "id",
        "language",
        "name",
        "resource_instance_id"
    ),
    "shop_seller_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "role",
        "seller_id",
        "shop_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "user_age_range": (
        "age_max",
        "age_min",
        "user_id"
    ),
    "user_language": (
        "lang_code_id",
        "user_id"
    ),
    "group_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_interest",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "parent_group_id",
        "sequence",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_level_ml_table": (
        "description",
        "id",
        "is_description_approved",
        "is_name_approved",
        "lang_code",
        "name",
        "profile_level_id"
    ),
    "location_view": (
        "address_english",
        "address_local_language",
        "city_id",
        "coordinate",
        "country_id",
        "county_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "identifier",
        "is_approved",
        "is_community_active",
        "is_test_data",
        "location_id",
        "name",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "region_id",
        "start_timestamp",
        "state_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "label_question_table": (
        "created_timestamp",
        "created_user_id",
        "id",
        "label_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "media_type_table": (
        "id"
    ),
    "media_type_ml_table": (
        "id",
        "lang_code",
        "media_type_id",
        "title"
    ),
    "label_profile_table_old": (
        "created_timestamp",
        "created_user_id",
        "id",
        "label_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_profile_level_table": (
        "id",
        "profile_id",
        "profile_level_id"
    ),
    "logger_view": (
        "action_id",
        "activity",
        "activity_id",
        "api_type",
        "api_type_id",
        "client_ip_v4",
        "client_ip_v6",
        "component_category",
        "component_id",
        "component_name",
        "component_type",
        "computer_language",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "destination_emails",
        "developer_email",
        "end_datetime",
        "error_stack",
        "field_id",
        "field_value_new",
        "field_value_old",
        "function_name",
        "group_id",
        "group_id1",
        "group_id2",
        "is_test_data",
        "location_id",
        "logger_id",
        "message",
        "message_id",
        "message_template_id",
        "message_template_text_block_id",
        "payload",
        "process_id",
        "profile_id",
        "question_id",
        "recipient",
        "record",
        "relationship_type_id",
        "return_code",
        "return_message",
        "returned_value",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "severity_id",
        "smartlink_identifier",
        "source_email",
        "sql_formatted",
        "sql_parameters",
        "sql_statement",
        "start_datetime",
        "state_id",
        "status_id",
        "stderr",
        "stdout",
        "testing_framework",
        "thread_id",
        "timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "profile_approved_en_main_view": (
        "experience_years_min",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_id",
        "profile_type_id",
        "visibility_id"
    ),
    "profile_approved_en_view": (
        "experience_years_min",
        "id_profile",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_type_id",
        "visibility_id"
    ),
    "profile_ml_en_view": (
        "experience_years_min",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_id",
        "profile_type_id",
        "title",
        "visibility_id"
    ),
    "profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "identifier",
        "internal_description",
        "is_approved",
        "is_business_profile",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "main_phone_id",
        "number",
        "preferred_lang_code",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile_id",
        "profile_type_id",
        "stars",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "group_approved_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_category_id",
        "group_id",
        "is_group_approved",
        "is_interest",
        "is_job_title",
        "location_id",
        "location_list_id",
        "members_visibility_id",
        "non_members_visibility_id",
        "parent_group_id",
        "parent_group_list_id",
        "sequence",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_view": (
        "coordinate",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_profile_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "group_category_id",
        "group_id",
        "hashtag",
        "identifier",
        "is_activity",
        "is_approved",
        "is_building",
        "is_campaign",
        "is_city",
        "is_community",
        "is_continent",
        "is_country",
        "is_county",
        "is_event",
        "is_first_name",
        "is_geo",
        "is_group_approved",
        "is_interest",
        "is_job_title",
        "is_language",
        "is_last_name",
        "is_marital_status",
        "is_neighbourhood",
        "is_official",
        "is_organization",
        "is_reddit_group",
        "is_region",
        "is_relationship",
        "is_religion",
        "is_role",
        "is_skill",
        "is_sport",
        "is_state",
        "is_street",
        "is_test_data",
        "is_zip_code",
        "location_id",
        "location_list_id",
        "main_group_type_id",
        "members_visibility_id",
        "name",
        "new_group_id",
        "non_members_visibility_id",
        "number",
        "parent_group_id",
        "parent_group_list_id",
        "profile_id",
        "profile_list_id",
        "sequence",
        "start_timestamp",
        "system_group_name",
        "system_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "component_view": (
        "api_type",
        "component_category",
        "component_id",
        "component_type",
        "github_repo",
        "name",
        "parent_component_id",
        "programming_language",
        "programming_language_id",
        "repo_name",
        "testing_framework"
    ),
    "queue_item_view": (
        "action_id",
        "class_parameters_json",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "execution_end_time",
        "execution_start_time",
        "function_parameters_json",
        "identifier",
        "is_test_data",
        "item_description",
        "number",
        "process_id",
        "queue_item_id",
        "queue_status_id",
        "return_code",
        "return_json",
        "return_message",
        "returned_value",
        "server_ip_v4",
        "server_ip_v6",
        "start_timestamp",
        "stderr",
        "stdout",
        "thread_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt"
    ),
    "signature_table": (
        "is_long",
        "lang_code",
        "profile_id",
        "signature_id",
        "template_html"
    ),
    "importer_vacancy_view": (
        "Country",
        "created_timestamp",
        "data_source_type.title",
        "entity_type_id",
        "title",
        "vacancy_name"
    ),
    "business_profile_view": (
        "contact_id",
        "Country",
        "data_imported_json",
        "data_source.title",
        "department",
        "email1",
        "entity_type_id",
        "entity_type_ml.title",
        "full_name",
        "job_title",
        "name_prefix",
        "notes",
        "organization",
        "phone1"
    ),
    "operational_hours_view": (
        "created_timestamp",
        "created_user_id",
        "day_of_week",
        "end_timestamp",
        "from_time",
        "is_business_hours",
        "is_test_data",
        "location_id",
        "occurence_id",
        "operational_hours_id",
        "operational_hours_type",
        "profile_id",
        "start_timestamp",
        "until_time",
        "updated_timestamp",
        "updated_user_id"
    ),
    "dialog_workflow_state_view": (
        "dialog_workflow_state_id",
        "lang_code",
        "message_template_id",
        "next_state_id",
        "next_state_id_if_there_is_no_feedback",
        "no_feedback_milliseconds",
        "parameter1",
        "parent_state_id",
        "result_figure_max",
        "result_figure_min",
        "result_logical",
        "state_id",
        "variable1_id",
        "workflow_action_id"
    ),
    "email_address_view": (
        "created_timestamp",
        "created_user_id",
        "email_address",
        "email_address_approved",
        "email_address_id",
        "end_timestamp",
        "is_test_data",
        "is_verified",
        "number",
        "profile_id_old",
        "start_timestamp",
        "text_block_id",
        "updated_timestamp",
        "updated_user_id",
        "verification_expiration",
        "verification_pin"
    ),
    "location_profile_view": (
        "created_effective_profile_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "location_id",
        "location_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "group_profile_view": (
        "consumer_category_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_year",
        "end_timestamp",
        "group_id",
        "group_profile_id",
        "group_profile_type_id",
        "is_test_data",
        "participant_category_id",
        "profile_id",
        "relationship_type_id",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "supplier_category_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "file_type_view": (
        "file_type",
        "file_type_id"
    ),
    "storage_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "file_extension_id",
        "file_type_id",
        "filename",
        "height",
        "is_test_data",
        "path",
        "ratio_x",
        "ration_y",
        "region",
        "start_timestamp",
        "storage_id",
        "storage_type_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "width"
    ),
    "location_profile_ml_view": (
        "created_effective_profile_id",
        "lang_code",
        "location_id",
        "location_profile_id",
        "profile_id",
        "title",
        "visibility_id"
    ),
    "profile_reaction_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "from_profile_id",
        "profile_reaction_id",
        "reaction_id",
        "start_timestamp",
        "to_profile_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_orm_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "test_orm_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "community_waiting_list_table": (
        "community_waiting_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "position",
        "profile_id",
        "start_timestamp",
        "timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "age_group_table": (
        "age_group_id",
        "max_age",
        "min_age",
        "age_group_id",
        "max_age",
        "min_age",
        "name"
    ),
    "age_group_group_table": (
        "age_group_group_id",
        "age_group_id",
        "group_id",
        "sequence"
    ),
    "visibility_table": (
        "isEnable",
        "visibility_id"
    ),
    "visibility_ml_table": (
        "lang_code",
        "title",
        "visibility_id",
        "visibility_ml_id"
    ),
    "community_waiting_list_view": (
        "community_waiting_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "position",
        "profile_id",
        "start_timestamp",
        "timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "field_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "database_field_name",
        "database_sub_field_name",
        "database_sub_field_value",
        "description",
        "end_timestamp",
        "field_id",
        "identifier",
        "internal_description",
        "is_ml_table",
        "is_test_data",
        "name",
        "number",
        "processing_database_field_name",
        "processing_id",
        "start_timestamp",
        "table_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "variable_id",
        "visibility_id"
    ),
    "profile_profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "is_potential",
        "is_system_recommendation",
        "is_test_data",
        "is_user_request",
        "job_title",
        "number",
        "profile_id1",
        "profile_id2",
        "profile_profile_id",
        "profile_profile_role_id",
        "relationship_type_id",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "system_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "user_external_level"
    ),
    "profile_user_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "profile_id",
        "profile_user_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id"
    ),
    "person_view": (
        "birthday_date",
        "birthday_original",
        "birthday_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "day",
        "end_timestamp",
        "father_name",
        "first_name",
        "gender_id",
        "identifier",
        "is_approved",
        "is_first_name_approved",
        "is_identity_confirmed",
        "is_last_name_approved",
        "is_last_name_auto_generated",
        "is_last_name_sure",
        "is_nickname_approved",
        "is_rip",
        "is_system_data",
        "is_test_data",
        "last_coordinate",
        "last_location_id",
        "last_name",
        "last_name_original",
        "location_id",
        "main_full_number_normalized",
        "month",
        "name",
        "nickname",
        "number",
        "person.main_email_address",
        "person.main_email_address_id",
        "person.main_profile_id",
        "person.preferred_lang_code",
        "person_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "year",
        "year_cira"
    ),
    "action_view": (
        "action_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "filename",
        "folder_name",
        "function_module",
        "function_name",
        "is_queue_action",
        "is_smartlink_action",
        "name",
        "package_name",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_view": (
        "area_code",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "extension",
        "full_number_normalized",
        "international_code",
        "is_test_data",
        "local_number_normalized",
        "number",
        "number_original",
        "phone_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "package_bank_table": (
        "amount",
        "invoice_id",
        "message_channel_id",
        "package_bank_id",
        "package_id"
    ),
    "profile_account_table": (
        "profile_account_id",
        "profile_id"
    ),
    "invoice_table": (
        "campaign_id",
        "from_profile_account_id",
        "invoice_id",
        "to_profile_accoint_id",
        "usd"
    ),
    "package_template_bank_table": (
        "amount",
        "message_channel_id",
        "package_template_bank_id",
        "package_template_id"
    ),
    "package_table": (
        "created_timestamp",
        "created_user_id",
        "effectove_user_id",
        "end_timestamp",
        "invoice_id",
        "package_id",
        "package_template_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "package_template_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "package_template_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "package_template_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "package_template_id",
        "package_template_ml_id",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "package_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "package_id",
        "package_ml_id",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_view": (
        "account_name",
        "additional_name",
        "address1_city",
        "address1_country",
        "address1_postal_code",
        "address1_state",
        "address1_street",
        "address2_city",
        "address2_country",
        "address2_postal_code",
        "address2_state",
        "address2_street",
        "anniversary",
        "birthday",
        "cira",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "day",
        "department",
        "display_as",
        "email1",
        "email2",
        "email3",
        "end_timestamp",
        "first_name",
        "full_name",
        "identifier",
        "is_test_data",
        "job_title",
        "job_title_id",
        "job_title_ml_id",
        "last_name",
        "last_sync_timestamp",
        "main_profile_id",
        "month",
        "name_prefix",
        "name_suffix",
        "nickname",
        "notes",
        "number",
        "organization",
        "organization_profile_id",
        "original_first_name",
        "original_last_name",
        "owner_profile_id",
        "person_id",
        "phone1",
        "phone2",
        "phone3",
        "photo_file_name",
        "photo_url",
        "source",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "uploaded_timestamp",
        "visibility_id",
        "website1",
        "website2",
        "website3",
        "year"
    ),
    "entity_type_view": (
        "created_timestamp",
        "created_user_id",
        "criteria_schema_name",
        "criteria_table_name",
        "criteria_view_name",
        "end_timestamp",
        "entity_type_id",
        "is_item_type",
        "is_test_data",
        "name",
        "schema_name",
        "start_timestamp",
        "system_id",
        "table_name",
        "updated_timestamp",
        "updated_user_id",
        "view_name",
        "view_with_deleted_and_test_data"
    ),
    "entity_type_ml_en_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_name",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_ml_view": (
        "coordinate",
        "country_id",
        "is_title_approved",
        "iso",
        "iso3",
        "lang_code",
        "name",
        "nicename",
        "numcode",
        "phonecode",
        "title"
    ),
    "county_ml_view": (
        "coordinate",
        "county_id",
        "is_title_approved",
        "lang_code",
        "state_id",
        "title"
    ),
    "neighborhood_ml_view": (
        "city_id",
        "group.name",
        "group_id",
        "lang_code",
        "name",
        "neighborhood.end_timestamp",
        "neighborhood_id",
        "title"
    ),
    "region_ml_view": (
        "country_id",
        "group.name",
        "group_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "region_id",
        "title"
    ),
    "state_ml_view": (
        "is_title_approved",
        "lang_code",
        "state_id",
        "title"
    ),
    "accommodation_table": (
        "accommodation_id"
    ),
    "accommodation_deal_table": (
        "accommodation_deal_id",
        "accommodation_id",
        "end_date",
        "hospitlity",
        "room_type",
        "start_date"
    ),
    "advertisement_type_table": (
        "advertisement_type_id"
    ),
    "event_price_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "description_approved",
        "event_price_id",
        "event_price_ml_id",
        "lang_code",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_view": (
        "access_start_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "facebook_event_url",
        "genre",
        "identifier",
        "is_can_be_paid_at_the_entrance_in_cash",
        "is_event_approved",
        "is_family",
        "is_paid",
        "is_require_end_user_arrival_confirmation",
        "is_require_end_user_registration",
        "is_require_organizer_registration_confirmation",
        "is_show_end_timestamp",
        "is_test_data",
        "location_id",
        "meetup_event_url",
        "name",
        "new_event_id",
        "number",
        "organizers_profile_id",
        "parent_event_id",
        "registration_url",
        "segement",
        "start_timestamp",
        "sub_genre",
        "sub_type",
        "timezone",
        "type",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "website_url"
    ),
    "event_ml_view": (
        "event_id",
        "lang_code",
        "title"
    ),
    "variable_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "field_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "person_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "visibility_id"
    ),
    "api_limit_view": (
        "api_limit_id",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "external_limit_code",
        "external_limit_name",
        "hard_limit_unit",
        "hard_limit_value",
        "is_test_data",
        "region",
        "soft_limit_unit",
        "soft_limit_value",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url_to_see_the limit",
        "user_external_id"
    ),
    "api_call_view": (
        "api_call_id",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "endpoint_url",
        "http_status_code",
        "incoming_message_json",
        "is_network",
        "is_test_data",
        "number",
        "outgoing_body_json",
        "outgoing_body_significant_fields_hash",
        "outgoing_header_json",
        "response_body_json",
        "session_token",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "api_type_view": (
        "action_id",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "endpoint_url",
        "expiration_unit",
        "expiration_value",
        "from_subsystem_id",
        "from_system_id",
        "github_repo",
        "hander_function_name",
        "http_method",
        "interval_max_seconds",
        "interval_min_seconds",
        "is_enabled",
        "is_test_data",
        "message_channel_provider_id",
        "name",
        "start_timestamp",
        "status_code_when_not_found",
        "to_subsystem_id",
        "to_system_id",
        "type",
        "updated_timestamp",
        "updated_user_id",
        "usage_and_billing_url"
    ),
    "api_type_field_view": (
        "api_type_field_id",
        "api_type_id",
        "field_name",
        "field_significant"
    ),
    "criteria_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "entity_type_id",
        "group_id",
        "group_list_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "location_id",
        "location_list_id",
        "name",
        "organization_name",
        "profile_list_id",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_timestamp",
        "template_condition",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id"
    ),
    "recruitment_agency_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "recruitment_agency_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_recruitment_agency": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "profile_id",
        "profile_recruitment_agency_id",
        "recruitment_agency_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_recruitment_agency": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_recruitment_agency_id",
        "recruitment_agency_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "occurrence_list_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "occurrence_list_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "occurrence_list_occurrence_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "occurrence_id",
        "occurrence_list_id",
        "occurrence_list_occurrence_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "occurrence_table": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "end_timestamp",
        "ends_on",
        "is_all_day",
        "num_occurrences",
        "occurrence_id",
        "repeat_every",
        "repeat_on_friday",
        "repeat_on_monday",
        "repeat_on_saturday",
        "repeat_on_sunday",
        "repeat_on_thursday",
        "repeat_on_tuesday",
        "repeat_on_wednesday",
        "repeat_unit",
        "start_datetime",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "consumer_category_table": (
        "consumer_category_id"
    ),
    "participant_category_table": (
        "participant_category_id"
    ),
    "api_type_field_table": (
        "api_type_field_id",
        "api_type_id",
        "field_name",
        "field_significant"
    ),
    "message_outbox_view": (
        "action_id",
        "campaign_id",
        "class_parameters_json",
        "component_id",
        "compound_message",
        "compound_message_json",
        "compound_message_json_version",
        "created_timestamp",
        "exection_end_time",
        "exection_start_time",
        "function_parameters_json",
        "identifier",
        "is_moderator_approved",
        "is_require_moderator",
        "is_test_data",
        "label_id",
        "message_id",
        "message_label_title",
        "number",
        "process_id",
        "queue_status_id",
        "queue_status_title",
        "return_code",
        "server_ip_v4",
        "server_ip_v6",
        "stderr",
        "stdout",
        "thread_id",
        "updated_timestamp",
        "user_jwt"
    ),
    "message_sent_view": (
        "actual_send_timestamp",
        "identifier",
        "label_id",
        "message_id",
        "number",
        "visibility_id"
    ),
    "importer_per_entity_type_view": (
        "count(importer.importer_table.entity_type_id)",
        "created_timestamp",
        "entity_type_id",
        "entity_type_ml.title"
    ),
    "visibility_id_values_view": (
        "isEnable",
        "title",
        "visibility_id"
    ),
    "relationship_classification_table_old": (
        "id"
    ),
    "profile_metrics_type_table": (
        "profile_metrics_type_id",
        "system_id"
    ),
    "gender_ml_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "gender_id",
        "gender_ml_id",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_url_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_url_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_id",
        "url_old",
        "url_type_id_old"
    ),
    "text_block_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "fields_extracted_json",
        "identifier",
        "number",
        "profile_id",
        "seq",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text",
        "text_block_id",
        "text_block_type_id",
        "text_hash",
        "text_without_empty_lines",
        "text_without_empty_lines_hash",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "visibility_id"
    ),
    "text_block_type_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "regex",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text_block_type_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_regex_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "lang_code",
        "regex",
        "start_timestamp",
        "text_block_type_id",
        "text_block_type_regex_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_ml_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_description_approved",
        "is_name_approved",
        "lang_code",
        "name",
        "start_timestamp",
        "text_block_type_id",
        "text_block_type_ml_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_ml_en_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_description_approved",
        "is_name_approved",
        "lang_code",
        "name",
        "regex",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text_block_type_id",
        "text_block_type_ml_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_ml_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_description_approved",
        "is_name_approved",
        "lang_code",
        "name",
        "regex",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text_block_type_id",
        "text_block_type_ml_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_metrics_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_metrics_id",
        "profile_metrics_type",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "replication_group_configuration_version": (
        "name",
        "version"
    ),
    "ro_replica_status_internal": (
        "Active_lsn",
        "Average_replay_latency_in_usec",
        "Cpu",
        "Current_read_lsn",
        "Current_replay_latency_in_usec",
        "Durable_lsn",
        "Highest_lsn_received",
        "Iops",
        "Is_current",
        "Last_error_timestamp",
        "Last_transport_error",
        "Last_updated_timestamp",
        "Log_buffer_sequence_number",
        "Log_stream_speed_in_KiB_per_second",
        "Max_replay_latency_in_usec",
        "Oldest_read_view_lsn",
        "Oldest_read_view_trx_id",
        "Pending_Read_IOs",
        "Read_IOs",
        "Replica_lag_in_msec",
        "Server_id",
        "Session_id",
        "Writer_reader_latency_in_usec"
    ),
    "ro_replica_status": (
        "Active_lsn",
        "Average_replay_latency_in_usec",
        "Cpu",
        "Current_read_lsn",
        "Current_replay_latency_in_usec",
        "Durable_lsn",
        "Highest_lsn_received",
        "Iops",
        "Is_current",
        "Last_error_timestamp",
        "Last_transport_error",
        "Last_updated_timestamp",
        "Log_buffer_sequence_number",
        "Log_stream_speed_in_KiB_per_second",
        "Max_replay_latency_in_usec",
        "Oldest_read_view_lsn",
        "Oldest_read_view_trx_id",
        "Pending_Read_IOs",
        "Read_IOs",
        "Replica_lag_in_msec",
        "Server_id",
        "Session_id",
        "Writer_reader_latency_in_usec"
    ),
    "channel_provider_table": (
        "channel_provider_id",
        "message_channel_id",
        "message_provider_id"
    ),
    "subscription_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "subscription_id",
        "subscription_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "employer_profile_profile_table": (
        "created_timestamp",
        "created_user_id",
        "employer_name",
        "employer_profile_id",
        "employer_profile_profile_id",
        "end_timestamp",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_promotional_message_to_person_table": (
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "criteria_promotional_message_to_person_id",
        "end_timestamp",
        "multiply_from_yesterday",
        "promotional_message_to_person_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "buyer_ml_table": (
        "buyer_id",
        "buyer_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "buyer_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "group_category_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "group_category_id",
        "group_category_ml_id",
        "is_main_title",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_vacancy_table": (
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "criteria_vacancy_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "vacancy_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "test_location_profile_mapping_ml_table": (
        "lang_code",
        "test_location_profile_id",
        "test_location_profile_ml_id",
        "title",
        "title_approved"
    ),
    "job_title_vacancy_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title_id",
        "job_title_vacancy_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "vacancy_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id",
        "vacancy_ml_id",
        "vacancy_name",
        "vacancy_name_approved"
    ),
    "test_typeorm_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "test_typeorm_id",
        "test_typeorm_ml_id",
        "title",
        "unique_int_column",
        "updated_timestamp",
        "updated_user_id"
    ),
    "base_entity_table": (
        "base_entity_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_sent_table": (
        "message_channel_id",
        "message_id",
        "message_sent_id",
        "request_to_send_timestamp",
        "sender_profile_id",
        "sent_timestamp"
    ),
    "message_template_table": (
        "message_template_id",
        "name"
    ),
    "advertisement_message_template_table": (
        "advertisement_id",
        "advertisement_message_template_id",
        "message_template_id"
    ),
    "location_news_table": (
        "location_id",
        "location_news_id",
        "news_id"
    ),
    "profile_preference_group_table": (
        "name",
        "profile_preference_group_id"
    ),
    "profile_preference_group_group_table": (
        "profile_preference_group_group_id",
        "profile_preference_group_id",
        "profile_preferences_id"
    ),
    "profile_preference_group_ml_table": (
        "description",
        "is_main",
        "lang_code",
        "name",
        "profile_preference_group_id",
        "profile_preference_group_ml_id"
    ),
    "profile_preference_ml_table": (
        "description",
        "is_main",
        "lang_code",
        "name",
        "profile_preference_id",
        "profile_preference_ml_id"
    ),
    "sport_event_ml_table": (
        "description",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "lang_code",
        "name",
        "sport_event_id",
        "sport_event_ml_id"
    ),
    "sport_event_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_description_approved",
        "is_name_approve",
        "name",
        "sport_event_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "news_table": (
        "news_id"
    ),
    "news_ml_table": (
        "content_html",
        "content_text",
        "is_context_html_approved",
        "is_context_text_approved",
        "is_title_approved",
        "lang_code",
        "news_id",
        "news_ml_id",
        "title"
    ),
    "table_history_table": (
        "datetime",
        "is_successful",
        "MAX_EXECUTION_TIME Miliseconds",
        "MySQL Workbench Duration Seconds",
        "MySQL Workbench Seconds",
        "rows",
        "schema",
        "sql_statement",
        "table",
        "table_history_id"
    ),
    "relationship_type_view": (
        "created_timestamp",
        "created_user_id",
        "is_direct_connection",
        "is_directional",
        "is_user_level",
        "party1_gender_id",
        "party1_older_max_years",
        "party1_older_min_years",
        "party1_profile_category_id",
        "party2_gender_id",
        "party2_profile_category_id",
        "past_present_future",
        "profile1_older_than_profile2",
        "relationship_category_id",
        "relationship_type_classification_id",
        "relationship_type_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "logger_function_start_end_view": (
        "component.component_name",
        "component.developer_email",
        "component_id",
        "created_timestamp",
        "function_name",
        "logger.component_name",
        "logger.developer_email",
        "logger_id",
        "process_id",
        "session",
        "severity_description",
        "severity_id",
        "thread_id"
    ),
    "profile_category_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "profile_category_id",
        "profile_category_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_category_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "profile_category_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_view": (
        "campaign_id",
        "days_of_week",
        "end_hour",
        "end_timestamp",
        "is_test_data",
        "max_audience",
        "max_exposure_per_day",
        "message_template_id_old",
        "minimal_days_between_messages_to_the_same_recipient",
        "name",
        "occurrence_id",
        "start_hour",
        "start_timestamp"
    ),
    "group_storage_blob_table": (
        "created_timestamp",
        "created_user_id",
        "group_storage_id",
        "image",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_storage_table": (
        "created_timestamp",
        "created_user_id",
        "file_type_id",
        "group_id",
        "group_storage_id",
        "storage_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_star_subscription_view": (
        "action_id",
        "action_star_subscription_id",
        "action_stars",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "file_type_id",
        "is_conditional",
        "is_test_data",
        "number",
        "start_timestamp",
        "subscription_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_user_view": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "end_timestamp",
        "start_datetime",
        "start_timestamp",
        "subscription_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_subscription_id"
    ),
    "phone_ml_table": (
        "created_timestamp",
        "created_user_id",
        "lang_code",
        "phone_id",
        "phone_ml_id",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "search_result_table": (
        "entity_type",
        "entity_type_id",
        "result_json",
        "result_text",
        "search_id",
        "search_result_id"
    ),
    "search_request_table": (
        "search_request_id",
        "search_request_json",
        "search_request_text"
    ),
    "importer_recent_view": (
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type_id",
        "entity_type_ml.title",
        "importer_country_name",
        "importer_id"
    ),
    "user_email_address_view": (
        "email_address_id",
        "email_address_ml.title",
        "email_address_profile.profile_id",
        "profile_id",
        "user.created_timestamp",
        "user.first_name",
        "user.last_name",
        "user.main_email_address",
        "user_id"
    ),
    "user_general_view": (
        "brand_id",
        "email_address",
        "last_login_timestamp",
        "person.birthday_date",
        "person.is_approved",
        "person.main_email_address",
        "person_id",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile.phone.full_number_normalized",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "role_name",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.active_location_id",
        "user.created_timestamp",
        "user.first_name",
        "user.is_approved",
        "user.is_test_data",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_id",
        "username"
    ),
    "message_type_ml_table": (
        "body",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "message_type_id",
        "message_type_ml_id",
        "start_timestamp",
        "subeject",
        "type_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_type_table": (
        "channel_type_id",
        "message_type_table_id",
        "version"
    ),
    "message_read_table": (
        "channel_id",
        "message_id",
        "message_read_id",
        "profile_id",
        "timestamp"
    ),
    "message_content_ml_table": (
        "body",
        "channel_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "message_content_id",
        "message_content_ml_id",
        "start_timestamp",
        "subeject",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_content_table": (
        "content_type_id",
        "message_content_id",
        "message_id",
        "version"
    ),
    "relationship_type_classification_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "relationship_type_classification_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "relationship_type_classification_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "relationship_type_classification_id",
        "relationship_type_classification_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "profile_type_id",
        "profile_type_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_list_profile_table_2delete": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "profile_id",
        "profile_list_id",
        "profile_list_profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_user_table": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "end_timestamp",
        "start_datetime",
        "start_timestamp",
        "subscription_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_subscription_id"
    ),
    "profile_media_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "image",
        "media_licenses_id",
        "media_source_url",
        "media_type_id",
        "profile_id",
        "profile_media_id",
        "start_timestamp",
        "storage_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_media_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "lang_code",
        "profile_media_id",
        "profile_media_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_external_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "event_external_id",
        "external_event_identifier",
        "is_approved",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id",
        "url"
    ),
    "user_external_view": (
        "access_token_old",
        "additional_fields_json",
        "base_credentials_token_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "credential_storage_id",
        "customer_id",
        "description",
        "email_address_id",
        "end_timestamp",
        "environment_id_old",
        "expiry_old",
        "handle",
        "identifier",
        "is_refresh_token_valid",
        "is_test_data",
        "main_profile_id",
        "number",
        "oauth_callback_confirmed",
        "oauth_state",
        "oauth_token",
        "oauth_token_secret",
        "project_id",
        "refresh_token_old",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "token_2delete",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_id",
        "user_external_id",
        "username"
    ),
    "message_outgoing_table_shall_we_delete_it": (
        "actual_sent_timestamp",
        "bcc_profile_id",
        "cc_profile_id",
        "from_profile_id",
        "message_channel_id",
        "message_outgoing_id",
        "message_type_id",
        "request_to_send_timestamp",
        "scheduled_sent_timestamp",
        "sending_message_result_id",
        "timestamp",
        "to_profile_id"
    ),
    "profile_user_external_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "profile_id",
        "profile_user_external_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "activity_ml_table_old": (
        "activity_id",
        "activity_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_list_table": (
        "activity_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_star_subscription_table": (
        "activity_id",
        "activity_star_subscription_id",
        "activity_stars",
        "subscription_id"
    ),
    "language_user_table": (
        "lang_code",
        "language_user_id",
        "language_user_type",
        "rank",
        "user_id"
    ),
    "user_preference_table": (
        "is_active",
        "preferences_internal_name",
        "user_id",
        "user_preference_id",
        "value"
    ),
    "preference_table": (
        "default_value",
        "is_active",
        "preference_id",
        "preferences_internal_name"
    ),
    "profile_preference_table": (
        "is_active",
        "preferences_internal_name",
        "profile_id",
        "profile_preference_id",
        "value"
    ),
    "label_message_view": (
        "label_id",
        "label_message_id",
        "message_id"
    ),
    "queue_status_ml_table": (
        "lang_code",
        "queue_status_id",
        "queue_status_ml_id",
        "title"
    ),
    "api_parameter_table": (
        "api_parameters_id",
        "api_type_id",
        "name",
        "type"
    ),
    "currency_ml_table": (
        "created_timestamp",
        "created_user_id",
        "currency_id",
        "currency_ml_id",
        "end_timestamp",
        "is_main",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "api_budget_table": (
        "api_budget_id",
        "priceset_id",
        "subsystem_id",
        "system_id",
        "text",
        "unit",
        "url"
    ),
    "logger_dialog_workflow_state_history_view": (
        "created_timestamp",
        "profile_id",
        "state_id",
        "timestamp",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "dialog_workflow_state_history_view_old": (
        "profile_id",
        "state_id",
        "timestamp",
        "variable_id",
        "variable_value_new"
    ),
    "smartlink_view": (
        "action_id",
        "campaign_id",
        "clicked_timestamp",
        "content_message_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id",
        "end_timestamp",
        "from_contact_id",
        "from_email_address_id",
        "from_email_address_old",
        "from_normalized_phone",
        "from_person_id",
        "from_phone_id",
        "from_profile_id",
        "from_user_id",
        "identifier",
        "is_test_data",
        "lang_code",
        "message_template_id",
        "number",
        "smartlink_id",
        "smartlink_message_id",
        "smartlink_type_id",
        "start_timestamp",
        "to_contact_id",
        "to_email_address_id",
        "to_email_address_old",
        "to_group_id",
        "to_normalized_phone_old",
        "to_person_id",
        "to_phone_id",
        "to_profile_id",
        "to_user_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_redirect"
    ),
    "city_ml_view": (
        "city_id",
        "city_ml_id",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "is_main",
        "lang_code",
        "location_city_name",
        "name",
        "phonecode",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "price_set_range_table": (
        "price_set_range_id"
    ),
    "product_deal_table": (
        "created_timestamp",
        "created_user_id",
        "currency_id",
        "end_timestamp",
        "max_price_set_id",
        "min_price_set_id",
        "price_set_id",
        "product_deal_id",
        "product_id",
        "quantity",
        "regular_price",
        "shop_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url"
    ),
    "event_status_table": (
        "event_status_id"
    ),
    "event_status_ml_table": (
        "created_timestamp",
        "created_user_id",
        "event_status_id",
        "event_status_ml_id",
        "lang_code",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "product_table": (
        "product_id"
    ),
    "product_url_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "product_id",
        "product_url_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "url_type_id"
    ),
    "event_storage_table": (
        "event_id",
        "event_storage_id",
        "is_official",
        "storage_id"
    ),
    "message_channel_provider_table": (
        "message_channel_provider_id",
        "provider_name",
        "supplier_id"
    ),
    "message_channel_provider_ml_table": (
        "lang_code",
        "message_channel_provider_ml_id",
        "message_provider_id",
        "message_provider_name"
    ),
    "person_kid_table": (
        "number_of_kids",
        "person_id",
        "person_kids_id"
    ),
    "phone_profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "label_phone_id",
        "phone_id",
        "phone_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_profile_main_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "label_phone_id",
        "phone_id",
        "phone_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_status_table": (
        "campaign_status_id"
    ),
    "campaign_ml_table": (
        "campaign_id",
        "campaign_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_status_ml_table": (
        "campaign_status_id",
        "campaign_status_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "role_user_table": (
        "role_id",
        "role_user_id",
        "user_id"
    ),
    "workflow_action_ml_view": (
        "lang_code",
        "title",
        "workflow_action_id",
        "workflow_action_ml_id"
    ),
    "field_alias_view": (
        "field_alias_id",
        "field_id",
        "field_name",
        "subsystem_id",
        "system_id",
        "visibility_id"
    ),
    "email_address_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "email_address_profile_id",
        "end_timestamp",
        "is_main",
        "is_sure",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_profile_verification_table": (
        "phone_profile_id",
        "phone_profile_verification_id"
    ),
    "test_mysql_ml_view": (
        "created_effective_profile_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_mysql_id",
        "test_mysql_ml_id",
        "title",
        "unique_int_column",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_mysql_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "point",
        "polygon",
        "start_timestamp",
        "test_mysql_id",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "event_event_external_table": (
        "created_event_id",
        "created_timestamp",
        "end_timestamp",
        "event_event_external_id",
        "event_external_id",
        "event_id",
        "start_timestamp",
        "updated_event_id",
        "updated_timestamp"
    ),
    "event_event_external_view": (
        "created_event_id",
        "created_timestamp",
        "end_timestamp",
        "event_event_external_id",
        "event_external_id",
        "event_id",
        "start_timestamp",
        "updated_event_id",
        "updated_timestamp"
    ),
    "campaign_smartlink_table": (
        "campaign_id",
        "campaign_smartlink_id",
        "smartlink_id"
    ),
    "campaign_smartlink_view": (
        "campaign_id",
        "campaign_smartlink_id",
        "smartlink_id"
    ),
    "profile_storage_type_table": (
        "file_type_id",
        "name",
        "profile_storage_type_id"
    ),
    "entity_moderation_history_table": (
        "entity_id",
        "entity_moderation_history_id",
        "entity_type",
        "is_approved",
        "is_need_more_info",
        "logger_id",
        "session",
        "update_entity_profile_id"
    ),
    "entity_moderation_table": (
        "create_entity_logger_id",
        "create_entity_profile_id",
        "create_session",
        "entity_id",
        "entity_moderation_id",
        "entity_type",
        "is_approved",
        "is_need_more_info",
        "stars_logger_id"
    ),
    "role_user_view": (
        "role_id",
        "role_user_id",
        "user_id"
    ),
    "role_ml_view": (
        "lang_code",
        "name",
        "role_id",
        "role_ml_id"
    ),
    "role_user_general_view": (
        "lang_code",
        "name",
        "user_id"
    ),
    "role_ml_table": (
        "lang_code",
        "name",
        "role_id",
        "role_ml_id"
    ),
    "group_list_member_view": (
        "group_id",
        "group_list_id",
        "group_list_member_id"
    ),
    "location_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "lang_code",
        "location_type_id",
        "location_type_ml_id",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "user_login_history_view": (
        "login_timestamp",
        "user.main_email_address",
        "user_created_timestamp",
        "user_first_name",
        "user_id",
        "user_last_name"
    ),
    "table_type_table": (
        "description",
        "name",
        "table_type_id"
    ),
    "action_logger_view": (
        "action_id",
        "count(logger.action_id)",
        "destination.system.title",
        "name",
        "source.system.title"
    ),
    "logger_action_view": (
        "action_id",
        "destination system name",
        "name",
        "source system name",
        "timestamp"
    ),
    "programming_language_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "programming_language_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "item_type_view": (
        "entity_type_id",
        "item_type_id"
    ),
    "item_rendered_view": (
        "item_id",
        "item_rendered_id",
        "item_type_id",
        "profile_id",
        "user_id",
        "view_duration",
        "view_timestamp"
    ),
    "entity_type_ml_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_ml_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "item_rendered_general_view": (
        "entity_type_id",
        "entity_type_ml.title",
        "item_id",
        "item_rendered_id",
        "item_type_id",
        "profile.main_email_address",
        "profile_id",
        "title",
        "user.main_email_address",
        "user_id",
        "view_duration",
        "view_timestamp"
    ),
    "message_template_message_template_text_block_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "message_template_id",
        "message_template_message_template_text_block_id",
        "message_template_message_template_text_block_type_id",
        "message_template_text_block_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_template_view": (
        "message_template_id",
        "name"
    ),
    "message_template_text_block_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "criteria_set_id",
        "end_timestamp",
        "is_test_data",
        "message_template_text_block_id",
        "message_template_text_block_type_id",
        "question_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_template_text_block_ml_view": (
        "default_body_template",
        "default_subject_template",
        "email_body_html_template",
        "email_subject_template",
        "lang_code",
        "message_template_text_block_id",
        "message_template_text_block_ml_id",
        "sms_body_template",
        "whatsapp_body_template"
    ),
    "question_ml_view": (
        "lang_code",
        "question_id",
        "question_ml_id",
        "title"
    ),
    "question_view": (
        "answer_expiration_days",
        "default_question_possible_answer_id",
        "default_value",
        "hint",
        "is_required",
        "is_visible",
        "label_question_id",
        "name",
        "question_id",
        "question_type_id",
        "schema_attributes",
        "table_id",
        "uischema_attributes",
        "variable_id",
        "where_statement"
    ),
    "star_transaction_view": (
        "action_id",
        "action_stars",
        "activity_type_id",
        "api_type_id",
        "balance_after_transaction",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "star_transaction_id",
        "subscription_id",
        "user_id"
    ),
    "platform_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "number",
        "platform_id",
        "platform_name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "first_name_alias_table": (
        "first_name_alias_id",
        "first_name_id1",
        "first_name_id2"
    ),
    "smartlink_type_view": (
        "action_id",
        "dialog_workflow_script_id_old",
        "expiration_unit",
        "expiration_value",
        "is_test_data",
        "smartlink.name",
        "smartlink_type_id",
        "url_redirect_template"
    ),
    "event_group_table": (
        "event_group_id",
        "event_id",
        "group_id"
    ),
    "url_type_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_main_title",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id",
        "url_type_id",
        "url_type_ml_id"
    ),
    "brand_url_table": (
        "brand_id",
        "brand_url_id",
        "url",
        "url_type_id"
    ),
    "criteria_group_table": (
        "criteria_group_id",
        "criteria_id",
        "group_id"
    ),
    "sql_structure2code_table": (
        "name",
        "sql_structure2code_id"
    ),
    "message_profile_view": (
        "message_id",
        "message_profile_id",
        "message_status_id",
        "profile_id",
        "reaction_id",
        "user_id"
    ),
    "item_rendered_table": (
        "item_id",
        "item_rendered_id",
        "item_type_id",
        "profile_id",
        "user_id",
        "view_duration",
        "view_timestamp"
    ),
    "item_profile_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "exposure_time_miliseconds",
        "is_test_data",
        "item_id",
        "item_profile_id",
        "message_id",
        "profile_id",
        "seen_body_end_timestamp",
        "seen_body_start_timestamp",
        "seen_summary_end_timestamp",
        "seen_summary_start_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "item_profile_general_view": (
        "item_id",
        "item_type_id",
        "profile_id",
        "recommendation_model_id",
        "source_name"
    ),
    "group_list_member_table": (
        "group_id",
        "group_list_id",
        "group_list_member_id"
    ),
    "database_engine_table": (
        "database_engine_id",
        "engine",
        "is_data_api",
        "is_serverless",
        "product",
        "serverless_version",
        "supplier_id"
    ),
    "environment_repo_table": (
        "environment_id",
        "environment_repo_id",
        "repo_id",
        "status"
    ),
    "device_table": (
        "created_timestamp",
        "created_user_id",
        "device_id",
        "end_timestamp",
        "os_id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "version"
    ),
    "profile_metrics_type_ml_table": (
        "description",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "name",
        "profile_metrics_type_id",
        "profile_metrics_type_ml_id"
    ),
    "event_external_approved_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "event_external_id",
        "external_event_identifier",
        "is_approved",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id",
        "url"
    ),
    "dialog_workflow_script_ml_view": (
        "dialog_workflow_script_id",
        "dialog_workflow_script_ml_id",
        "lang_code",
        "title"
    ),
    "dialog_workflow_script_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id",
        "end_timestamp",
        "is_test_data",
        "max_state_id",
        "name",
        "start_state_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_page_table": (
        "is_approved",
        "parent_profile_id",
        "profile_id",
        "profile_page_id"
    ),
    "profile_page_ml_table": (
        "is_main",
        "lang_code",
        "profile_page_id",
        "profile_page_ml_id",
        "title"
    ),
    "event_external_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "event_external_id",
        "external_event_identifier",
        "is_approved",
        "is_test_data",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id",
        "url"
    ),
    "workflow_action_ml_table": (
        "lang_code",
        "title",
        "workflow_action_id",
        "workflow_action_ml_id"
    ),
    "internet_domain_view": (
        "domain",
        "internet_domain_id",
        "is_commercial_webmail",
        "is_email_assign_profile_organization_to_people",
        "is_url_assign_profile_organization_to_people",
        "organization_id",
        "profile_id",
        "profile_list_id",
        "top_level_domain"
    ),
    "device_user_external_table": (
        "device_id",
        "devicel_user_externa_id",
        "user_external_id"
    ),
    "contact_email_address_view": (
        "contact_email_address_id",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "end_timestamp",
        "is_test_data",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "dialog_workflow_general_view": (
        "description",
        "dialog_workflow_script_id",
        "dialog_workflow_script_ml_id",
        "dialog_workflow_state_id",
        "start_state_id",
        "state_id",
        "title",
        "variable name",
        "variable1_id",
        "workflow action name",
        "workflow_action_id"
    ),
    "node_table": (
        "cloud_region_id",
        "cloud_region_name",
        "description",
        "environment_id",
        "max_bigint_id",
        "min_bigint_id",
        "node_id"
    ),
    "cloud_vendor_table": (
        "cloud_vendor_id",
        "name",
        "seller_id",
        "cloud_vendor_id",
        "seller_id"
    ),
    "cloud_region_table": (
        "cloud_region_id",
        "cloud_vendor_id",
        "name",
        "cloud_region_id",
        "cloud_vendor_id",
        "name"
    ),
    "campaign_message_template_table": (
        "campaign_id",
        "campaign_message_template_id",
        "message_template_id",
        "percent"
    ),
    "relationship_type_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "gender1",
        "gender2",
        "is_direct_connection",
        "is_directional",
        "is_user_level",
        "party1_gender_id",
        "party1_older_max_years",
        "party1_older_min_years",
        "party1_profile_category_id",
        "party1_profile_category_title",
        "party2_gender_id",
        "party2_profile_category_id",
        "party2_profile_category_title",
        "past_present_future",
        "profile1_older_than_profile2",
        "relationship_category_id",
        "relationship_type_classification_id",
        "relationship_type_classification_title",
        "relationship_type_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_type_table": (
        "activity_type_id"
    ),
    "activity_type_ml_table": (
        "activity_type_id",
        "activity_type_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_template_rfp_table": (
        "message_template_id",
        "message_template_rfp_id",
        "rfp_id"
    ),
    "buyer_table": (
        "buyer_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_type_table": (
        "service_type_id"
    ),
    "seller_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "seller_id",
        "seller_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "rfp_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "rfp_id",
        "rfp_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "rfp_requirements_table": (
        "field_id",
        "max_value",
        "min_value",
        "pririty",
        "rfp_id",
        "rfp_requirements_id"
    ),
    "rfp_referral_table": (
        "buyer_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "message_id",
        "number",
        "press_link_in_content_message_timestamp",
        "rfp_id",
        "rfp_referral_id",
        "seller_approach_buyer_using_message_to_email_timestamp",
        "seller_approach_buyer_using_message_to_phone_timestamp",
        "seller_exposed_buyer_address_timestmap",
        "seller_exposed_buyer_email_address_timestmap",
        "seller_exposed_buyer_phone_timestmap",
        "seller_exposed_rfp_details_timestmap",
        "seller_id",
        "seller_interested_in_rfp_timestamp",
        "seller_interested_that_buyer_will_approach_seller_timestamp",
        "seller_interested_to_approach_buyer_timestamp",
        "smartlink_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "repo_dependency_table": (
        "dependent_on_repo_id",
        "repo_dependency_id",
        "repo_id"
    ),
    "source_code_table": (
        "filename",
        "path",
        "programming_language_id",
        "repo_id",
        "source_code_id",
        "tag",
        "version"
    ),
    "message_template_text_block_type_ml_view": (
        "lang_code",
        "message_template_text_block_type_id",
        "message_template_text_block_type_ml_id",
        "title"
    ),
    "operational_hours_table": (
        "created_timestamp",
        "created_user_id",
        "day_of_week",
        "end_timestamp",
        "from_time",
        "is_business_hours",
        "is_test_data",
        "location_id",
        "occurence_id",
        "operational_hours_id",
        "operational_hours_type",
        "profile_id",
        "start_timestamp",
        "until_time",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_reaction_table": (
        "contact_id",
        "contact_reaction_id",
        "created_effective_contact_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "reaction_id",
        "start_timestamp",
        "updated_effective_contact_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_metrics_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_metrics_id",
        "profile_metrics_type",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "person_in_all_table_view": (
        "COUNT(person_id)"
    ),
    "contact_note_text_block_view": (
        "contact_note_id",
        "contact_note_text_block_id",
        "seq",
        "text_block_id"
    ),
    "json_version_source_code_table": (
        "json_version_id",
        "json_version_source_code_id",
        "source_code_id"
    ),
    "json_version_table": (
        "field_name",
        "json_name",
        "json_version_id",
        "table_id",
        "version"
    ),
    "event_news_table": (
        "event_id",
        "event_news_id",
        "news_id"
    ),
    "group_news_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_news_id",
        "hashtag",
        "identifier",
        "is_test_data",
        "label_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_profile_email_table": (
        "channel_id",
        "event_id",
        "event_profile_email_id",
        "is_list_with_full_names",
        "is_number_of_participants",
        "number_of_hours_before_event",
        "profile_id"
    ),
    "brand_location_table": (
        "brand_id",
        "brand_location_id",
        "location_id"
    ),
    "error_table": (
        "description",
        "error_id",
        "identifier",
        "source_code_id"
    ),
    "avatar_table": (
        "avatar_id",
        "name",
        "profile_id",
        "subsystem",
        "system"
    ),
    "contact_group_view": (
        "attribute",
        "consumer_category_id",
        "contact_group_id",
        "contact_group_type_id",
        "contact_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "is_main_in_category",
        "is_recommende_by_ai",
        "is_recommended_by_contact_notes",
        "is_sure",
        "is_test_data",
        "months",
        "participant_category_id",
        "relationship_type_id",
        "seq",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "supplier_category_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_importer_view": (
        "contact_id",
        "contact_importer_id",
        "import_contact_id"
    ),
    "contact_note_view": (
        "block_type_id_old",
        "contact_id",
        "contact_note_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_passed_regex_old",
        "is_test_data",
        "last_parsing_timestmp_old",
        "note",
        "number",
        "seq_old",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "contact_person_view": (
        "contact_id",
        "contact_person_id",
        "person_id"
    ),
    "contact_phone_view": (
        "contact_id",
        "contact_phone_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "phone_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_profile_view": (
        "contact_id",
        "contact_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_potential",
        "is_test_data",
        "number",
        "profile_id",
        "relationship_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "importer_recent_contact_view": (
        "created_timestamp",
        "data_source_type.title",
        "data_source_type_id",
        "entity_id",
        "entity_type_id",
        "entity_type_ml.title",
        "first_name",
        "importer.is_test_data",
        "importer_country_ml.title",
        "importer_id",
        "last_name",
        "user_external_id"
    ),
    "importer_recent_contact_email_address_view": (
        "contact_email_address_id",
        "contact_id",
        "created_timestamp",
        "data_source_title",
        "data_source_type_id",
        "email_address",
        "email_address_approved",
        "email_address_ml.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "importer_country_name",
        "importer_id",
        "last_name"
    ),
    "importer_recent_contact_phone_view": (
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type.title",
        "first_name",
        "importer_country_name",
        "importer_id",
        "label.name",
        "label_phone_ml.title",
        "last_name",
        "local_number_normalized",
        "number_original"
    ),
    "brand_type_table": (
        "brand_type_id",
        "name",
        "other companies in this domain"
    ),
    "profile_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_type_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_type_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_type_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_profile_role_table": (
        "profile_profile_role_id"
    ),
    "profile_profile_role_ml_table": (
        "lang_code",
        "profile_profile_role_id",
        "profile_profile_role_ml_table",
        "title"
    ),
    "profile_history_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "field_id",
        "new_value",
        "old_value",
        "profile_history_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_user_external_view": (
        "contact_id",
        "contact_user_external_id",
        "user_external_id"
    ),
    "permission_repo_table": (
        "permision_action",
        "permission_repo_id",
        "permission_table",
        "repo_id",
        "repo_name_old"
    ),
    "activity_activity_table": (
        "activity_activity_id",
        "activity_id",
        "parent_activity_id"
    ),
    "product_metrics_type_table": (
        "product_metrics_type_id",
        "system_id"
    ),
    "product_metrics_type_ml_table": (
        "description",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "name",
        "product_metrics_type_id",
        "product_metrics_type_ml_id"
    ),
    "product_metrics_table": (
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "is_test_data",
        "product_metrics_id",
        "product_metrics_type",
        "regions_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "sql_report_table": (
        "is_approved",
        "max_time_seconds",
        "min_time_seconds",
        "number_of_records",
        "number_of_records_limit",
        "product_metrics_type_id",
        "sql_report_id",
        "sql_statement",
        "time_limit_seconds"
    ),
    "environnment_sql_report_table": (
        "environnment_id",
        "environnment_sql_report_id",
        "product_metric_id",
        "sql_report_id"
    ),
    "table_list_old": (
        "table_id"
    ),
    "data_source_external_user_ml_table": (
        "created_timestamp",
        "created_user_id",
        "data_source_id",
        "data_source_name",
        "data_source_name_approved",
        "data_source_type_ml_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_milestone_table_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_sure",
        "is_test_data",
        "number",
        "profile_id",
        "profile_milestone_id",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "importer_general_view": (
        "country_ml.title",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "handle",
        "importer.is_test_data",
        "importer_id",
        "is_test_data",
        "location_id",
        "profile_id",
        "start_timestamp",
        "system_ml.title",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "importer_recent_contact_location_view": (
        "contact's_location_id",
        "contact.city_ml.title",
        "contact_address_english",
        "contact_address_local_language",
        "contact_city_id",
        "contact_country_id",
        "contact_country_name",
        "contact_id",
        "contact_location_id",
        "created_timestamp",
        "data_source.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "importer_country_name",
        "importer_id",
        "last_name",
        "person's_location_id",
        "person.city_ml.title",
        "person_address_english",
        "person_address_local_language",
        "person_country_name"
    ),
    "location_person_table": (
        "location_id",
        "location_person_id",
        "person_id"
    ),
    "contact_location_view": (
        "contact_id",
        "contact_location_id",
        "location_id"
    ),
    "location_person_view": (
        "location_id",
        "location_person_id",
        "person_id"
    ),
    "brand_group_view": (
        "brand_group_id",
        "brand_id",
        "group_id"
    ),
    "brand_group_table": (
        "brand_group_id",
        "brand_id",
        "group_id",
        "is_group_part_of_brand"
    ),
    "importer_recent_contact_group_view": (
        "contact.first_name",
        "contact.last_name",
        "contact_group_id",
        "contact_id",
        "created_timestamp",
        "data_source.title",
        "entity_id",
        "entity_type_ml.title",
        "group_id",
        "group_type",
        "importer.is_test_data",
        "importer_country_name",
        "importer_id",
        "title"
    ),
    "importer_recent_contact_text_block_view": (
        "contact_id",
        "contact_note",
        "contact_note_id",
        "contact_note_text_block_id",
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type_ml.title",
        "field_profile_text_block.field_id",
        "first_name",
        "group_title",
        "importer_country_name",
        "importer_id",
        "last_name",
        "seq",
        "text_block.text",
        "text_block_id",
        "text_block_type_id",
        "text_block_type_name"
    ),
    "location_general_view": (
        "address_english",
        "address_local_language",
        "city_id",
        "city_ml.is_title_approved",
        "city_ml.title",
        "coordinate",
        "country_id",
        "country_ml.is_title_approved",
        "country_name",
        "county_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_approved",
        "is_community_active",
        "is_test_data",
        "location.is_title_approved",
        "location_id",
        "location_ml.title",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "region_id",
        "start_timestamp",
        "state_id",
        "state_ml.is_title_approved",
        "state_ml.title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_person_table": (
        "email_address_id",
        "email_address_person_id",
        "person_id"
    ),
    "person_phone_view": (
        "person_id",
        "person_phone_id",
        "phone_id"
    ),
    "message_template_general_view": (
        "action_id",
        "action_name",
        "default_body_template",
        "default_question_possible_answer_id",
        "default_subject_template",
        "email_body_html_template",
        "email_subject_template",
        "field_id",
        "field_name",
        "message_template.name",
        "message_template_id",
        "message_template_text_block.criteria.name",
        "message_template_text_block.criteria_id_old",
        "message_template_text_block.criteria_set_id",
        "message_template_text_block_id",
        "message_template_text_block_is_visible",
        "message_template_text_block_ml_lang_code",
        "message_template_text_block_name",
        "message_template_text_block_seq",
        "message_template_text_block_type.title",
        "possible_answer",
        "question.title",
        "question_hint",
        "question_id",
        "question_is_required",
        "question_ml_lang_code",
        "question_type_id",
        "question_type_name",
        "question_type_schema_attributes",
        "question_type_uischema_attributes",
        "schema_attributes",
        "sms_body_template",
        "uischema_attributes",
        "variable_id",
        "variable_ml_title",
        "variable_name",
        "whatsapp_body_template"
    ),
    "group_group_category_table": (
        "group_category_id",
        "group_group_category_id",
        "group_id"
    ),
    "importer_recent_contact_profile_group_view": (
        "contact_id",
        "contact_profile_id",
        "created_timestamp",
        "data_source.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "group_id",
        "group_profile_id",
        "importer_country_name",
        "importer_id",
        "last_name",
        "profile_id",
        "title"
    ),
    "importer_recent_contact_profile_view": (
        "contact_id",
        "contact_profile_id",
        "created_timestamp",
        "data_source_type_ml.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "gender",
        "importer_country_name",
        "importer_id",
        "last_name",
        "profile_id"
    ),
    "importer_recent_contact_first_name_view": (
        "contact_id",
        "created_timestamp",
        "data_source.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "first_name_ml.first_name",
        "gender_id",
        "importer_country_name",
        "importer_id",
        "last_name"
    ),
    "api_call_general_view": (
        "api_call_id",
        "api_type.name",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "endpoint_url",
        "handle",
        "http_status_code",
        "incoming_message_json",
        "is_network",
        "is_test_data",
        "outgoing_body_json",
        "outgoing_body_significant_fields_hash",
        "outgoing_header_json",
        "response_body_json",
        "session_token",
        "user_external_id",
        "username"
    ),
    "profile_general_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "internal_description",
        "is_approved",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "main_phone_id",
        "number",
        "person.first_name",
        "person.last_name",
        "person.person_id",
        "preferred_lang_code",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_person_id",
        "profile.main_user_id",
        "profile.name",
        "profile_id",
        "profile_ml.title",
        "profile_type_id",
        "profile_type_ml.title",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "message_template_function_table": (
        "description",
        "is_expression",
        "is_function",
        "message_template_function_id",
        "name"
    ),
    "sql2code_table": (
        "refresh_unit",
        "refresh_value",
        "schema_name",
        "sql2code_id",
        "view_table_name"
    ),
    "rfp_type_table": (
        "name",
        "rfp_type_id"
    ),
    "recruitment_rfp_table": (
        "recruitment_rfp_id"
    ),
    "group_general_view": (
        "group.end_timestamp",
        "group.is_test_data",
        "group.name",
        "group_category_title",
        "group_id",
        "group_id1 (parent)",
        "group_id2 (child)",
        "group_ml.end_timestamp",
        "group_ml.is_test_data",
        "group_ml_id",
        "group_type_name",
        "is_contact_classification",
        "is_location",
        "is_main_title",
        "lang_code",
        "new_group_id",
        "title"
    ),
    "group_type_table": (
        "examples_old",
        "group_type_id",
        "is_contact_classification",
        "is_gender",
        "is_job_title",
        "is_location",
        "is_organization",
        "is_organization_classification",
        "name"
    ),
    "importer_recent_contact_user_external_view": (
        "contact_id",
        "contact_user_external_id",
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "importer_country_name",
        "importer_id",
        "last_name"
    ),
    "importer_view": (
        "created_timestamp",
        "created_user_id",
        "data_imported_json",
        "data_source_external_user_id_old",
        "data_source_instance_id",
        "data_source_type_id_old",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "google_calendar_id",
        "google_people_api_resource_name",
        "importer_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external_country_id_old",
        "user_external_id"
    ),
    "planned_database_schema_change_table": (
        "change",
        "database_version",
        "new_field_name",
        "new_object",
        "old_field_name",
        "old_object",
        "planned_database_schema_change_",
        "schema",
        "team_members_involved"
    ),
    "importer_recent_contact_profile_milestone_view": (
        "contact_id",
        "contact_profile_id",
        "created_timestamp",
        "data_source_type_ml.title",
        "entity_id",
        "entity_type_ml.title",
        "first_name",
        "importer_country_name",
        "importer_id",
        "is_sure",
        "last_name",
        "profile_id",
        "profile_milestone_id",
        "start_timestamp",
        "text",
        "text_block_id"
    ),
    "contact_internet_domain_view": (
        "contact_id",
        "contact_internet_domain_id",
        "internet_domain_id"
    ),
    "form_general_view": (
        "default_body_template",
        "default_question_possible_answer_id",
        "default_subject_template",
        "field_name",
        "form_id",
        "form_message_seq_in_page",
        "form_message_template_id",
        "form_name",
        "form_page",
        "max_message_template",
        "message_template.name",
        "message_template_id",
        "message_template_text_block.criteria_set_id",
        "message_template_text_block_id",
        "message_template_text_block_is_visible",
        "message_template_text_block_ml_lang_code",
        "message_template_text_block_name",
        "message_template_text_block_seq",
        "min_message_template",
        "possible_answer",
        "question_id",
        "question_is_required",
        "question_ml_lang_code",
        "question_title",
        "question_type_id",
        "question_type_name",
        "question_type_schema_attributes",
        "question_type_uischema_attributes",
        "schema_attribute",
        "uischema_attribute",
        "variable_id",
        "variable_ml_title",
        "variable_name"
    ),
    "question_general_view": (
        "field_name",
        "label_question_id",
        "question.name",
        "question_default_value",
        "question_id",
        "question_ml_default_value",
        "question_type_id",
        "question_type_name",
        "question_type_schema_attributes",
        "question_typeuischema_attributes",
        "title",
        "value",
        "variable_id",
        "variable_ml_title",
        "variable_name"
    ),
    "action_item_view": (
        "action_id",
        "action_item_id",
        "action_item_type_id",
        "active_main_profile_id",
        "activity_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "other_group_id",
        "other_person_id",
        "other_profile_id",
        "passive_main_contact_id1",
        "passive_main_profile_id",
        "ranking",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "action_item_ml_view": (
        "action_item_id",
        "action_item_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_template_text_block_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "message_template_text_block_type_id",
        "name",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_phone_view": (
        "label_id",
        "label_phone_id",
        "phone_id"
    ),
    "label_profile_view": (
        "label_id",
        "label_profile_id",
        "profile_id"
    ),
    "email_address_label_table": (
        "email_address_id",
        "email_address_label_id",
        "label_id"
    ),
    "email_address_label_view": (
        "email_address_id",
        "email_address_label_id",
        "label_id",
        "email_address_id",
        "email_address_label_id",
        "label_id"
    ),
    "group_category_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_category_id",
        "group_type_id",
        "is_hierarchy",
        "is_location",
        "is_organization",
        "is_prioritze",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "organization_profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "job_title_id",
        "job_title_ml_id",
        "number",
        "organization_id",
        "organization_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "organization_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_approved",
        "is_main",
        "is_test_data",
        "location_id",
        "main_group_id",
        "members_visibility_scope_id",
        "name",
        "Non_members_visibility_profile_id",
        "non_members_visibility_scope_id",
        "number",
        "organization_id",
        "parent_organization_id",
        "point",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "organization_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "organization_id",
        "organization_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_ml_approved_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "form_transaction_table": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "class_parameters_json",
        "compound_form",
        "compound_form_json",
        "compound_form_json_version",
        "created_timestamp",
        "created_user_id",
        "email_form_id",
        "end_timestamp",
        "form_id",
        "form_transaction_id",
        "form_type_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "is_gen_ai",
        "is_moderator_approved",
        "is_require_moderator",
        "is_test_data",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_status_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "number",
        "occurrence_id",
        "pin",
        "previous_form_id",
        "process_id",
        "profile_id",
        "provider_id",
        "push_notification",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_form",
        "returned_value",
        "scheduled_sent_timestamp",
        "sender_email",
        "sender_profile_id",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_profile_id",
        "to_profile_list_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "visibility_id"
    ),
    "dialog_workflow_script_ml_table": (
        "dialog_workflow_script_id",
        "dialog_workflow_script_ml_id",
        "lang_code",
        "title"
    ),
    "dialog_workflow_action_en_general_view": (
        "description",
        "lang_code",
        "name",
        "title",
        "workflow_action_id"
    ),
    "dialog_workflow_state_ml_table": (
        "dialog_workflow_state_id",
        "dialog_workflow_state_ml_id",
        "lang_code",
        "title"
    ),
    "label_phone_table_old": (
        "label_id",
        "label_phone_id",
        "phone_id"
    ),
    "group_ml_also_not_approved_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_ml_not_approved_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_job_title_general_view": (
        "group_id",
        "group_name",
        "job_titile.name",
        "job_title_id",
        "job_title_ml.title"
    ),
    "table_definition_view": (
        "contact_mapping_table_id",
        "entity_type_id1",
        "entity_type_id2",
        "is_created_updated_fields",
        "is_entity",
        "is_graph",
        "is_history",
        "is_logical_delete_fields",
        "is_main_column",
        "is_mapping_table",
        "is_metrics",
        "is_ml_child_table",
        "is_ml_parent_table",
        "is_number_column",
        "is_one_table_per_all_tenants",
        "is_pii",
        "is_sensative",
        "is_single_and_multi_value",
        "is_test_data_column",
        "is_tree",
        "is_visibility",
        "name_of_single_value_column",
        "profile_mapping_table_id",
        "schema",
        "sensative_columns",
        "seq",
        "table_definition_id",
        "table_list_id",
        "table_name",
        "table_type",
        "view_name"
    ),
    "user_to_be_reviewed_view": (
        "active_location_id",
        "birthday_moved_to_person",
        "birthday_original_moved_to_person",
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "current_time_zone",
        "default_time_zone",
        "end_timestamp",
        "first_name",
        "gender_id",
        "identity_id",
        "interested_in_gender_id",
        "is_approved",
        "is_test_data",
        "last_name",
        "name_approved",
        "number",
        "password",
        "preferred_lang_code",
        "presence",
        "region",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user.main_email_address",
        "user_approval_to_get_emails_from_us",
        "user_id",
        "username"
    ),
    "message_channel_table": (
        "chanel_type_id",
        "message_channel_id",
        "name",
        "version"
    ),
    "text_block_type_regex_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "lang_code",
        "regex",
        "start_timestamp",
        "text_block_type_id",
        "text_block_type_regex_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_channel_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "message_channel_id",
        "message_channel_ml_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item_ml_table": (
        "action_item_id",
        "action_item_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_ml_table": (
        "action_item_type_id",
        "action_item_type_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item_profile_view": (
        "action_item_id",
        "action_item_ml_id",
        "action_item_profile_id",
        "active_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "passive_contact_id",
        "passive_profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item_priority_table": (
        "action_item_priority_id",
        "default_action_item_ranking"
    ),
    "calendar_item_ml_table": (
        "calendar_item_id",
        "calendar_item_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_hashag_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "event_hashag_id",
        "event_id",
        "hashtag"
    ),
    "participant_table": (
        "number_of_participants",
        "paied",
        "participant_id",
        "price",
        "profile_id"
    ),
    "calendar_ml_table": (
        "calendar_id",
        "calendar_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "calendar_item_resource_table": (
        "calendar_item_id",
        "calendar_item_resource_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_mandatory",
        "is_optional",
        "is_test_data",
        "number",
        "resource_id",
        "resource_status_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "calendar_profile_table": (
        "attendee_status_id",
        "calendar_item_id",
        "calendar_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_mandatory",
        "is_optional",
        "is_test_data",
        "number",
        "person_id",
        "profile_id",
        "profile_person_status_id",
        "resource_id",
        "resource_status_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "calendar_item_profile_person_table": (
        "attendee_status_id",
        "calendar_item_id",
        "calendar_item_profile_person_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_mandatory",
        "is_optional",
        "is_test_data",
        "number",
        "person_id",
        "profile_id",
        "profile_person_status_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "activity_table_old": (
        "activity_id",
        "activity_type_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "parent_activity_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "calendar_table": (
        "calendar_id",
        "calendar_type",
        "created_effective_profile_id",
        "effective_profile_id",
        "name"
    ),
    "user_not_deleted_view": (
        "active_location_id",
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "current_time_zone",
        "default_time_zone",
        "end_timestamp",
        "first_name",
        "gender_id",
        "identity_id",
        "interested_in_gender_id",
        "is_approved",
        "is_test_data",
        "last_name",
        "name_approved",
        "number",
        "password",
        "password_clear_text",
        "password_encrypted",
        "preferred_lang_code",
        "presence",
        "region",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user.main_email_address",
        "user_approval_to_get_emails_from_us",
        "user_id",
        "username"
    ),
    "gender_list_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "gender_list_id",
        "gender_list_id2",
        "name",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "gender_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "gender_id",
        "is_test_data",
        "name",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "gender_list_member_view": (
        "gender_id",
        "gender_list_id",
        "gender_list_member_id",
        "gender_list_name",
        "gender_name",
        "gender_title"
    ),
    "gender_ml_view": (
        "gender_id",
        "lang_code",
        "title"
    ),
    "first_name_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "first_name_id",
        "gender_id",
        "gender_list_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "first_name_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "first_name_id",
        "first_name_ml_id",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "last_name_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "last_name_id",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "last_name_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "last_name_id",
        "last_name_ml_id",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "organization_general_view": (
        "organization_id",
        "title"
    ),
    "organization_ml_not_deleted_view": (
        "lang_code",
        "name",
        "organization_id",
        "organization_ml_id",
        "title"
    ),
    "calendar_item_type_ml_table": (
        "calendar_item_type_id",
        "calendar_item_type_ml_id",
        "land_code",
        "title"
    ),
    "calendar_item_type_table": (
        "calendar_item_type_id",
        "name"
    ),
    "table_definition_general_view": (
        "schema_main",
        "schema_mapping",
        "table_defintion_id_main",
        "table_defintion_id_mapping",
        "table_name_main",
        "table_name_mapping"
    ),
    "email_address_profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "email_address_profile_id",
        "end_timestamp",
        "is_main",
        "is_sure",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "field_text_block_type_view": (
        "comment",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "field_id",
        "field_text_block_type_id",
        "index_in_regex",
        "is_multiple",
        "lang_code",
        "regex",
        "start_timestamp",
        "text_block_type_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_general_view": (
        "lang_code",
        "name",
        "regex",
        "system_entity_id",
        "system_id",
        "text_block_type_id",
        "text_block_type_name",
        "text_block_type_regex_regex"
    ),
    "person_profile_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "person_profile_id",
        "person_profile_role_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_ml_not_deleted_view": (
        "coordinate",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_profile_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "group_category_id",
        "group_id",
        "hashtag",
        "identifier",
        "is_approved",
        "is_building",
        "is_campaign",
        "is_city",
        "is_continent",
        "is_country",
        "is_county",
        "is_event",
        "is_first_name",
        "is_geo",
        "is_interest",
        "is_job_title",
        "is_last_name",
        "is_neighbourhood",
        "is_official",
        "is_organization",
        "is_region",
        "is_state",
        "is_street",
        "is_test_data",
        "is_zip_code",
        "lang_code",
        "location_id",
        "location_list_id",
        "main_group_type_id",
        "members_visibility_id",
        "name",
        "non_members_visibility_id",
        "number",
        "parent_group_id",
        "parent_group_list_id",
        "profile_id",
        "sequence",
        "start_timestamp",
        "system_group_name",
        "system_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "user_external_general_view": (
        "access_token",
        "access_token_old",
        "cloud_project",
        "created_timestamp",
        "created_user_id",
        "email_account_id",
        "email_account_name",
        "email_address_id",
        "end_timestamp",
        "environment_id_old",
        "expiry_old",
        "handle",
        "is_test_data",
        "number",
        "oauth_callback_confirmed",
        "oauth_token",
        "oauth_token_secret",
        "person_id",
        "phone.full_number_normalized",
        "profile_id",
        "profile_user_external_id",
        "refresh_token",
        "refresh_token_old",
        "scope_json",
        "start_timestamp",
        "subsystem_id",
        "system.name",
        "system_id",
        "token_2delete",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "user_id",
        "username"
    ),
    "mailbox_profile_table": (
        "mailbox_id",
        "mailbox_profile_id",
        "name",
        "profile_id"
    ),
    "email_account_table": (
        "domain",
        "email_account_id",
        "imap_host",
        "imap_port",
        "name",
        "password",
        "smtp_host",
        "smtp_port",
        "username"
    ),
    "email_account_mailbox_table": (
        "email_account_id",
        "email_account_mailbox_id",
        "mailbox_id"
    ),
    "mailbox_table": (
        "mailbox_id",
        "name"
    ),
    "profile_mailbox_general_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "gender_id",
        "internal_description",
        "is_approved",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id_old",
        "mailbox_profile_id",
        "mailbox_profile_name",
        "main_person_id",
        "main_person_id_in_profile_table",
        "main_phone_id",
        "main_user_id",
        "number",
        "person_first_name",
        "person_id_in_person_table",
        "person_last_name",
        "preferred_lang_code",
        "profile_id",
        "profile_main_email",
        "profile_ml.title",
        "profile_name",
        "profile_type_id",
        "stars",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "email_account_user_external_table": (
        "default_group_id",
        "email_account_id",
        "email_account_user_external_id",
        "email_address_id",
        "name",
        "user_external_id"
    ),
    "user_external_email_account_general_view": (
        "access_token_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_account.name",
        "email_account_id",
        "email_address_id",
        "end_timestamp",
        "environment_id_old",
        "expiry_old",
        "handle",
        "is_test_data",
        "number",
        "oauth_callback_confirmed",
        "oauth_token",
        "oauth_token_secret",
        "refresh_token_old",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "token_2delete",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "username"
    ),
    "url_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "url_id",
        "url_type"
    ),
    "profile_url_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_url_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_id",
        "url_old",
        "url_type_id_old"
    ),
    "person_contact_general_view": (
        "contact_created_timestamp",
        "contact_email_address1",
        "contact_first_name",
        "contact_id",
        "contact_person_type_id",
        "contact_person_type_name",
        "first_name",
        "last_name",
        "person.main_email_address",
        "person_created_timestamp",
        "person_id"
    ),
    "user_external_test_data_view": (
        "access_token_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_account.name",
        "email_account_id",
        "email_address_id",
        "end_timestamp",
        "environment_id_old",
        "expiry_old",
        "handle",
        "is_test_data",
        "number",
        "oauth_callback_confirmed",
        "oauth_token",
        "oauth_token_secret",
        "refresh_token_old",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "token_2delete",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "username"
    ),
    "person_general_view": (
        "number_of_contacts",
        "number_of_phones",
        "number_of_profiles",
        "person.birthday_date",
        "person.created_timestamp",
        "person.end_timestamp",
        "person.first_name",
        "person.is_test_data",
        "person.last_name",
        "person.main_email_address",
        "person.person_id",
        "person_profile.end_timestamp",
        "profile.end_timestamp",
        "profile.is_test_data",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile.profile_id",
        "user.user_id"
    ),
    "person_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "person_profile_id",
        "person_profile_role_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "person_with_missing_data_view": (
        "person.end_timestamp",
        "person.first_name",
        "person.last_name",
        "person.main_email_address",
        "person_id",
        "person_profile.end_timestamp",
        "profile.end_timestamp",
        "profile.is_test_data",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile_id"
    ),
    "phone_profile_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "full_number_normalized",
        "is_main",
        "label_phone_id",
        "person.first_name",
        "phone_id",
        "phone_is_test_data",
        "phone_profile_id",
        "phone_profile_is_test_data",
        "profile.main_email_address",
        "profile.name",
        "profile_id",
        "profile_is_test_data",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_with_empty_full_number_normalized_view": (
        "area_code",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "extension",
        "full_number_normalized",
        "international_code",
        "is_test_data",
        "local_number_normalized",
        "number",
        "number_original",
        "phone_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_view": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_criteria_set_id",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "channel_id",
        "class_parameters_json",
        "component_id",
        "compound_message",
        "compound_message_json",
        "compound_message_json_version",
        "conversation_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_message_id",
        "end_timestamp",
        "from_api_type_id",
        "from_email",
        "from_phone_number",
        "from_profile_id",
        "from_provider_id",
        "from_subsystem_id",
        "from_system_id",
        "from_user_external_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "identifier",
        "inforu_application_id_old",
        "inforu_channel_old",
        "inforu_customer_id_old",
        "inforu_customer_param_old",
        "inforu_incoming_json_old",
        "inforu_keyword_old",
        "inforu_mo_session_id_old",
        "inforu_network_old",
        "inforu_project_id_old",
        "inforu_short_code_old",
        "inforu_type_old",
        "inforu_value_old",
        "is_gen_ai",
        "is_moderator_approved",
        "is_pin",
        "is_push_notification",
        "is_require_moderator",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_id",
        "message_message_relationship_id",
        "message_status_id",
        "message_template_id",
        "message_type_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "next_dialog_workflow_state_id",
        "number",
        "occurrence_id",
        "original_message_id",
        "parent_message_id",
        "process_id",
        "profile_id",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_message",
        "returned_value",
        "scheduled_sent_timestamp",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_phone_number",
        "to_profile_id",
        "to_profile_list_id",
        "to_provider_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "visibility_id"
    ),
    "user_pii_general_view": (
        "first_name",
        "last_name",
        "main_profile_id",
        "password",
        "password_clear_text",
        "password_encrypted",
        "user.is_test_data",
        "user.main_email_address",
        "user_id",
        "username"
    ),
    "user_pii_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "password_clear_text",
        "password_encrypted",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_pii_id"
    ),
    "action_item_type_ml_table": (
        "action_item_type_id",
        "action_item_type_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item_type_table": (
        "action_item_type_id",
        "channel_type_id",
        "version"
    ),
    "action_item_category_table": (
        "action_item_category_table_id",
        "channel_category_id",
        "version"
    ),
    "action_item_category_ml_table": (
        "action_item_category_id",
        "action_item_category_ml_id",
        "body",
        "category_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "subeject",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_text_block_view": (
        "contact_id",
        "contact_text_block_id",
        "is_test_data",
        "text_block_id"
    ),
    "profile_level_table": (
        "profile_id"
    ),
    "profile_user_table_view": (
        "profile_id",
        "profile_user_id",
        "user_id"
    ),
    "user_to_delete_view": (
        "is_system",
        "is_test_data",
        "user_id"
    ),
    "profile_to_delete_view": (
        "profile_id",
        "profile_is_system",
        "profile_is_test_data",
        "user_id",
        "user_is_system",
        "user_is_test_data"
    ),
    "table_size_history_table": (
        "brand_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "environment_name",
        "identifier",
        "is_test_data",
        "number",
        "number_of_records",
        "start_timestamp",
        "table_name",
        "table_schema",
        "table_size_history_id",
        "timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "login_history_table": (
        "browser",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "effective_profile_id",
        "end_timestamp",
        "identifier",
        "ipv4",
        "ipv6",
        "is_test_data",
        "login_history_id",
        "number",
        "operating_system",
        "real_profile_id",
        "session",
        "start_timestamp",
        "timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id"
    ),
    "contact_person_table": (
        "contact_id",
        "contact_person_id",
        "contact_person_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "field_text_block_type_table": (
        "comment",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "field_id",
        "field_text_block_type_id",
        "index_in_regex",
        "is_multiple",
        "lang_code",
        "regex",
        "start_timestamp",
        "text_block_type_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_block_type_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "regex",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text_block_type_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_importer_table": (
        "contact_id",
        "contact_importer_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "import_contact_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "import_contact_table_old": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "facebook_profile_name",
        "facebook_profile_number",
        "google_account",
        "identifier",
        "import_contact_id",
        "importer_table_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_internet_domain_table": (
        "contact_id",
        "contact_internet_domain_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "internet_domain_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_note_text_block_table": (
        "contact_note_id",
        "contact_note_text_block_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "seq",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_note_group_table": (
        "contact_note_group_id",
        "contact_note_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_user_external_table": (
        "contact_id",
        "contact_user_external_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "contact_profile_table": (
        "contact_id",
        "contact_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_potential",
        "is_test_data",
        "number",
        "profile_id",
        "relationship_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_text_block_table": (
        "contact_id",
        "contact_text_block_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_location_table": (
        "contact_id",
        "contact_location_id",
        "contact_location_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "label_location_id",
        "location_id",
        "location_type_str",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_location_type_table": (
        "contact_location_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_location_type_ml_table": (
        "contact_location_type_id",
        "contact_location_type_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "land_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_person_type_table": (
        "contact_person_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_group_type_table": (
        "contact_group_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_group_type_ml_table": (
        "contact_group_type_id",
        "contact_group_type_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "message_id",
        "message_profile_id",
        "message_status_id",
        "number",
        "profile_id",
        "reaction_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id"
    ),
    "item_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "exposure_time_miliseconds",
        "identifier",
        "is_test_data",
        "item_id",
        "item_profile_id",
        "item_type_id",
        "message_id",
        "number",
        "profile_id",
        "recommendation_model_id",
        "seen_body_end_timestamp",
        "seen_body_start_timestamp",
        "seen_summary_end_timestamp",
        "seen_summary_start_timestamp",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "event_profile_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "event_profile_type_id",
        "event_profile_type_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "profile_id",
        "profile_text_block_id",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_profile_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_profile_type_id",
        "group_profile_type_ml_id",
        "identifier",
        "is_test_data",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "employer_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "employer_profile_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "employer_profile_vacancy_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "employer_profile_id",
        "employer_profile_vacancy_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "email_address_profile_verification_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_profile_id",
        "email_address_profile_verification_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_text_block_id",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "person_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "person_text_block_id",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "field_profile_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "field_id",
        "field_profile_text_block_id",
        "identifier",
        "is_test_data",
        "number",
        "profile_id",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "email_address_text_block_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "text_table": (
        "component_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "start_timestamp",
        "text_category_id",
        "text_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_system",
        "is_test_data",
        "label_id",
        "label_profile_id",
        "number",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item_priority_view": (
        "action_item_priority_id",
        "default_action_item_ranking"
    ),
    "location_profile_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "lang_code",
        "location_profile_id",
        "location_profile_ml_id",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_mysql_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_mysql_id",
        "test_mysql_ml_id",
        "title",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "contact_note_table": (
        "block_type_id_old",
        "contact_id",
        "contact_note_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_passed_regex_old",
        "is_test_data",
        "last_parsing_timestmp_old",
        "note",
        "number",
        "seq_old",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "first_name_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "first_name_id",
        "gender_id",
        "gender_list_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "first_name_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "first_name_id",
        "first_name_ml_id",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "last_name_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_name_approved",
        "is_test_data",
        "lang_code",
        "last_name_id",
        "last_name_ml_id",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "profile_milestone_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_sure",
        "is_test_data",
        "number",
        "profile_id",
        "profile_milestone_id",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "variable_ml_table": (
        "lang_code",
        "title",
        "variable_id",
        "variable_ml_id",
        "visibility_id"
    ),
    "gender_list_member_table": (
        "gender_id",
        "gender_list_id",
        "gender_list_member_id"
    ),
    "identifier_view": (
        "entity_type_id",
        "identifier",
        "identifier_id",
        "is_test_data",
        "number"
    ),
    "importer_entity_types_imported_group_by_source_view": (
        "Data Source Title",
        "Entity Type Title",
        "Number of Entries"
    ),
    "identifier_group_by_entity_type_view": (
        "entity_type_ml.title",
        "Number of Entries"
    ),
    "identifier_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "entity_type_id",
        "identifier",
        "identifier_id",
        "is_test_data",
        "number",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp"
    ),
    "item_type_table": (
        "entity_type_id",
        "is_personal",
        "item_type_id"
    ),
    "question_type_table": (
        "component_name",
        "name",
        "parent_question_type_id",
        "question_type_id",
        "schema_attributes",
        "uischema_attributes"
    ),
    "question_type_view": (
        "component_name",
        "name",
        "parent_question_type_id",
        "question_type_id",
        "schema_attributes",
        "uischema_attributes"
    ),
    "question_possible_answer_table": (
        "question_id",
        "question_possible_answer_id"
    ),
    "question_possible_answer_ml_table": (
        "action_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_approved",
        "is_main",
        "is_test_data",
        "lang_code",
        "number",
        "question_id_old",
        "question_possible_answer_id",
        "question_possible_answer_ml_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "question_possible_answer_ml_view": (
        "action_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_approved",
        "is_main",
        "is_test_data",
        "lang_code",
        "number",
        "question_id_old",
        "question_possible_answer_id",
        "question_possible_answer_ml_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "question_possible_answer_view": (
        "question_id",
        "question_possible_answer_id"
    ),
    "question_table": (
        "answer_expiration_days",
        "default_question_possible_answer_id",
        "default_value",
        "hint",
        "is_required",
        "is_visible",
        "label_question_id",
        "name",
        "question_id",
        "question_type_id",
        "schema_attributes",
        "table_id",
        "uischema_attributes",
        "variable_id",
        "where_statement"
    ),
    "logger_logger_general_view": (
        "activity",
        "activity_id",
        "client_ip_v4",
        "client_ip_v6",
        "component_id",
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "error_stack",
        "group_id",
        "location_id",
        "logger_id",
        "message",
        "name",
        "payload",
        "profile_id",
        "relationship_type_id",
        "server_ip_v4",
        "server_ip_v6",
        "severity_id",
        "start_datetime",
        "state_id",
        "status_id",
        "timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "logger_variable_general_view": (
        "profile.main_email_address",
        "profile_id",
        "timestamp",
        "variable name",
        "variable_id",
        "variable_value_new",
        "variable_value_old",
        "workflow_state_id"
    ),
    "importer_contact_google_people_api_sync_view": (
        "contact_id",
        "last_sync_time"
    ),
    "contact_email_address_table": (
        "contact_email_address_id",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "end_timestamp",
        "is_test_data",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_review_view": (
        "email_address",
        "email_address_approved",
        "email_address_id",
        "is_test_data"
    ),
    "phone_review_view": (
        "created_timestamp",
        "full_number_normalized",
        "is_test_data",
        "local_number_normalized",
        "number_original",
        "phone_id"
    ),
    "contact_phone_table": (
        "contact_id",
        "contact_phone_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "phone_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_phone_general_view": (
        "contact.is_test_data",
        "contact_id",
        "contact_phone_id",
        "email1",
        "first_name",
        "full_number_normalized",
        "last_name",
        "person_id",
        "phone.is_test_data",
        "phone_id"
    ),
    "reaction_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "image",
        "is_test_data",
        "name",
        "number",
        "reaction_id",
        "reaction_type_id",
        "start_timestamp",
        "text",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "reaction_category_table": (
        "reaction_category_id"
    ),
    "reaction_reaction_category_table": (
        "reaction_category_id",
        "reaction_id",
        "reaction_reaction_category_id"
    ),
    "reaction_review_view": (
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "reaction_id",
        "reaction_ml_id",
        "title"
    ),
    "label_message_table": (
        "label_id",
        "label_message_id",
        "message_id"
    ),
    "importer_entity_types_imported_group_by_date_source_view": (
        "created_user_id",
        "Data Source Title",
        "data_source_type_id",
        "date",
        "entity_type_id",
        "entity_type_title",
        "first_name",
        "last_name",
        "Number of Entries",
        "user.main_email_address",
        "username"
    ),
    "api_call_by_date_api_type_name_view": (
        "api_type.name",
        "api_type_id",
        "count(*)",
        "DATE(created_timestamp)",
        "username"
    ),
    "person_ml_view": (
        "first_name",
        "is_first_name_approved",
        "is_last_name_approved",
        "is_test_data",
        "lang_code",
        "last_name",
        "person_id",
        "person_ml_id"
    ),
    "environment_ml_table": (
        "description",
        "description_approved",
        "environment_id",
        "environment_ml_id",
        "lang_code",
        "name",
        "name_approved"
    ),
    "api_type_user_external_view": (
        "api_type_id",
        "api_type_user_external_id",
        "environment_id",
        "is_test_data",
        "user_external_id"
    ),
    "event_group_view": (
        "event_group_id",
        "event_id",
        "group_id"
    ),
    "event_location_view": (
        "event_id",
        "event_location_id",
        "location_id"
    ),
    "logger_function_start_end_group_by_component_view": (
        "component_id",
        "COUNT(*)",
        "function_name",
        "logger.component_name",
        "severity_description"
    ),
    "api_call_group_by_date_view": (
        "COUNT(*)",
        "DATE(created_timestamp)"
    ),
    "api_call_group_without_user_extenal_id_by_date_view": (
        "COUNT(*)",
        "DATE(created_timestamp)"
    ),
    "api_call_api_type_user_external_count_view": (
        "api_type.name",
        "api_type_id",
        "COUNT(*)",
        "created_user_id",
        "DATE(created_timestamp)",
        "user_external_id",
        "username"
    ),
    "group_deleted_view": (
        "group_category_title",
        "group_id",
        "group_id2 (child)",
        "group_id2 (parent)",
        "group_is_test_data",
        "group_type_name",
        "is_contact_classification",
        "is_location",
        "title"
    ),
    "message_template_message_template_text_block_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "default_body_template",
        "default_subject_template",
        "end_timestamp",
        "is_test_data",
        "message_template_id",
        "message_template_message_template_text_block_id",
        "message_template_message_template_text_block_type_id",
        "message_template_text_block_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "entity_type_general_view": (
        "contact_mapping_table_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_id1",
        "entity_type_id2",
        "is_created_updated_fields",
        "is_entity",
        "is_graph",
        "is_history",
        "is_logical_delete_fields",
        "is_main_column",
        "is_mapping_table",
        "is_metrics",
        "is_ml_child_table",
        "is_ml_parent_table",
        "is_number_column",
        "is_one_table_per_all_tenants",
        "is_pii",
        "is_sensative",
        "is_single_and_multi_value",
        "is_test_data_column",
        "is_title_approved",
        "is_tree",
        "is_visibility",
        "lang_code",
        "name_of_single_value_column",
        "profile_mapping_table_id",
        "schema",
        "sensative_columns",
        "seq",
        "start_timestamp",
        "table_definition_id",
        "table_list_id",
        "table_name",
        "TABLE_ROWS",
        "table_type",
        "title",
        "updated_timestamp",
        "updated_user_id",
        "view_name"
    ),
    "campaign_message_template_view": (
        "campaign_id",
        "campaign_message_template_id",
        "message_template_id",
        "percent"
    ),
    "criteria_general_view": (
        "city_list_id",
        "country_id",
        "criteria.location_id",
        "criteria_id",
        "criteria_name",
        "entity_type.name",
        "entity_type.title",
        "entity_type_id",
        "gender_list.name",
        "gender_list_id",
        "group.location.country.name",
        "group.location.id",
        "group.location.name",
        "group_id",
        "group_list.name",
        "group_list_id",
        "location_type_id",
        "max_age",
        "min_age",
        "title"
    ),
    "group_list_view": (
        "created_timestamp",
        "created_user_id",
        "group_list_id",
        "is_and",
        "name",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_list_table": (
        "created_timestamp",
        "created_user_id",
        "group_list_id",
        "is_and",
        "name",
        "updated_timestamp",
        "updated_user_id"
    ),
    "phone_general_view": (
        "area_code",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "extension",
        "full_number_normalized",
        "international_code",
        "is_test_data",
        "is_verified",
        "local_number_normalized",
        "number",
        "number_original",
        "phone_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_phone_general_view": (
        "area_code",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "extension",
        "full_number_normalized",
        "international_code",
        "is_test_data",
        "is_verified",
        "label.name",
        "local_number_normalized",
        "number",
        "number_original",
        "phone_id",
        "phone_profile_id",
        "profile_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_phone_table": (
        "created_timestamp",
        "created_user_id",
        "label_id",
        "label_phone_id",
        "phone_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_job_title_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_job_title_id",
        "group_ml_id",
        "job_title_id",
        "job_title_ml_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "relationship_type_ml_title2_view": (
        "coordinate",
        "entity_id",
        "title2"
    ),
    "relationship_type_ml_title1_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "goods_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "goods_id",
        "goods_ml_id",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "timeline_potential_items_view": (
        "coordinate",
        "created_user_id",
        "entity_id",
        "entity_type_id",
        "lang_code",
        "location_id",
        "title",
        "updated_timestamp",
        "visibility_id"
    ),
    "goods_location_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "goods_id",
        "goods_location_id",
        "identifier",
        "is_test_data",
        "location_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "goods_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "goods_id",
        "group_list_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "group_location_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_location_type_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_location_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_location_id",
        "group_location_type_id",
        "identifier",
        "is_test_data",
        "label_location_id",
        "location_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_potential_view": (
        "contact_location_id",
        "contact_ml.titile",
        "contact_ml_id",
        "coordinate",
        "created_user_id",
        "entity_id",
        "entity_type_id",
        "lang_code",
        "location_id",
        "title",
        "updated_timestamp",
        "visibility_id"
    ),
    "event_location_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "event_location_id",
        "identifier",
        "is_test_data",
        "location_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "event_potential_view": (
        "coordinate",
        "created_user_id",
        "entity_id",
        "entity_type_id",
        "lang_code",
        "location_id",
        "title",
        "updated_timestamp",
        "visibility_id"
    ),
    "group_potential_view": (
        "coordinate",
        "created_user_id",
        "entity_id",
        "entity_type_id",
        "lang_code",
        "location_id",
        "title",
        "updated_timestamp",
        "visibility_id"
    ),
    "goods_potential_view": (
        "coordinate",
        "created_user_id",
        "entity_id",
        "entity_type_id",
        "lang_code",
        "location_id",
        "title",
        "updated_timestamp",
        "visibility_id"
    ),
    "contact_ml_table": (
        "contact_id",
        "contact_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "data_source_external_user_table": (
        "created_timestamp",
        "created_user_id",
        "data_source_external_user_id",
        "data_source_id",
        "end_timestamp",
        "is_test_data",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "device_profile_table": (
        "created_timestamp",
        "created_user_id",
        "device_id",
        "device_profile_id",
        "end_timestamp",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "data_source_service_table_old": (
        "data_source_service_id",
        "name"
    ),
    "criteria_set_general_view": (
        "child_criteria_set_id",
        "child_name",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "criteria.is_test_data",
        "criteria_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_name",
        "group_id",
        "group_list_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "location_id",
        "location_list_id",
        "name",
        "organization_name",
        "parent_criteria_set_id",
        "parent_name",
        "profile_list_id",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_timestamp",
        "template_condition",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id"
    ),
    "person_phone_general_view": (
        "area_code",
        "full_number_normalized",
        "internationa_code",
        "person.is_test_data",
        "person_id",
        "person_phone_id",
        "phone.is_test_data",
        "phone_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "full_number_normalized",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "person_phone_id",
        "phone_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "real_estate_criteria_table": (
        "agent_profile_id",
        "budget_local_currency",
        "budget_usd",
        "buyer_profile_id",
        "cash_guarentee_local_currency",
        "criteria_id",
        "currency",
        "has_cheq",
        "name",
        "requested_entrance_date"
    ),
    "email_address_person_view": (
        "email_address_id",
        "email_address_person_id",
        "person_id"
    ),
    "email_address_person_general_view": (
        "contact_email_address_id",
        "contact_id",
        "email_address",
        "email_address_approved",
        "email_address_id",
        "email_address_person_id",
        "first_name",
        "last_name",
        "person_id"
    ),
    "contact_note_contact_table_old": (
        "contact_id",
        "contact_note_contact_id",
        "contact_note_id"
    ),
    "data_source_category_table": (
        "data_source_category_id",
        "name"
    ),
    "data_source_category_ml_table": (
        "created_timestamp",
        "created_user_id",
        "data_source_category_id",
        "data_source_category_ml_id",
        "data_source_name",
        "data_source_name_approved",
        "end_timestamp",
        "lang_code",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "data_source_type_ml_en_view": (
        "data_source_type_id",
        "name",
        "subsystem_id",
        "subsystem_ml.title",
        "system_id",
        "system_ml.tite",
        "title"
    ),
    "system_ml_table": (
        "lang_code",
        "system_id",
        "system_ml_id",
        "title"
    ),
    "subsystem_ml_table": (
        "lang_code",
        "subsystem_id",
        "subsystem_ml_id",
        "title"
    ),
    "data_source_type_ml_table": (
        "created_timestamp",
        "created_user_id",
        "data_source_type_id",
        "data_source_type_ml_id",
        "end_timestamp",
        "is_test_data",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subsystem_table": (
        "name",
        "subsystem_id",
        "system_id"
    ),
    "group_profile_general_view": (
        "consumer_category_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "group.name",
        "group_id",
        "group_profile_id",
        "group_profile_type_id",
        "identifier",
        "is_approved_by_group_admin",
        "is_recommended",
        "is_request_by_the_user",
        "is_sure",
        "is_test_data",
        "months",
        "number",
        "participant_category_id",
        "profile.name",
        "profile_id",
        "rank",
        "relationship_type_id",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "supplier_category_id",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "data_source_type__field_table": (
        "data_source_field_id",
        "data_source_type_id",
        "external_field_name",
        "field_id",
        "position"
    ),
    "entity_type_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_ml_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "api_call_group_by_api_type_date_view": (
        "api_type_id",
        "api_type_name",
        "COUNT(*)",
        "DATE(created_timestamp)"
    ),
    "logger_start_end_group_by_component_date_view": (
        "component_id",
        "component_name",
        "COUNT(*)",
        "DATE(created_timestamp)"
    ),
    "system_table": (
        "name",
        "parent_system_id",
        "system_id"
    ),
    "logger_sessions_group_by_day_view": (
        "COUNT(*)",
        "date"
    ),
    "logger_records_of_last_session_view": (
        "class_name",
        "component_id",
        "component_name",
        "created_timestamp",
        "error_stack",
        "function_name",
        "line_number",
        "logger_id",
        "message",
        "payload",
        "profile_id",
        "session",
        "severity_id",
        "status_id",
        "user_id"
    ),
    "logger_last_session_view": (
        "created_timestamp",
        "created_user_id",
        "session"
    ),
    "logger_sessions_order_by_timestamp_view": (
        "COUNT(*)",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "profile_id",
        "session",
        "user.main_email_address",
        "user_id"
    ),
    "storage_recent_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "ext",
        "file_extension.name",
        "file_extension_id",
        "file_type",
        "file_type_id",
        "file_type_ml.title",
        "filename",
        "height",
        "path",
        "ratio_x",
        "ration_y",
        "region",
        "start_timestamp",
        "stoage.is_test_data",
        "storage_id",
        "storage_type_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "width"
    ),
    "file_type_ml_table": (
        "description",
        "file_type_id",
        "file_type_ml_id",
        "is_description_approved",
        "is_main",
        "is_title_approved",
        "lang_code",
        "title"
    ),
    "file_extension_table": (
        "ext",
        "file_extension_id",
        "name"
    ),
    "file_type_ml_view": (
        "description",
        "file_type_id",
        "file_type_ml_id",
        "is_description_approved",
        "is_main",
        "is_title_approved",
        "lang_code",
        "title"
    ),
    "severity_table": (
        "created_timestamp",
        "created_user_id",
        "end_datetime",
        "level",
        "name",
        "severity_id",
        "start_datetime",
        "updated_timestamp",
        "updated_user_id"
    ),
    "severity_ml_table": (
        "description",
        "lang_code",
        "severity_id",
        "severity_ml_id",
        "title"
    ),
    "api_type_table": (
        "action_id",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "endpoint_url",
        "expiration_unit",
        "expiration_value",
        "from_subsystem_id",
        "from_system_id",
        "github_repo",
        "hander_function_name",
        "http_method",
        "interval_max_seconds",
        "interval_min_seconds",
        "is_enabled",
        "is_test_data",
        "message_channel_provider_id",
        "name",
        "start_timestamp",
        "status_code_when_not_found",
        "to_subsystem_id",
        "to_system_id",
        "type",
        "updated_timestamp",
        "updated_user_id",
        "usage_and_billing_url"
    ),
    "profile_ml_table": (
        "about",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "lang_code",
        "profile_id",
        "profile_ml_id",
        "start_timestamp",
        "title",
        "title_approved",
        "updated_timestamp",
        "updated_user_id"
    ),
    "api_call_table": (
        "api_call_id",
        "api_type_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "endpoint_url",
        "http_status_code",
        "incoming_message_json",
        "is_network",
        "is_test_data",
        "number",
        "outgoing_body_json",
        "outgoing_body_significant_fields_hash",
        "outgoing_header_json",
        "response_body_json",
        "session_token",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "neighborhood_view": (
        "city_id",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "group_id",
        "is_test_data",
        "neighborhood_id",
        "number",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_ml_table": (
        "email_address_id",
        "email_address_ml_id",
        "is_test_data",
        "lang_code",
        "title",
        "title_approved"
    ),
    "profile_storage_type_ml_table": (
        "lang_code",
        "profile_storage_type_id",
        "profile_storage_type_ml_id",
        "title"
    ),
    "criteria_set_table": (
        "condition_type_id",
        "criteria_id",
        "criteria_set_id",
        "include_exclude_type_id",
        "name"
    ),
    "importer_recent_general_view": (
        "country_ml.title",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "handle",
        "importer.is_test_data",
        "importer_id",
        "is_test_data",
        "location_id",
        "profile_id",
        "start_timestamp",
        "system_ml.title",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "api_call_recent_general_view": (
        "api_call_id",
        "api_type_id",
        "created_timestamp",
        "endpoint_url",
        "http_status_code",
        "incoming_message_json",
        "is_network",
        "is_test_data",
        "outgoing_body_json",
        "outgoing_body_significant_fields_hash",
        "outgoing_header_json",
        "response_body_json",
        "session_token",
        "user_external_id"
    ),
    "data_source_instance_recent_general_view": (
        "computer_name",
        "created_timestamp",
        "data_source_instance_id",
        "data_source_type_id",
        "end_index",
        "end_timestamp",
        "file_or_api",
        "filename",
        "import_or_last_sync_timestamp",
        "imported_file_created_timestamp",
        "is_test_data",
        "name",
        "path",
        "start_index",
        "start_timestamp",
        "user_external_id"
    ),
    "url_type_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "file_type_id",
        "name",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id",
        "url_prefix_old",
        "url_type_id"
    ),
    "data_source_instance_general_view": (
        "computer_name",
        "created_timestamp",
        "data_source_instance_id",
        "data_source_type_id",
        "end_index",
        "end_timestamp",
        "file_or_api",
        "filename",
        "import_or_last_sync_timestamp",
        "imported_file_created_timestamp",
        "is_test_data",
        "name",
        "path",
        "start_index",
        "start_timestamp",
        "user_external_id"
    ),
    "importer_recent_contact_url_view": (
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type_id",
        "entity_type_ml.title",
        "importer_country_name",
        "importer_id",
        "normalized_url",
        "url",
        "url_type",
        "user_external_id",
        "username"
    ),
    "importer_recent_contact_storage_view": (
        "created_timestamp",
        "data_source_type.title",
        "entity_id",
        "entity_type_id",
        "entity_type_ml.title",
        "filename",
        "importer_country_name",
        "importer_id",
        "storage.url",
        "storage_id"
    ),
    "campaign_criteria_set_view": (
        "campaign_criteria_set_id",
        "campaign_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "criteria_set_id",
        "end_timestamp",
        "is_test_data",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_ml_view": (
        "email_address_id",
        "email_address_ml_id",
        "is_test_data",
        "lang_code",
        "title",
        "title_approved"
    ),
    "campaign_criteria_set_general_view": (
        "campaign.name",
        "campaign_criteria_set_id",
        "campaign_id",
        "criteria_id_old",
        "criteria_set.name",
        "criteria_set_id",
        "is_test_data"
    ),
    "test_mysql_without_is_main_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_mysql_id",
        "test_mysql_ml_id",
        "title",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_mysql_without_is_main_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_mysql_id",
        "test_mysql_ml_id",
        "title",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "location_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "location_id",
        "location_profile_id",
        "location_type_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "county_ml_general_view": (
        "county_id",
        "county_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "criteria_set_view": (
        "condition_type_id",
        "criteria_id",
        "criteria_set_id",
        "include_exclude_type_id",
        "name"
    ),
    "profile_storage_view": (
        "profile_id",
        "profile_storage_id",
        "storage_id"
    ),
    "location_list_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "location_list_id",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_profile_general_view": (
        "address_english",
        "address_local_language",
        "city_id",
        "coordinate",
        "country_id",
        "county_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "identifier",
        "is_approved",
        "is_community_active",
        "is_test_data",
        "location_id",
        "location_profile_id",
        "name",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "profile_id",
        "region_id",
        "start_timestamp",
        "state_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "label_location_view": (
        "is_approved",
        "label_id",
        "label_location_id",
        "location_id"
    ),
    "location_profile_label_general_view": (
        "address_english",
        "address_local_language",
        "city_id",
        "coordinate",
        "country_id",
        "county_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "identifier",
        "is_approved",
        "is_community_active",
        "is_test_data",
        "label_id",
        "location_id",
        "name",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "profile_id",
        "region_id",
        "start_timestamp",
        "state_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "label_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "label_id",
        "label_ml_id",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_general_view": (
        "coordinate",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "group.name",
        "group.title",
        "group_id",
        "identifier",
        "is_test_data",
        "iso",
        "iso3",
        "name",
        "nicename",
        "numcode",
        "phonecode",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_set_parents_table": (
        "condition_type_id",
        "criteria_set_id",
        "criteria_set_parents_id",
        "include_exclude_type_id",
        "parent_criteria_set_id"
    ),
    "location_field_ml_table": (
        "field_name",
        "is_test_data",
        "lang_code",
        "location_id",
        "locaton_field_ml_id",
        "prime",
        "title",
        "title_approved"
    ),
    "country_ml_table": (
        "country_id",
        "country_id_old",
        "country_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "county_ml_table": (
        "county_id",
        "county_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "state_table": (
        "coordinate",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "end_timestmap",
        "group_id",
        "is_test_data",
        "name",
        "number",
        "state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "county_ml_with_deleted_and_test_data_view": (
        "county_id",
        "county_ml_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "state_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_ml_with_deleted_and_test_data_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "test_state_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "entity_type_table": (
        "created_timestamp",
        "created_user_id",
        "criteria_schema_name",
        "criteria_table_name",
        "criteria_view_name",
        "end_timestamp",
        "entity_type_id",
        "is_item_type",
        "is_test_data",
        "name",
        "schema_name",
        "start_timestamp",
        "system_id",
        "table_name",
        "updated_timestamp",
        "updated_user_id",
        "view_name",
        "view_with_deleted_and_test_data"
    ),
    "activity_group_location_occurrence_profile_table": (
        "activity_group_location_occurrence_profile_id",
        "activity_id",
        "activity_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_list_id",
        "location_id",
        "location_list_id",
        "occurrence_id",
        "occurrence_list_id",
        "profile_id",
        "profile_list_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "county_table": (
        "coordinate",
        "country_id",
        "county_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "number",
        "state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "region_ml_with_deleted_and_test_data_view": (
        "country_id",
        "created_timestamp",
        "created_user_id",
        "group_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "region.end_timestamp",
        "region_id",
        "region_ml.end_timestamp",
        "region_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "neighborhood_ml_with_deleted_and_test_data_view": (
        "city_id",
        "created_timestamp",
        "created_user_id",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "neighborhood_id",
        "neighborhood_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_profile_view": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_id",
        "criteria_profile_id",
        "identifier",
        "is_approved",
        "is_test_data",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "location_ml_view": (
        "created_timestamp",
        "created_user_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "location_id",
        "location_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_ml_table": (
        "created_timestamp",
        "created_user_id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "location_id",
        "location_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "city_ml_table": (
        "city_id",
        "city_ml_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "neighborhood_ml_table": (
        "created_timestamp",
        "created_user_id",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "neighborhood_id",
        "neighborhood_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "neighborhood_table": (
        "city_id",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "neighborhood_id",
        "number",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "region_table": (
        "coordinate",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "number",
        "region_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "region_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "region_id",
        "region_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_view": (
        "coordinate",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "group_id",
        "identifier",
        "is_test_data",
        "iso",
        "iso3",
        "name",
        "nicename",
        "numcode",
        "phonecode",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_template_text_block_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "lang_code",
        "message_template_text_block_type_id",
        "message_template_text_block_type_ml_id",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "email_address_profile_general_view": (
        "contact_email_address_id",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_address_id",
        "email_address_of_contact",
        "email_address_of_person",
        "email_address_of_profile",
        "email_address_person_id",
        "email_address_profile_id",
        "end_timestamp",
        "is_main",
        "is_sure",
        "person_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_set_profile_view": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_set_id",
        "criteria_set_profile_id",
        "identifier",
        "is_test_data",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "form_type_table": (
        "channel_type_id",
        "form_type_table_id",
        "name",
        "version"
    ),
    "form_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "form_id",
        "form_type_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_list_member_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "location_id",
        "location_list_id",
        "location_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_list_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "label_list_id",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "label_list_id",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_list_member_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "label_id",
        "label_list_id",
        "label_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_list_member_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "label_id",
        "label_list_id",
        "label_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "city_list_member_table": (
        "city_id",
        "city_list_id",
        "city_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "city_list_member_view": (
        "city_id",
        "city_list_id",
        "city_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "state_list_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "start_timestamp",
        "state_list_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "state_list_member_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "state_id",
        "state_list_id",
        "state_list_member_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "state_list_member_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "state_id",
        "state_list_id",
        "state_list_member_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_list_table": (
        "country_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_list_member_table": (
        "country_id",
        "country_list_id",
        "country_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "country_list_member_view": (
        "country_id",
        "country_list_id",
        "country_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "county_list_table": (
        "county_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "county_list_member_table": (
        "county_id",
        "county_list_id",
        "county_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "county_list_member_view": (
        "county_id",
        "county_list_id",
        "county_list_member_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "reaction_type_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "image",
        "is_test_data",
        "name",
        "number",
        "reaction_id",
        "reaction_type_id",
        "start_timestamp",
        "text",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "reaction_type_ml_view": (
        "lang_code",
        "name",
        "reaction_type_id",
        "title"
    ),
    "organization_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "organization_id",
        "organization_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_general_view": (
        "action_id",
        "action_stars",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "destination_subsystem_id",
        "destination_system_id",
        "end_timestamp",
        "expiration_unit",
        "expiration_value",
        "filename",
        "folder_name",
        "function_module",
        "function_name",
        "is_queue_action",
        "is_smartlink_action",
        "is_test_data",
        "name",
        "package_name",
        "source_system_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "reaction_type_ml_general_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "image",
        "is_test_data",
        "name",
        "number",
        "reaction_id",
        "reaction_type_id",
        "reaction_type_ml.end_timestamp",
        "reaction_type_ml.is_test_data",
        "reaction_type_ml.lang_code",
        "start_timestamp",
        "text",
        "title",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "reaction_type_ml_with_deleted_and_test_data_view": (
        "end_timestamp",
        "is_test_data",
        "lang_code",
        "name",
        "reaction_type_id",
        "reaction_type_ml.end_timestamp",
        "reaction_type_ml.is_test_data",
        "title"
    ),
    "reaction_ml_with_deleted_and_test_data_view": (
        "image",
        "lang_code",
        "reaction.is_test_data",
        "reaction.title",
        "reaction_category_id",
        "reaction_category_ml.is_test_data",
        "reaction_category_ml.lang_code",
        "reaction_category_ml.title",
        "reaction_id",
        "reaction_ml.is_test_data",
        "reaction_type_id",
        "reaction_type_ml.is_test_data",
        "reaction_type_ml.lang_code",
        "text",
        "title",
        "value"
    ),
    "reaction_ml_general_view": (
        "image",
        "is_main",
        "is_test_data",
        "lang_code",
        "reaction.title",
        "reaction_category_id",
        "reaction_category_ml.title",
        "reaction_id",
        "reaction_type_id",
        "text",
        "title",
        "value"
    ),
    "smartlink_type_general_view": (
        "action_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id",
        "dialog_workflow_script_id_old",
        "end_timestamp",
        "expiration_unit",
        "expiration_value",
        "identifier",
        "is_test_data",
        "number",
        "smartlink.name",
        "smartlink_type_id",
        "start_state_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "reaction_category_ml_table": (
        "description",
        "id",
        "is_test_data",
        "lang_code",
        "reaction_category_id",
        "title"
    ),
    "reaction_ml_view_old": (
        "description",
        "image",
        "lang_code",
        "reaction_id",
        "reaction_ml_id",
        "reaction_table.is_test_data",
        "reaction_type_id",
        "text",
        "title",
        "updated_timestamp",
        "value"
    ),
    "reaction_ml_view": (
        "description",
        "image",
        "lang_code",
        "reaction_id",
        "reaction_ml_id",
        "reaction_type_id",
        "title",
        "value"
    ),
    "reaction_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "identifier",
        "is_description_approved",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "reaction_id",
        "reaction_ml_id",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "data_source_instance_table": (
        "computer_name",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_type_id",
        "end_index",
        "end_timestamp",
        "file_or_api",
        "filename",
        "identifier",
        "import_or_last_sync_timestamp",
        "imported_file_created_timestamp",
        "is_test_data",
        "name",
        "number",
        "path",
        "start_index",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "visibility_id"
    ),
    "system_entity_table": (
        "system_entity_id",
        "system_id",
        "title"
    ),
    "system_view": (
        "name",
        "parent_system_id",
        "system_id"
    ),
    "system_ml_view": (
        "lang_code",
        "system_id",
        "system_ml_id",
        "title"
    ),
    "url_type_regex_table": (
        "example",
        "url_type_id",
        "url_type_regex",
        "url_type_regex_id"
    ),
    "field_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "database_field_name",
        "database_sub_field_name",
        "database_sub_field_value",
        "description",
        "end_timestamp",
        "field_id",
        "identifier",
        "internal_description",
        "is_ml_table",
        "is_test_data",
        "name",
        "number",
        "processing_database_field_name",
        "processing_id",
        "start_timestamp",
        "table_id",
        "table_name",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "variable.name",
        "variable_id",
        "visibility_id"
    ),
    "variable_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "field_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "person_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "variable_id",
        "visibility_id"
    ),
    "data_source_type_general_view": (
        "created_timestamp",
        "created_user_id",
        "data_source_category_id",
        "data_source_type_id",
        "data_source_type_ml.title",
        "end_timestamp",
        "is_test_data",
        "name",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "smartlink_table": (
        "action_id",
        "campaign_id",
        "clicked_timestamp",
        "content_message_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id",
        "end_timestamp",
        "from_contact_id",
        "from_email_address_id",
        "from_email_address_old",
        "from_normalized_phone",
        "from_person_id",
        "from_phone_id",
        "from_profile_id",
        "from_user_id",
        "identifier",
        "is_test_data",
        "lang_code",
        "message_template_id",
        "number",
        "smartlink_id",
        "smartlink_message_id",
        "smartlink_type_id",
        "start_timestamp",
        "to_contact_id",
        "to_email_address_id",
        "to_email_address_old",
        "to_group_id",
        "to_normalized_phone_old",
        "to_person_id",
        "to_phone_id",
        "to_profile_id",
        "to_user_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_redirect"
    ),
    "smartlink_type_table": (
        "action_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id_old",
        "end_timestamp",
        "expiration_unit",
        "expiration_value",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "smartlink_type_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_redirect_template"
    ),
    "location_ml_with_deleted_and_test_data_view": (
        "address_english",
        "address_local_language",
        "city_id",
        "coordinate",
        "country_id",
        "county_id",
        "created_timestamp",
        "group_id",
        "identifier",
        "is_approved",
        "is_community_active",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "location.created_effective_profile_id",
        "location.created_effective_user_id",
        "location.created_real_user_id",
        "location.created_timestamp",
        "location.created_user_id",
        "location.end_timestamp",
        "location.is_test_data",
        "location.start_timestamp",
        "location.updated_effective_profile_id",
        "location.updated_effective_user_id",
        "location.updated_real_user_id",
        "location.updated_timestamp",
        "location.updated_user_id",
        "location.visibility_id",
        "location_id",
        "location_ml.created_user_id",
        "location_ml_id",
        "name",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "region_id",
        "state_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "form_message_template_table": (
        "criteria_set_id",
        "form_id",
        "form_message_template_id",
        "max_message_template",
        "message_template_id",
        "min_message_template",
        "page",
        "seq"
    ),
    "url_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "normalized_url",
        "number",
        "start_timestamp",
        "storage_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "url_id",
        "url_type"
    ),
    "logger_all_fields_updates": (
        "internal_description",
        "name",
        "variable.name",
        "Website"
    ),
    "person_profile_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "identifier",
        "is_test_data",
        "number",
        "person.end_timestamp",
        "person.first_name",
        "person.last_name",
        "person.main_email_address",
        "person_id",
        "person_profile.end_timestamp",
        "person_profile_id",
        "person_profile_role_id",
        "profile.end_timestamp",
        "profile.is_test_data",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile.profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_message_template_general_view": (
        "campaign_id",
        "campaign_message_template.percent",
        "email_body_html_template",
        "email_subject_template",
        "message_template.name",
        "message_template_id",
        "message_template_text_block_name",
        "message_template_text_block_seq",
        "message_template_text_block_type.title",
        "question.name",
        "question.title",
        "question_id",
        "sms_body_template",
        "whatsapp_body_template"
    ),
    "job_title_ml_en_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title.name",
        "job_title_id",
        "job_title_ml.is_title_approved",
        "job_title_ml.lang_code",
        "job_title_ml.title",
        "job_title_ml_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "job_title_vacancy_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title_id",
        "job_title_vacancy_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "vacancy_id"
    ),
    "last_name_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "last_name_id",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "contact_recent_general_view": (
        "contact_id",
        "email1",
        "first_name",
        "last_name",
        "updated_timestamp"
    ),
    "logger_last_session_general_view": (
        "created_timestamp",
        "session"
    ),
    "group_type_ml_table": (
        "group_type_id",
        "group_type_ml_id",
        "lang_code",
        "title"
    ),
    "state_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "state_abbr",
        "state_abbr_approved",
        "state_id",
        "state_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "table_definition_table": (
        "contact_mapping_table_id",
        "description",
        "end_timestamp",
        "entity_type_id1",
        "entity_type_id2",
        "insert_is_undelete",
        "is_created_updated_fields",
        "is_dynamic",
        "is_entity",
        "is_graph",
        "is_history",
        "is_logical_delete_fields",
        "is_main_column",
        "is_mapping_table",
        "is_metadata",
        "is_metadata_whole_table",
        "is_metrics",
        "is_ml_child_table",
        "is_ml_parent_table",
        "is_number_column",
        "is_one_table_per_all_tenants",
        "is_pii",
        "is_sensative",
        "is_single_and_multi_value",
        "is_test_data_column",
        "is_tree",
        "is_user_data",
        "is_user_data_whole_table",
        "is_visibility",
        "name_of_single_value_column",
        "profile_mapping_table_id",
        "schema",
        "sensative_columns",
        "seq",
        "table_definition_id",
        "table_list_id",
        "table_name",
        "table_type",
        "view_name"
    ),
    "campaign_type_table": (
        "campaign_type_id",
        "name"
    ),
    "campaign_campaign_type_table": (
        "campaign_campaign_type_id",
        "campaign_id",
        "campaign_type_id"
    ),
    "group_group_type_view": (
        "group_group_type_id",
        "group_id",
        "group_type_id"
    ),
    "role_table": (
        "group_id",
        "name",
        "role_id"
    ),
    "group_type_ml_view": (
        "group_type_id",
        "group_type_ml_id",
        "lang_code",
        "name",
        "title"
    ),
    "message_status_type_table": (
        "message_status_id",
        "name"
    ),
    "message_status_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "lang_code",
        "message_status_type_id",
        "message_status_type_ml_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "person_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "end_timestamp",
        "first_name",
        "identifier",
        "is_first_name_approved",
        "is_last_name_approved",
        "is_test_data",
        "lang_code",
        "last_name",
        "number",
        "person_id",
        "person_ml_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "data_source_instance_view": (
        "computer_name",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_type_id",
        "end_index",
        "end_timestamp",
        "file_or_api",
        "filename",
        "identifier",
        "import_or_last_sync_timestamp",
        "imported_file_created_timestamp",
        "is_test_data",
        "name",
        "number",
        "path",
        "start_index",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "visibility_id"
    ),
    "action_item_profile_table_old": (
        "action_item_id",
        "action_item_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "profile_id1",
        "profile_id2",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "profile_general_circ_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "internal_description",
        "is_approved",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "main_person_id",
        "main_phone_id",
        "number",
        "person.first_name",
        "person.last_name",
        "person.person_id",
        "preferred_lang_code",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_person_id",
        "profile.main_user_id",
        "profile.name",
        "profile_id",
        "profile_ml.title",
        "profile_type_id",
        "profile_type_ml.title",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "contact_group_table": (
        "attribute",
        "consumer_category_id",
        "contact_group_id",
        "contact_group_type_id",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_main_in_category",
        "is_potential",
        "is_recommende_by_ai",
        "is_recommended_by_contact_notes",
        "is_sure",
        "is_test_data",
        "months",
        "number",
        "participant_category_id",
        "relationship_type_id",
        "seq",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "supplier_category_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "job_title_profile_view": (
        "end_timestamp",
        "job_title_id",
        "job_title_ml_id",
        "job_title_profile_id",
        "profile_id"
    ),
    "importer_recent_contact_job_title_view": (
        "contact.job_title",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "entity_id",
        "entity_type_id",
        "group_ml.title",
        "handle",
        "importer_id",
        "is_test_data",
        "job_title_ml_from_contact.job_title_ml_id",
        "job_title_ml_from_contact.title",
        "job_title_ml_from_profile.job_title_ml_id",
        "job_title_ml_from_profile.title",
        "job_title_profile_id",
        "location_id",
        "name",
        "profile_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "group_profile_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_profile_type_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "importer_recent_contact_action_item_view": (
        "action_item.name",
        "action_item_ml.title",
        "action_item_profile_id",
        "activity_id",
        "country_ml.title",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "handle",
        "importer_id",
        "is_test_data",
        "location_id",
        "profile.name",
        "profile_id",
        "start_timestamp",
        "system_ml.title",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "campaign_message_profile_general_view": (
        "campaign_id",
        "campaign_message_profile_id",
        "channel_id",
        "channel_id_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "message_id",
        "number",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "campaign_message_profile_view": (
        "campaign_id",
        "campaign_message_profile_id",
        "channel_id_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "message_id",
        "number",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "profile_user_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_active",
        "is_main",
        "profile_id",
        "profile_user_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_id"
    ),
    "action_item_general_view": (
        "action_id",
        "action_item_id",
        "action_item_type_id",
        "active_main_profile_id",
        "activity_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "other_group_id",
        "other_person_id",
        "other_profile_id",
        "passive_main_contact_id",
        "passive_main_profile_id",
        "ranking",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "action_item_profile_table": (
        "action_item_id",
        "action_item_ml_id",
        "action_item_profile_id",
        "active_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "passive_contact_id",
        "passive_profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "job_title_ml_with_deleted_and_test_data_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title.name",
        "job_title_id",
        "job_title_ml.created_timestamp",
        "job_title_ml.created_user_id",
        "job_title_ml.end_timestamp",
        "job_title_ml.is_title_approved",
        "job_title_ml.lang_code",
        "job_title_ml.start_timestamp",
        "job_title_ml.title",
        "job_title_ml.updated_timestamp",
        "job_title_ml.updated_user_id",
        "job_title_ml_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "job_title_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title.name",
        "job_title_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "job_title_ml_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "job_title_id",
        "job_title_ml.is_main",
        "job_title_ml.is_title_approved",
        "job_title_ml.lang_code",
        "job_title_ml.title",
        "job_title_ml_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subsystem_ml_genetal_view": (
        "name",
        "parent_system_id",
        "subsystem_ml.title",
        "system_id",
        "system_ml.title"
    ),
    "dialog_workflow_script_with_forms_view": (
        "dialog_workflow_state_id",
        "lang_code",
        "name",
        "next_state_id_if_there_is_no_feedback",
        "parameter1",
        "parent_state_id",
        "state_id",
        "title",
        "workflow_action_id"
    ),
    "message_template_message_template_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "message_template_id",
        "message_template_message_template_text_block_id",
        "message_template_text_block_id",
        "message_template_text_block_type_id",
        "seq",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "workflow_action_table": (
        "description",
        "name",
        "software_development_status",
        "workflow_action_id"
    ),
    "importer_table": (
        "created_timestamp",
        "created_user_id",
        "data_imported_json",
        "data_source_external_user_id_old",
        "data_source_instance_id",
        "data_source_type_id_old",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "google_calendar_id",
        "google_people_api_resource_name",
        "importer_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external_country_id_old",
        "user_external_id"
    ),
    "profile_variable_table": (
        "answer_number",
        "answer_text",
        "created_timestamp",
        "expiration_timestamp",
        "profile_id",
        "profile_variable_id",
        "variable_id"
    ),
    "user_general_circles_view": (
        "brand_id",
        "person.birthday_date",
        "person.is_approved",
        "person.main_email_address",
        "person_id",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.phone.full_number_normalized",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "role_name",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.active_location_id",
        "user.created_timestamp",
        "user.first_name",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_id",
        "username"
    ),
    "profile_question_table": (
        "answer_number",
        "answer_text",
        "created_timestamp",
        "expiration_timestamp",
        "profile_id",
        "profile_question_id",
        "question_id",
        "user_id"
    ),
    "content_set_table": (
        "content_set_id",
        "name",
        "parent_content_set_id"
    ),
    "message_channel_content_set_profile_table": (
        "content_set_id",
        "is_opt_out",
        "is_subscribed",
        "message_channel_content_set_profile_table",
        "message_channel_id",
        "profile_id"
    ),
    "criteria_person_table_old": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_id",
        "criteria_person_id",
        "identifier",
        "is_approved",
        "is_test_data",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "dialog_workflow_script_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "dialog_workflow_script_id",
        "dialog_workflow_script_status_text",
        "end_timestamp",
        "is_test_data",
        "max_state_id",
        "min_state_id",
        "name",
        "start_state_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "dialog_workflow_state_table": (
        "channel_id",
        "checked_in",
        "criteria_id",
        "criteria_set_id",
        "dialog_workflow_state_id",
        "end_timestamp",
        "lang_code",
        "message_template_id",
        "name",
        "next_state_id",
        "next_state_id_if_there_is_no_feedback",
        "no_feedback_milliseconds",
        "parameter1",
        "parent_state_id",
        "remark",
        "result_figure_max",
        "result_figure_min",
        "result_logical",
        "state_id",
        "variable1_id",
        "variable2_id",
        "version",
        "workflow_action_id"
    ),
    "job_title_profile_table": (
        "end_timestamp",
        "job_title_id",
        "job_title_ml_id",
        "job_title_profile_id",
        "profile_id"
    ),
    "group_job_title_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_job_title_id",
        "group_ml_id",
        "job_title_id",
        "job_title_ml_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "seller_category_table": (
        "seller_category_id"
    ),
    "cloud_region_county_table": (
        "cloud_region_country_id",
        "cloud_region_id",
        "county_id"
    ),
    "brand_cloud_region_county_table": (
        "brand_cloud_region_country_id",
        "brand_id",
        "cloud_region_id",
        "county_id",
        "error_message"
    ),
    "criteria_profile_general_view": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_id",
        "criteria_profile_id",
        "identifier",
        "is_approved",
        "is_main",
        "is_test_data",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "advertisement_costs_table": (
        "activity_id",
        "advertisement_costs_id",
        "category_id",
        "cost_max_usd",
        "cost_min_usd",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "model",
        "per",
        "source",
        "start_timestamp",
        "subscription_id",
        "updated_timestamp",
        "updated_user_id",
        "vertical_id"
    ),
    "advertisement_type_ml_table": (
        "advertisement_type_id",
        "advertisement_type_ml_id",
        "lang_code",
        "title"
    ),
    "real_estate_listing_view": (
        "agent_name",
        "agent_office_phone",
        "building_size",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "land_size",
        "num_of_bathrooms",
        "num_of_bedrooms",
        "number",
        "price",
        "property_type",
        "real_estate_listing_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "importer_recent_real_estate_view": (
        "agent_name",
        "agent_office_phone",
        "building_size",
        "country_ml.title",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "handle",
        "identifier",
        "importer.is_test_data",
        "importer_id",
        "is_test_data",
        "land_size",
        "location_id",
        "num_of_bathrooms",
        "num_of_bedrooms",
        "number",
        "price",
        "profile_id",
        "property_type",
        "re.created_effective_profile_id",
        "re.created_effective_user_id",
        "re.created_real_user_id",
        "re.created_timestamp",
        "re.created_user_id",
        "re.end_timestamp",
        "re.is_test_data",
        "re.start_timestamp",
        "re.updated_effective_profile_id",
        "re.updated_effective_user_id",
        "re.updated_real_user_id",
        "re.updated_timestamp",
        "re.updated_user_id",
        "re.visibility_id",
        "real_estate_listing_id",
        "start_timestamp",
        "system_ml.title",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "person_promotional_message_ml_table": (
        "created_timestamp",
        "created_user_id",
        "lang_code",
        "person_promotional_message_id",
        "person_promotional_message_ml_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "real_estate_listing_table": (
        "agent_name",
        "agent_office_phone",
        "building_size",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "land_size",
        "num_of_bathrooms",
        "num_of_bedrooms",
        "number",
        "price",
        "property_type",
        "real_estate_listing_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "label_location_general_view": (
        "is_approved",
        "label.name",
        "label_id",
        "label_location_id",
        "label_ml.title",
        "location.name",
        "location_id",
        "location_ml.title"
    ),
    "location_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "location_type_id",
        "name",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_location_type_view": (
        "location_id",
        "location_location_type_id",
        "location_ml_id",
        "location_type_id",
        "location_type_ml_id"
    ),
    "location_location_type_geneal_view": (
        "location.name",
        "location_id",
        "location_location_type_id",
        "location_ml.title",
        "location_ml_id",
        "location_type.name",
        "location_type_id",
        "location_type_ml.title",
        "location_type_ml_id"
    ),
    "location_location_type_geneal_review_view": (
        "location.name",
        "location_id",
        "location_location_type_id",
        "location_ml.title",
        "location_ml_id",
        "location_type.name",
        "location_type_id",
        "location_type_ml.title",
        "location_type_ml_id"
    ),
    "location_location_type_table": (
        "location_id",
        "location_location_type_id",
        "location_ml_id",
        "location_type_id",
        "location_type_ml_id"
    ),
    "message_template_text_block_question_review_view": (
        "default_body_template",
        "lang_code",
        "message_template_text_block_id",
        "name",
        "question_id"
    ),
    "question_ml_table": (
        "default_value",
        "gender_id",
        "lang_code",
        "question_id",
        "question_ml_id",
        "title"
    ),
    "conversation_ml_table": (
        "conversation_id",
        "conversation_ml_table",
        "lang_code",
        "title"
    ),
    "location_table": (
        "address_english",
        "address_local_language",
        "city_id",
        "city_ml_id",
        "coordinate",
        "country_id",
        "county_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "identifier",
        "is_approved",
        "is_community_active",
        "is_official",
        "is_test_data",
        "location_id",
        "name",
        "neighborhood_id",
        "number",
        "phonecode",
        "postal_code",
        "region_id",
        "start_timestamp",
        "state_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "city_list_table": (
        "city_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "city_table": (
        "city_id",
        "coordinate",
        "country_id",
        "country_id_old",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_city_id",
        "number",
        "phonecode",
        "state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_table": (
        "coordinate",
        "country_id_old",
        "county_id_old",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "entity_type_id",
        "gender_list_id_old",
        "group_id",
        "group_list_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "label_id",
        "location_id",
        "location_list_id",
        "location_type_id",
        "max_age_old",
        "max_height_old",
        "max_kids_age_old",
        "max_number_of_kids_old",
        "min_age_old",
        "min_height_old",
        "min_kids_age_old",
        "min_number_of_kids_old",
        "name",
        "number_of_partners_old",
        "organization_name",
        "partner_experience_level_old",
        "profile_list_id",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_timestamp",
        "template_condition",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id"
    ),
    "organization_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_approved",
        "is_main",
        "is_test_data",
        "location_id",
        "main_group_id",
        "members_visibility_scope_id",
        "name",
        "Non_members_visibility_profile_id",
        "non_members_visibility_scope_id",
        "number",
        "organization_id",
        "parent_organization_id",
        "point",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "organization_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "job_title_id",
        "job_title_ml_id",
        "number",
        "organization_id",
        "organization_ml_id",
        "organization_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "contact_table": (
        "account_name",
        "additional_name",
        "address1_city",
        "address1_country",
        "address1_postal_code",
        "address1_state",
        "address1_street",
        "address2_city",
        "address2_country",
        "address2_postal_code",
        "address2_state",
        "address2_street",
        "anniversary",
        "birthday",
        "cira",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "day",
        "department",
        "display_as",
        "email1",
        "email2",
        "email3",
        "end_timestamp",
        "first_name",
        "full_name",
        "identifier",
        "is_test_data",
        "job_title",
        "job_title_id",
        "job_title_ml_id",
        "last_name",
        "last_sync_timestamp",
        "main_profile_id",
        "month",
        "name_prefix",
        "name_suffix",
        "nickname",
        "notes",
        "number",
        "organization",
        "organization_profile_id",
        "original_first_name",
        "original_last_name",
        "owner_profile_id",
        "person_id",
        "phone1",
        "phone2",
        "phone3",
        "photo_file_name",
        "photo_url",
        "source",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "uploaded_timestamp",
        "visibility_id",
        "website1",
        "website2",
        "website3",
        "year"
    ),
    "contact_organization_general_view": (
        "contact.first_name",
        "contact.job_title_ml.job_title_ml_id",
        "contact.job_title_ml.title",
        "contact.last_name",
        "contact_id",
        "contact_profile_id",
        "group.name",
        "group_id",
        "group_ml.title",
        "job_title_ml.title",
        "organization_ml.title",
        "organization_profile.job_title_ml.title",
        "profile.job_title_ml.job_title_ml_id",
        "profile.job_title_ml.title",
        "profile_id"
    ),
    "location_criteria_table": (
        "city_id",
        "city_list_id",
        "country_id",
        "county_id",
        "criteria_id",
        "location_id_old",
        "location_list_id_old",
        "location_type_id",
        "neighbourhood_id",
        "state_id",
        "street_id"
    ),
    "smartlink_general_view": (
        "campaign_id",
        "created_user.username",
        "created_user_id",
        "dialog_workflow_script_id_old",
        "from_email_address_old",
        "identifier",
        "lang_code",
        "smartlink.action_id",
        "smartlink.created_timestamp",
        "smartlink.is_test_data",
        "smartlink.name",
        "smartlink.number",
        "smartlink_id",
        "smartlink_type.action_id",
        "smartlink_type_id",
        "to_email_address_old"
    ),
    "smartlink_general_recent_view": (
        "campaign_id",
        "created_user.username",
        "created_user_id",
        "dialog_workflow_script_id_old",
        "from_email_address_old",
        "identifier",
        "lang_code",
        "smartlink.action_id",
        "smartlink.created_timestamp",
        "smartlink.is_test_data",
        "smartlink.name",
        "smartlink.number",
        "smartlink_id",
        "smartlink_type.action_id",
        "smartlink_type_id",
        "to_email_address_old"
    ),
    "logger_general_view": (
        "component_id",
        "created_timestamp",
        "logger_id",
        "name",
        "programming_language",
        "programming_language_id",
        "server_ip_v4",
        "server_ip_v6"
    ),
    "logger_general_recent_view": (
        "component_id",
        "created_timestamp",
        "logger_id",
        "name",
        "programming_language",
        "programming_language_id",
        "server_ip_v4",
        "server_ip_v6"
    ),
    "activity_table": (
        "action_id",
        "action_item_category_id",
        "activity_id",
        "activity_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "parent_activity_id",
        "ranking",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "star_transaction_table": (
        "action_id",
        "action_stars",
        "activity_type_id",
        "api_type_id",
        "balance_after_transaction",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "invitation_id",
        "is_conditional",
        "is_test_data",
        "number",
        "parent_star_transaction_id",
        "person_id",
        "person_id2",
        "profile_id",
        "profile_id2",
        "session",
        "star_transaction_id",
        "start_timestamp",
        "subscription_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_id2"
    ),
    "subject_table": (
        "name",
        "subject_id"
    ),
    "lesson_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "duration_minutes",
        "end_timestamp",
        "lesson_id",
        "local_curency_id",
        "platform",
        "platform_id",
        "price_local_currency",
        "start_timestamp",
        "subject_id",
        "teacher_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "course_table": (
        "business_education_profile_id",
        "course_id",
        "name",
        "semester",
        "subject_id"
    ),
    "lang_code_table": (
        "lang_code",
        "lang_code_id",
        "name"
    ),
    "group_user_external_table": (
        "group_id",
        "group_user_external_id",
        "user_external_id"
    ),
    "importer_recent_general_facebook_view": (
        "country_ml.title",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "data_source_instance_name",
        "data_source_type_id",
        "data_source_type_id_old",
        "data_source_type_ml.title",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "handle",
        "importer.is_test_data",
        "importer_id",
        "is_test_data",
        "location_id",
        "profile_id",
        "start_timestamp",
        "system_ml.title",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external.system_id",
        "user_external_country_id_old",
        "user_external_id",
        "username"
    ),
    "group_label_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_label_id",
        "hashtag",
        "identifier",
        "is_test_data",
        "label_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_label_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_label_id",
        "hashtag",
        "identifier",
        "is_test_data",
        "label_id",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "queue_item_table": (
        "action_id",
        "class_parameters_json",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "execution_end_time",
        "execution_start_time",
        "function_parameters_json",
        "identifier",
        "is_test_data",
        "item_description",
        "number",
        "process_id",
        "queue_item_id",
        "queue_status_id",
        "return_code",
        "return_json",
        "return_message",
        "returned_value",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "start_timestamp",
        "stderr",
        "stdout",
        "thread_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt"
    ),
    "group_url_table": (
        "created_timestamp",
        "group_id",
        "group_url_id",
        "is_active",
        "is_tal_il_joined_temp",
        "number_of_members_temp",
        "system_id_temp",
        "url_id",
        "url_temp"
    ),
    "subscription_table": (
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_paid",
        "name",
        "start_timestamp",
        "subscription_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_action_limit_template_table": (
        "action_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "fix_limit_units",
        "fixed_limit",
        "hard_limit_value",
        "limit_unit",
        "soft_limit_value",
        "start_timestamp",
        "subscription_action_limit_template_id",
        "subscription_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "brand_ml_table": (
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "id",
        "is_main_title",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_profile_table": (
        "name",
        "test_profile_id"
    ),
    "test_location_profile_mapping_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "test_location_id",
        "test_location_profile_id",
        "test_profile_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_location_profile_mapping_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "start_timestamp",
        "test_location_id",
        "test_location_profile_id",
        "test_profile_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_general_view": (
        "action.name",
        "action_id",
        "brand_id",
        "brand_name",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "hard_limit_value",
        "is_paid",
        "start_timestamp",
        "subscription.name",
        "subscription_id",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_addons_table": (
        "action_id",
        "name",
        "subscription_addons_id"
    ),
    "action_general_without_subscription_view": (
        "action_id",
        "action_stars",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "destination_subsystem_id",
        "destination_system_id",
        "end_timestamp",
        "expiration_unit",
        "expiration_value",
        "filename",
        "folder_name",
        "function_module",
        "function_name",
        "is_queue_action",
        "is_smartlink_action",
        "is_test_data",
        "name",
        "package_name",
        "source_system_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "seller_ml_view": (
        "coordinate",
        "entity_type_id",
        "seller.name",
        "seller.title"
    ),
    "seller_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "name",
        "profile_id",
        "seller_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "shop_table": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "name",
        "owner_seller_id",
        "shop_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "shop_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "id",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "shop_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "shop_general_view": (
        "profile.name",
        "seller.name",
        "seller_ml.title",
        "shop.name"
    ),
    "profile_ml_general_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "internal_description",
        "is_approved",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "main_phone_id",
        "number",
        "person.first_name",
        "person.last_name",
        "person.person_id",
        "preferred_lang_code",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_person_id",
        "profile.main_user_id",
        "profile.name",
        "profile_id",
        "profile_ml.title",
        "profile_type_id",
        "profile_type_ml.title",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_location_mapping_view": (
        "is_test_data",
        "test_county_id",
        "test_county_location_mapping_id",
        "test_location_id"
    ),
    "test_county_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "person_promotional_message_view": (
        "created_timestamp",
        "created_user_id",
        "criteria_set_id",
        "dialog_workflow_state_id",
        "is_test_data",
        "join_group_list_id_if_choosen",
        "message_template_id",
        "name",
        "person_id",
        "person_promotional_message_id",
        "priority",
        "promotional_message_id",
        "sql_where",
        "updated_timestamp",
        "updated_user_id"
    ),
    "person_promotional_message_table": (
        "created_timestamp",
        "created_user_id",
        "criteria_set_id",
        "dialog_workflow_script_id",
        "dialog_workflow_state_id",
        "is_test_data",
        "join_group_list_id_if_choosen",
        "message_template_id",
        "name",
        "person_id",
        "person_promotional_message_id",
        "priority",
        "promotional_message_id",
        "question_id",
        "sql_where",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_costs_view": (
        "action.is_messaging",
        "action.name",
        "action_id",
        "hard_limit_value",
        "hard_limit_value*price_per_unit",
        "price_per_unit",
        "subscription_id",
        "subsctiption.name"
    ),
    "newsletter_table": (
        "duration",
        "name",
        "newsletter_id",
        "owner_profile_id",
        "owner_profile_name_temp",
        "root_message_id",
        "system_id",
        "url_id",
        "url_temp"
    ),
    "advertisement_table": (
        "advertisement_id",
        "is_actionable",
        "seller_id"
    ),
    "advertisement_campaign_table": (
        "advertisement_campaign_id",
        "advertisement_id",
        "campaign_id"
    ),
    "advertisement_storage_table": (
        "advertisement_id",
        "advertisement_storage_id",
        "html_map_tag",
        "is_actionable",
        "storge_id"
    ),
    "coupon_table": (
        "coupon_id",
        "event_id",
        "from.profile_id",
        "identifier",
        "name",
        "number",
        "quantity"
    ),
    "event_participant_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "event_participant_id",
        "participant_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "brand_event_table": (
        "brand_event_id",
        "brand_id",
        "event_id"
    ),
    "event_sponsor_table": (
        "event_id",
        "event_sponsor_id",
        "event_sponsorship_type",
        "sponsor_profile_id"
    ),
    "test_typeorm_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "new_test_typeorm_id",
        "start_timestamp",
        "test_typeorm_id",
        "unique_int_column",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_table": (
        "action_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "destination_subsystem_id",
        "destination_system_id",
        "dialog_workflow_script_id",
        "end_timestamp",
        "expiration_unit",
        "expiration_value",
        "filename",
        "folder_name",
        "function_module",
        "function_name",
        "is_messaging",
        "is_queue_action",
        "is_smartlink_action",
        "is_test_data",
        "name",
        "package_name",
        "source_system_id",
        "start_timestamp",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "event_ml_id",
        "identifier",
        "is_main_title",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "event_event_table": (
        "action_id",
        "event_event_id",
        "event_id1",
        "event_id2",
        "is_main",
        "is_move_everything_from_event1_to_event2",
        "relationship_type_id"
    ),
    "event_event_category_table": (
        "event_category_id",
        "event_event_category_id",
        "event_id"
    ),
    "event_category_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_category_id",
        "event_type_id",
        "is_hierarchy",
        "is_location",
        "is_organization",
        "is_prioritze",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_category_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "event_category_id",
        "event_category_ml_id",
        "is_main_title",
        "lang_code",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_type_table": (
        "event_type_id",
        "examples_old",
        "is_contact_classification",
        "is_gender",
        "is_job_title",
        "is_location",
        "is_organization",
        "is_organization_classification",
        "name"
    ),
    "event_type_ml_table": (
        "event_type_id",
        "event_type_ml_id",
        "lang_code",
        "title"
    ),
    "event_event_type_table": (
        "event_event_type_id",
        "event_id",
        "event_type_id"
    ),
    "event_general_test_data_view_temp": (
        "event.end_timestamp",
        "event.is_test_data",
        "event.name",
        "event_category_title",
        "event_id",
        "event_id1 (parent)",
        "event_id2 (child)",
        "event_ml.end_timestamp",
        "event_ml.is_test_data",
        "event_ml_id",
        "event_type_name",
        "is_contact_classification",
        "is_location",
        "is_main_title",
        "lang_code",
        "new_event_id",
        "title"
    ),
    "group_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "group_id",
        "group_ml_id",
        "identifier",
        "is_description_approved",
        "is_main_title",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "event_number_of_main_title_per_event_and_language_review_view": (
        "event_ml_id",
        "is_main_title",
        "title"
    ),
    "group_number_of_main_title_per_group_and_language_review_view": (
        "group_ml_id",
        "is_main_title",
        "title"
    ),
    "platform_invitation_view": (
        "action_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "from_profile_id",
        "identifier",
        "is_test_data",
        "number",
        "platform_invitation_id",
        "process_id",
        "queue_status_id",
        "start_timestamp",
        "to_contact_id",
        "to_email_address",
        "to_email_address_id",
        "to_phone_id",
        "to_phone_number",
        "to_profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt"
    ),
    "group_table": (
        "coordinate",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_profile_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "group_category_id",
        "group_id",
        "hashtag",
        "identifier",
        "is_activity",
        "is_approved",
        "is_building",
        "is_campaign",
        "is_city",
        "is_community",
        "is_continent",
        "is_country",
        "is_county",
        "is_event",
        "is_first_name",
        "is_geo",
        "is_group_approved",
        "is_interest",
        "is_job_title",
        "is_language",
        "is_last_name",
        "is_marital_status",
        "is_neighbourhood",
        "is_official",
        "is_organization",
        "is_reddit_group",
        "is_region",
        "is_relationship",
        "is_religion",
        "is_role",
        "is_skill",
        "is_sport",
        "is_state",
        "is_street",
        "is_test_data",
        "is_zip_code",
        "location_id",
        "location_list_id",
        "main_group_type_id",
        "members_visibility_id",
        "name",
        "new_group_id",
        "non_members_visibility_id",
        "number",
        "parent_group_id",
        "parent_group_list_id",
        "profile_id",
        "profile_list_id",
        "sequence",
        "start_timestamp",
        "system_group_name",
        "system_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "group_group_type_general_view": (
        "group.name",
        "group_group_type.end_timestamp",
        "group_group_type.is_test_data",
        "group_group_type_id",
        "group_id",
        "group_type.name",
        "group_type_id",
        "title"
    ),
    "event_profile_view": (
        "coupon_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "event_price_id",
        "event_profile_id",
        "event_profile_type",
        "from_profile_id",
        "identifier",
        "is_admin",
        "is_approved_by_admin",
        "is_can_be_paid_at_the_entrance_in_cash",
        "is_charged",
        "is_checked_in",
        "is_checkin_checker",
        "is_gave_payment_method",
        "is_invited",
        "is_join_virtually",
        "is_not_interested",
        "is_tentative",
        "is_test_data",
        "is_user_filled_registration",
        "is_will_arrive",
        "is_will_not_arrive",
        "number",
        "old_event_id",
        "qrcode_storage_id",
        "rank",
        "start_timestamp",
        "to_invited_profile_id",
        "to_registered_profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_profile_general_view": (
        "event.name",
        "event_id",
        "event_profile_id",
        "old_event.name",
        "old_event_id",
        "to_registered_profile.name",
        "to_registered_profile.profile_id"
    ),
    "event_profile_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_profile_type_id",
        "group_type_id",
        "identifier",
        "is_hierarchy",
        "is_include_logo",
        "is_location",
        "is_marketing_partner",
        "is_organization",
        "is_prioritze",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_group_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "end_timestamp",
        "group_group_type_id",
        "group_id",
        "group_type_id",
        "identifier",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "event_price_view": (
        "coupon_id",
        "created_timestamp",
        "created_user_id",
        "currency",
        "end_timestamp",
        "event_id",
        "event_price_id",
        "event_price_identifier",
        "event_price_number",
        "is_test_data",
        "price",
        "price_usd",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "subscription_user_profile_action_limit_table": (
        "action_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "hard_limit_value",
        "limit_unit",
        "profile_id",
        "soft_limit_value",
        "start_timestamp",
        "subscription_id",
        "subscription_profile_action_limit_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event_price_table": (
        "coupon_id",
        "created_timestamp",
        "created_user_id",
        "currency",
        "end_timestamp",
        "event_id",
        "event_price_id",
        "event_price_identifier",
        "event_price_number",
        "is_test_data",
        "price",
        "price_usd",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "age_group_group_view": (
        "age_group_group_id",
        "age_group_id",
        "group_id",
        "sequence"
    ),
    "event_profile_table": (
        "coupon_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "event_price_id",
        "event_profile_id",
        "event_profile_type",
        "from_profile_id",
        "identifier",
        "is_admin",
        "is_approved_by_admin",
        "is_can_be_paid_at_the_entrance_in_cash",
        "is_charged",
        "is_checked_in",
        "is_checkin_checker",
        "is_gave_payment_method",
        "is_invited",
        "is_join_virtually",
        "is_not_interested",
        "is_promoter",
        "is_tentative",
        "is_test_data",
        "is_user_filled_registration",
        "is_will_arrive",
        "is_will_not_arrive",
        "number",
        "old_event_id",
        "qrcode_storage_id",
        "rank",
        "start_timestamp",
        "to_invited_profile_id",
        "to_registered_profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "payment_request_table": (
        "amount",
        "first_reminder_timestmap",
        "from_profile_id",
        "last_reminder_timestamp",
        "payment_category",
        "payment_due_timestamp",
        "payment_request_id",
        "payment_url",
        "remark",
        "to_profile_id"
    ),
    "message_payment_request_table": (
        "message_id",
        "message_payment_request_id",
        "payment_request_id"
    ),
    "invoice_payment_request_table": (
        "invoice_id",
        "invoice_payment_request_id",
        "payment_request_id"
    ),
    "actual_payment_table": (
        "actual_payment_id",
        "payment_method_id",
        "price_id",
        "timestamp"
    ),
    "actual_payment_payment_request_table": (
        "actual_payment_id",
        "actual_payment_payment_request_id",
        "payment_request_id",
        "price_id"
    ),
    "payment_requst_text_block_table": (
        "payment_request_id",
        "payment_requst_text_block_id",
        "text_block_id"
    ),
    "action_item_payment_request_table": (
        "action_item_id",
        "action_item_payment_request_id",
        "payment_request_id"
    ),
    "user_external_pii_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "password_clear_text",
        "password_encrypted",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "user_external_pii_general_view": (
        "main_profile_id",
        "password_clear_text",
        "system_id",
        "user_external_id",
        "username"
    ),
    "payment_terms_service_offer_table": (
        "payment_terms_id",
        "payment_terms_service_offer_id",
        "service_offer_id"
    ),
    "marketplace_payment_terms_table": (
        "down_payment_percentage",
        "downpayment",
        "marketplace_payment_terms_id"
    ),
    "group_profile_table": (
        "consumer_category_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "group_id",
        "group_profile_id",
        "group_profile_type_id",
        "identifier",
        "is_admin",
        "is_approved_by_group_admin",
        "is_moderator",
        "is_recommended",
        "is_request_by_the_user",
        "is_sure",
        "is_test_data",
        "months",
        "number",
        "participant_category_id",
        "profile_id",
        "rank",
        "relationship_type_id",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "supplier_category_id",
        "text_block_id",
        "transaction_fee_amount",
        "transaction_fee_percentage",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_transaction_table": (
        "description",
        "group_profile_id",
        "is_paid",
        "name",
        "price_id",
        "service_transaction_id"
    ),
    "group_seller_table": (
        "consumer_category_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "group_id",
        "group_profile_type_id",
        "group_seller_id",
        "identifier",
        "is_admin",
        "is_approved_by_group_admin",
        "is_moderator",
        "is_recommended",
        "is_request_by_the_user",
        "is_sure",
        "is_test_data",
        "months",
        "number",
        "participant_category_id",
        "rank",
        "relationship_type_id",
        "seller_category_id",
        "seller_id",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "text_block_id",
        "transaction_fee_amount",
        "transaction_fee_percentage",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "service_transaction_transaction_benfit_table": (
        "benfit_service_transaction_id",
        "name",
        "service_transaction_id",
        "transaction_benefit_id"
    ),
    "goods_provider_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "id",
        "profile_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "transaction_benefit_table": (
        "goods_offer_id",
        "name",
        "service_offer_id",
        "transaction_benefit_id"
    ),
    "service_offer_table": (
        "base_price",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience",
        "main_currency_id",
        "price_id",
        "reaction_id",
        "reaction_value",
        "service_id",
        "service_offer_id",
        "service_provider_id",
        "service_seller_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "price_set_table": (
        "is_all_currencies",
        "price_set_id"
    ),
    "currency_table": (
        "code",
        "created_timestamp",
        "created_user_id",
        "currency_id",
        "end_timestamp",
        "is_digital_currency",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "digital_currency_table": (
        "digital_currency_id",
        "name"
    ),
    "service_messaging_general_view": (
        "action.name",
        "action_id",
        "name",
        "price_per_unit",
        "pricing_url_temp",
        "profile_id",
        "seller_id",
        "service_messaging_id",
        "system.name",
        "system_id",
        "unit_name"
    ),
    "service_messaging_table": (
        "action_id",
        "description",
        "max_number_of_characters",
        "price_per_unit",
        "pricing_url_temp",
        "quantity",
        "seller_id",
        "service_messaging_id",
        "system_id",
        "to_country_id",
        "unit_name"
    ),
    "profile_user_external_general_view": (
        "password_clear_text",
        "profile_id",
        "system_id",
        "user_external_id"
    ),
    "group.group_group_category_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_category_id",
        "group_type_id",
        "is_hierarchy",
        "is_location",
        "is_organization",
        "is_prioritze",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_group.group_group_view": (
        "action_id",
        "group_group_id",
        "group_id1",
        "group_id2",
        "is_main",
        "is_move_everything_from_group1_to_group2",
        "relationship_type_id"
    ),
    "activity_group_location_occurrence_profile_view": (
        "activity_group_location_occurrence_profile_id",
        "activity_id",
        "activity_list_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "group_list_id",
        "location_id",
        "location_list_id",
        "occurrence_id",
        "occurrence_list_id",
        "profile_id",
        "profile_list_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "group_group_view": (
        "action_id",
        "group_group_id",
        "group_id1",
        "group_id2",
        "is_main",
        "is_move_everything_from_group1_to_group2",
        "relationship_type_id"
    ),
    "job_title_profile_general_view": (
        "job_title.name",
        "job_title_id",
        "profile.name",
        "profile_id"
    ),
    "profile_profile_general_view": (
        "job_title",
        "job_title_id",
        "job_title_ml.title",
        "party1_for_party2_title",
        "profile1.name",
        "profile2.name",
        "profile_id1",
        "profile_id2",
        "relationship_type.name",
        "relationship_type_id"
    ),
    "component_test_table": (
        "component_id",
        "component_test_id",
        "name",
        "test_type_id",
        "test_type_name_temp"
    ),
    "message_channel_view": (
        "chanel_type_id",
        "message_channel_id",
        "name",
        "version"
    ),
    "user_user_extenal_general_view": (
        "brand_id",
        "email_address",
        "last_login_timestamp",
        "person.birthday_date",
        "person.is_approved",
        "person.main_email_address",
        "person_id",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.name",
        "profile.phone.full_number_normalized",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "role_name",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.active_location_id",
        "user.created_timestamp",
        "user.first_name",
        "user.is_approved",
        "user.is_test_data",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_external.username",
        "user_id",
        "username"
    ),
    "api_type_user_external_table": (
        "api_type_id",
        "api_type_user_external_id",
        "environment_id",
        "is_test_data",
        "user_external_id"
    ),
    "campaign_table": (
        "campaign_id",
        "content_set_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "days_of_week",
        "default_sender_profile_id",
        "dialog_workflow_script_id",
        "end_hour",
        "end_timestamp",
        "is_dialog_workflow",
        "is_test_data",
        "marketplace_rfp_id",
        "max_audience",
        "max_exposure_per_day",
        "message_template_id_old",
        "minimal_days_between_messages_to_the_same_recipient",
        "name",
        "occurrence_id",
        "start_hour",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_event_table": (
        "criteria_event_id",
        "criteria_set_id",
        "event_id"
    ),
    "campaign_event_table": (
        "campaign_event_id",
        "campaign_id",
        "event_id"
    ),
    "group_group_table": (
        "action_id",
        "group_group_id",
        "group_id1",
        "group_id2",
        "is_main",
        "is_move_everything_from_group1_to_group2",
        "relationship_type_id"
    ),
    "campaign_criteria_set_table": (
        "campaign_criteria_set_id",
        "campaign_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "criteria_set_id",
        "end_timestamp",
        "is_test_data",
        "max_recipients_per_campaign_criteria_set",
        "name",
        "number_of_channels_per_recipient",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "api_price_view": (
        "api_price_id",
        "api_type_id",
        "is_test_data",
        "max_quantity",
        "min_quantity",
        "price_set_id",
        "quantity_unit",
        "user_external_id"
    ),
    "api_price_table": (
        "api_price_id",
        "api_type_id",
        "is_test_data",
        "max_quantity",
        "min_quantity",
        "price_set_id",
        "quantity_unit",
        "user_external_id"
    ),
    "profile_storage_table": (
        "is_ai_approved",
        "is_manual_approved",
        "profile_id",
        "profile_storage_id",
        "storage_id"
    ),
    "reaction_type_ml_table": (
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_test_data",
        "lang_code",
        "reaction_type_id",
        "reaction_type_ml_id",
        "start_timestamp",
        "title",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_file_star_subscription_view": (
        "action_id",
        "COUNT(subscription_id)",
        "file_type_id"
    ),
    "action_star_subscription_table": (
        "action_id",
        "action_star_subscription_id",
        "action_stars",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "file_type_id",
        "is_conditional",
        "is_test_data",
        "number",
        "start_timestamp",
        "subscription_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_general_view": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_criteria_set_id",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "channel_id",
        "class_parameters_json",
        "component_id",
        "compound_message",
        "compound_message_json",
        "compound_message_json_version",
        "conversation_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_message_id",
        "end_timestamp",
        "from_email",
        "from_profile.profile-name",
        "from_profile_id",
        "from_provider_id",
        "from_subsystem_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "identifier",
        "inforu_application_id",
        "inforu_channel",
        "inforu_customer_id",
        "inforu_customer_param",
        "inforu_incoming_json",
        "inforu_keyword",
        "inforu_mo_session_id",
        "inforu_network",
        "inforu_project_id",
        "inforu_short_code",
        "inforu_type",
        "inforu_value",
        "is_gen_ai",
        "is_moderator_approved",
        "is_pin",
        "is_push_notification",
        "is_require_moderator",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_id",
        "message_message_relationship_id",
        "message_status_id",
        "message_template_id",
        "message_type_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "next_dialog_workflow_state_id",
        "number",
        "occurrence_id",
        "original_message_id",
        "parent_message_id",
        "process_id",
        "profile_id",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_message",
        "returned_value",
        "scheduled_sent_timestamp",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_profile.profile-name",
        "to_profile_id",
        "to_profile_list_id",
        "to_provider_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "visibility_id"
    ),
    "document_type_table": (
        "document_type_id",
        "name"
    ),
    "document_type_ml_table": (
        "document_type_id",
        "document_type_ml_id",
        "lang_code",
        "title"
    ),
    "brand_document_type_table": (
        "brand_document_type_table",
        "brand_id",
        "document_type_id",
        "is_enable",
        "next_document_number"
    ),
    "review_relationship_type_general_duplicate_description_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "gender1",
        "gender2",
        "is_direct_connection",
        "is_directional",
        "is_user_level",
        "Name_exp_11",
        "party1_gender_id",
        "party1_older_max_years",
        "party1_older_min_years",
        "party1_profile_category_id",
        "party1_profile_category_title",
        "party2_gender_id",
        "party2_profile_category_id",
        "party2_profile_category_title",
        "past_present_future",
        "profile1_older_than_profile2",
        "relationship_category_id",
        "relationship_type_classification_id",
        "relationship_type_classification_title",
        "relationship_type_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "relationship_type_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "description",
        "end_timestamp",
        "is_description_approved",
        "is_party1_for_party2_title_approved",
        "is_party2_for_party1_title_approved",
        "is_person_name_in_the_end",
        "lang_code",
        "party1_for_party2_title",
        "party2_for_party1_title",
        "relationship_type_id",
        "relationship_type_ml_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_inforu_table": (
        "inforu_application_id",
        "inforu_channel",
        "inforu_customer_id",
        "inforu_customer_param",
        "inforu_incoming_json",
        "inforu_keyword",
        "inforu_mo_session_id",
        "inforu_network",
        "inforu_project_id",
        "inforu_short_code",
        "inforu_type",
        "inforu_value",
        "message_id",
        "inforu_application_id",
        "inforu_channel",
        "inforu_customer_id",
        "inforu_customer_param",
        "inforu_incoming_json",
        "inforu_keyword",
        "inforu_mo_session_id",
        "inforu_network",
        "inforu_project_id",
        "inforu_short_code",
        "inforu_type",
        "inforu_value",
        "is_test_data",
        "message_id"
    ),
    "brand_environment_variable_table": (
        "brand_environment_variable_id",
        "brand_id",
        "environment_variable_name",
        "environment_variable_value"
    ),
    "agreement_table": (
        "agreement_id",
        "buyer_id",
        "seller_id",
        "agreement_id"
    ),
    "goods_offer_group_table": (
        "goods_offer_group_id",
        "goods_offer_id",
        "group_id",
        "platform_transaction_fee_from_buyer",
        "platform_transaction_fee_from_seller",
        "platform_transaction_percentage_from_buyer",
        "platform_transaction_percentage_from_seller"
    ),
    "goods_group_supplier_table": (
        "goods_group_supplier_id",
        "group_id",
        "platform_transaction_fee_value_from_buyer",
        "platform_transaction_fee_value_from_seller",
        "platform_transaction_percentage_from_buyer",
        "platform_transaction_percentage_from_seller",
        "supplier_id"
    ),
    "event_profile__label_table": (
        "event_profile__label_id",
        "event_profile_id",
        "label_id"
    ),
    "goods_offer_profile_table": (
        "end_timestamp",
        "goods_offer__profile_id",
        "goods_offer_id",
        "income_transaction_id",
        "price_duration",
        "price_id",
        "profile_id",
        "start_timestamp"
    ),
    "goods_offer_table": (
        "advertisement_campaign_id",
        "base_price",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "experience",
        "exposures",
        "goods_id",
        "goods_offer_id",
        "goods_provider_id",
        "goods_seller_id",
        "image_height",
        "image_width",
        "is_hot_deal",
        "main_currency_id",
        "main_price_id",
        "period",
        "reaction_id",
        "reaction_value",
        "start_timestamp",
        "units",
        "units_of_what",
        "updated_timestamp",
        "updated_user_id"
    ),
    "intent_table": (
        "intent_id",
        "is_event_registration",
        "is_natural",
        "is_negative",
        "is_positive",
        "text"
    ),
    "brand_profile_table": (
        "brand_id",
        "brand_profile_id",
        "is_partner",
        "profile_id"
    ),
    "brand_partner_profile_table": (
        "brand_id",
        "brand_profile_id",
        "percentage",
        "profile_id"
    ),
    "goods_offer__sku_table": (
        "description",
        "goods_offer__sku_id",
        "goods_offer_id",
        "name",
        "sku_id"
    ),
    "invoice__line_item_table": (
        "invoice__line_item_id",
        "invoice_id",
        "line_item_id",
        "name"
    ),
    "line_item__proforma_invoice_table": (
        "line_item__proforma_invoice_id"
    ),
    "line_item_ml_table": (
        "lang_code",
        "line_item_id",
        "line_item_ml_id",
        "title"
    ),
    "sku_table": (
        "name",
        "sku_id"
    ),
    "proforma_invoice_table": (
        "from_profile_id",
        "name",
        "number",
        "proforma_invoice_id",
        "to_profile_id"
    ),
    "service_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "name",
        "service_id",
        "service_type_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "event__food_needed_table": (
        "event__food_needed_id",
        "event_id",
        "quantity_per_event",
        "quantity_per_participant",
        "sku",
        "unit"
    ),
    "event__bring_your_own_food_table": (
        "event__bring_your_own_food_id",
        "event_id",
        "from_profile_id",
        "quantity",
        "sku",
        "unit",
        "your_own_food_id"
    ),
    "people_criteria_ext_view": (
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "gender_list_id",
        "group_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "max_age",
        "max_height",
        "max_kids_age",
        "max_number_of_kids",
        "min_age",
        "min_height",
        "min_kids_age",
        "min_number_of_kids",
        "name",
        "number_of_partners",
        "organization_name",
        "partner_experience_level",
        "profile_list_id",
        "start_date_type_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "workplace_profile_list_id"
    ),
    "message_template_category_table": (
        "description",
        "facebook_conversation_category",
        "message_template_category_id",
        "name"
    ),
    "message_template_text_block_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "criteria_set_id",
        "end_timestamp",
        "is_test_data",
        "is_version_locked",
        "is_visible",
        "message_template_category_id",
        "message_template_text_block_id",
        "message_template_text_block_type_id",
        "name",
        "question_id",
        "seq_old",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "version"
    ),
    "logger_general_recent_whatsapp_inforu_view": (
        "component_id",
        "logger_id",
        "name",
        "programming_language",
        "programming_language_id",
        "server_ip_v4",
        "server_ip_v6"
    ),
    "purchase_order_table": (
        "from_profile_id",
        "name",
        "number",
        "purchase_order_id",
        "to_profile_id"
    ),
    "rfp_item_proposal_table": (
        "max_price_id",
        "min_price_id",
        "prospect_profile_id",
        "rfp_item_id",
        "rfp_item_proposal_id",
        "supplier_profile_id"
    ),
    "action_item_table": (
        "action_id",
        "action_item_id",
        "action_item_type_id",
        "active_main_profile_id",
        "activity_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "due_date_bonus1",
        "due_date_regular",
        "end_timestamp",
        "financial_transaction_id",
        "github_repo_url",
        "identifier",
        "is_recommendation",
        "is_test_data",
        "jira_issue_id",
        "jira_issue_url",
        "name",
        "number",
        "other_group_id",
        "other_person_id",
        "other_profile_id",
        "passive_main_contact_id",
        "passive_main_profile_id",
        "price_id_bonus1",
        "price_id_reguar",
        "ranking",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "line_item_table": (
        "action_item_id",
        "description_temp",
        "event_id",
        "from_time",
        "hours",
        "line_item_id",
        "name",
        "project_id",
        "to_time"
    ),
    "service_receiver_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "service_provider_ml_view": (
        "entity_id",
        "name"
    ),
    "service_review_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "service_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "service_offer_ml_view": (
        "entity_id",
        "name"
    ),
    "shop_ml_view": (
        "coordinate",
        "entity_id",
        "name"
    ),
    "platform_invitation_table": (
        "action_id",
        "bcc_profile_id",
        "cc_profile_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "from_profile_id",
        "identifier",
        "invitation_method_id",
        "invitation_status_id",
        "is_test_data",
        "number",
        "platform_invitation_id",
        "process_id",
        "queue_status_id",
        "start_timestamp",
        "to_contact_id",
        "to_email_address",
        "to_email_address_id",
        "to_phone_id",
        "to_phone_number",
        "to_profile_id",
        "token",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt"
    ),
    "message_template__system_table": (
        "external_template_id",
        "message_template__system_id",
        "message_template_id",
        "sub_system_id",
        "system_id"
    ),
    "api_incoming_request_table": (
        "api_incoming_request_id",
        "body",
        "endpoint",
        "full_json",
        "headers",
        "httpMethod",
        "multiValueHeaders",
        "multiValueQueryStringParameters",
        "path",
        "pathParameters",
        "queryStringParameters",
        "requestContext",
        "resource",
        "server",
        "stageVariables"
    ),
    "conversation_view": (
        "conversation_id",
        "name"
    ),
    "profile__phone_general_view": (
        "full_number_normalized",
        "phone.phone_id",
        "phone_profile_id",
        "profile.name",
        "profile_id"
    ),
    "profile_list_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "name",
        "profile_list_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_recent_general_view": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_criteria_set_id",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "channel_id",
        "class_parameters_json",
        "compound_message",
        "compound_message_json",
        "compound_message_json_version",
        "conversation_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_message_id",
        "end_timestamp",
        "from_email",
        "from_profile.profile-name",
        "from_profile_id",
        "from_provider_id",
        "from_subsystem_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "identifier",
        "inforu_application_id",
        "inforu_channel",
        "inforu_customer_id",
        "inforu_customer_param",
        "inforu_incoming_json",
        "inforu_keyword",
        "inforu_mo_session_id",
        "inforu_network",
        "inforu_project_id",
        "inforu_short_code",
        "inforu_type",
        "inforu_value",
        "is_gen_ai",
        "is_moderator_approved",
        "is_pin",
        "is_push_notification",
        "is_require_moderator",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_id",
        "message_message_relationship_id",
        "message_status_id",
        "message_template_id",
        "message_type_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "next_dialog_workflow_state_id",
        "number",
        "occurrence_id",
        "original_message_id",
        "parent_message_id",
        "process_id",
        "profile_id",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_message",
        "returned_value",
        "scheduled_sent_timestamp",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_profile.profile-name",
        "to_profile_id",
        "to_profile_list_id",
        "to_provider_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "visibility_id"
    ),
    "conversation_table": (
        "conversation_id",
        "name",
        "conversation_id",
        "name",
        "profile_list_id"
    ),
    "conversation__profile_table": (
        "conversation__profile_id",
        "conversation_id",
        "end_timestamp",
        "is_get_popup",
        "is_mute",
        "last_dialog_workflow_state_id",
        "last_message_channel_id",
        "message_channel_id",
        "profile_id",
        "start_timestamp"
    ),
    "campaign_message_profile_table": (
        "campaign_id",
        "campaign_message_profile_id",
        "channel_id_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "last_dialog_workflow_state_id",
        "message_id",
        "number",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "calendar_url_table": (
        "calendar_url_id",
        "is_in_person",
        "length_minutes",
        "name",
        "platform_id",
        "profile_id",
        "profile_name_temp",
        "schedule_url_id",
        "schedule_url_temp"
    ),
    "marketplace_rfp_table": (
        "activity_id",
        "activity_list_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "goods_id",
        "identifier",
        "is_test_data",
        "is_to_prospect",
        "is_to_supplier",
        "location_id",
        "location_list_id",
        "marketplace_rfp_id",
        "name",
        "number",
        "service_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_id",
        "url_temp"
    ),
    "file_type_table": (
        "add_action_id",
        "file_type",
        "file_type_id",
        "is_identity_document",
        "is_system",
        "location_id",
        "location_list_id",
        "remove_action_id",
        "visibility_id"
    ),
    "profile_list_member_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_list_id",
        "profile_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "people_criteria_ext_table": (
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "gender_list_id",
        "group_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "max_age",
        "max_height",
        "max_kids_age",
        "max_number_of_kids",
        "min_age",
        "min_height",
        "min_kids_age",
        "min_number_of_kids",
        "name",
        "number_of_partners",
        "organization_name",
        "partner_experience_level",
        "profile_id",
        "profile_list_id",
        "start_date_type_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "workplace_profile_list_id"
    ),
    "people_criteria_ext_general_view": (
        "city_id",
        "coordinate",
        "country_id_old",
        "created_timestamp",
        "created_user_id",
        "criteria.criteria_id",
        "criteria.name",
        "end_timestamp",
        "entity_type_id",
        "gender_list_id",
        "group_id",
        "group_id_old",
        "group_list_id",
        "international_code",
        "international_code_old",
        "internet_domain_id",
        "internet_domain_id_old",
        "internet_domain_list_id",
        "internet_domain_list_id_old",
        "is_test_data",
        "job_group_list_id",
        "job_group_list_id_old",
        "job_location_list_id",
        "job_location_list_id_old",
        "job_types_id",
        "job_types_id_old",
        "label_id",
        "location_criteria.criteria_id",
        "location_id",
        "location_list_id",
        "max_age",
        "max_height",
        "max_kids_age",
        "max_number_of_kids",
        "min_age",
        "min_height",
        "min_kids_age",
        "min_number_of_kids",
        "number_of_partners",
        "organization_name",
        "organization_name_old",
        "partner_experience_level",
        "people_criteria.created_timestamp",
        "people_criteria.created_user_id",
        "people_criteria.criteria_id",
        "people_criteria.end_timestamp",
        "people_criteria.is_test_data",
        "people_criteria.name",
        "people_criteria.start_timestamp",
        "people_criteria.updated_timestamp",
        "people_criteria.updated_user_id",
        "people_criteria.user_external_id",
        "profile_list_id",
        "profile_list_id_old",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_date_type_id_old",
        "start_timestamp",
        "template_condition_2delete",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "vacancy_list_id_old",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id",
        "workplace_profile_list_id_old"
    ),
    "people_criteria_ext_review_view": (
        "criteria.name",
        "criteria_id",
        "max_age",
        "min_age",
        "people_criteria.name"
    ),
    "relationship_type_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "is_direct_connection",
        "is_directional",
        "is_people_group",
        "is_test_data",
        "is_user_level",
        "name",
        "party1_gender_id",
        "party1_older_max_years",
        "party1_older_min_years",
        "party1_profile_category_id",
        "party2_gender_id",
        "party2_goods_id",
        "party2_goods_offer_id",
        "party2_profile_category_id",
        "pary2_sku_id",
        "past_present_future",
        "profile1_older_than_profile2",
        "relationship_category_id",
        "relationship_type_classification_id",
        "relationship_type_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "techer_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "full_name",
        "is_test_data",
        "start_timestamp",
        "teacher_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "rfp_table": (
        "from_profile_id",
        "name",
        "rfp_id"
    ),
    "rfp_item_table": (
        "acitivity_id",
        "description",
        "mandatory_quantity",
        "max_price_id",
        "min_price_id",
        "name",
        "optional_quantity",
        "price_per",
        "rfp_id",
        "rfp_item_id",
        "sub_activity_id"
    ),
    "price_set_currency_table": (
        "country_id",
        "currency_id",
        "end_timestamp",
        "is_vat_included",
        "price",
        "price_set_currency_id",
        "price_set_id",
        "start_timestamp"
    ),
    "student_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "father_profile_id",
        "full_name",
        "is_test_data",
        "mother_profile_id",
        "profile_id",
        "start_timestamp",
        "student_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "teacher_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "full_name",
        "is_test_data",
        "profile_id",
        "start_timestamp",
        "teacher_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "brand_table": (
        "brand_id",
        "brand_profile_id",
        "brand_type_id",
        "competitors",
        "description",
        "google_ads_account",
        "name",
        "potential_partners",
        "root_group_id"
    ),
    "event_general_view": (
        "event.description",
        "event.end_timestamp",
        "event.name",
        "event.start_timestamp",
        "event_category_title",
        "event_id",
        "event_id1 (parent)",
        "event_id2 (child)",
        "event_ml.end_timestamp",
        "event_ml.is_test_data",
        "event_ml_id",
        "event_type_name",
        "facebook_event_url",
        "is_contact_classification",
        "is_location",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "location_id",
        "meetup_event_url",
        "new_event_id",
        "organizers_profile_id",
        "registration_url",
        "title",
        "website_url"
    ),
    "event_general_with_deleted_and_test_data_view": (
        "event.description",
        "event.end_timestamp",
        "event.name",
        "event.start_timestamp",
        "event_category_title",
        "event_id",
        "event_id1 (parent)",
        "event_id2 (child)",
        "event_ml.end_timestamp",
        "event_ml.is_test_data",
        "event_ml_id",
        "event_type_name",
        "facebook_event_url",
        "is_contact_classification",
        "is_location",
        "is_main_title",
        "is_test_data",
        "lang_code",
        "location.name",
        "location_id",
        "location_ml.lang_code",
        "location_ml.title",
        "meetup_event_url",
        "new_event_id",
        "organizers_profile_id",
        "registration_url",
        "title",
        "website_url"
    ),
    "profile_en_view_old": (
        "experience_years_min",
        "id_profile",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "main_person_id",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_type_id",
        "visibility_id"
    ),
    "profile_ml_approved_en_view_old": (
        "experience_years_min",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_id",
        "profile_type_id",
        "visibility_id"
    ),
    "profile_ml_approved_view_old": (
        "experience_years_min",
        "is_approved",
        "is_main",
        "is_rip",
        "last_dialog_workflow_state_id_old",
        "number",
        "preferred_lang_code",
        "profile.main_user_id",
        "profile_id",
        "profile_type_id",
        "visibility_id"
    ),
    "user_table": (
        "active_location_id",
        "birthday_moved_to_person",
        "birthday_original_moved_to_person",
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "current_time_zone",
        "default_time_zone",
        "end_timestamp",
        "first_name",
        "gender_id",
        "identity_id",
        "interested_in_gender_id",
        "is_approved",
        "is_system",
        "is_test_data",
        "last_name",
        "main_profile_id",
        "name_approved",
        "number",
        "password",
        "preferred_lang_code",
        "presence",
        "region",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user.main_email_address",
        "user_approval_to_get_emails_from_us",
        "user_id",
        "username"
    ),
    "user_pii_view": (
        "active_location_id",
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "current_time_zone",
        "default_time_zone",
        "end_timestamp",
        "first_name",
        "gender_id",
        "identity_id",
        "interested_in_gender_id",
        "is_approved",
        "is_test_data",
        "last_name",
        "name_approved",
        "number",
        "password",
        "password_clear_text",
        "password_encrypted",
        "preferred_lang_code",
        "presence",
        "region",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user.main_email_address",
        "user_approval_to_get_emails_from_us",
        "user_id",
        "username"
    ),
    "user_view": (
        "active_location_id",
        "birthday_moved_to_person",
        "birthday_original_moved_to_person",
        "brand_id",
        "created_timestamp",
        "created_user_id",
        "current_time_zone",
        "default_time_zone",
        "end_timestamp",
        "first_name",
        "gender_id",
        "identity_id",
        "interested_in_gender_id",
        "is_approved",
        "is_system",
        "is_test_data",
        "last_name",
        "main_profile_id",
        "name_approved",
        "number",
        "password",
        "preferred_lang_code",
        "presence",
        "region",
        "stars",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "user.main_email_address",
        "user_approval_to_get_emails_from_us",
        "user_id",
        "username"
    ),
    "conversation__profile_view": (
        "conversation__profile_id",
        "conversation_id",
        "end_timestamp",
        "is_get_popup",
        "is_mute",
        "last_message_channel_id",
        "profile_id",
        "start_timestamp"
    ),
    "conversation__profile_general_view": (
        "conversation.conversation_id",
        "conversation__profile_id"
    ),
    "login_history_general_view": (
        "brand_id",
        "login_history.created_timestamp",
        "person.is_approved",
        "person_id",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.created_timestamp",
        "user.first_name",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_id"
    ),
    "login_history_general_recent_view": (
        "brand_id",
        "login_history.created_timestamp",
        "person.is_approved",
        "person_id",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.created_timestamp",
        "user.first_name",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_id"
    ),
    "login_login_history_group_by_user_view": (
        "brand_id",
        "login_history.created_timestamp",
        "person.is_approved",
        "person_id",
        "profile.preferred_lang_code",
        "profile.stars",
        "profile_id",
        "subscription_id",
        "subscription_title",
        "user.active_location.country_ml.title",
        "user.created_timestamp",
        "user.first_name",
        "user.last_name",
        "user.main_email_address",
        "user.stars",
        "user_id"
    ),
    "profile_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_circa",
        "end_date_day",
        "end_date_month",
        "end_date_year",
        "end_timestamp",
        "is_potential",
        "is_profile1_knows_profile2",
        "is_profile1_requested_to_be_friend_with_profile2",
        "is_profile2_approved_profile1_request_to_be_friends",
        "is_star",
        "is_system_recommendation",
        "is_test_data",
        "is_user_request",
        "job_title",
        "job_title_id",
        "number",
        "profile_id1",
        "profile_id2",
        "profile_profile_id",
        "profile_profile_role_id",
        "relationship_type_id",
        "remark",
        "start_circa",
        "start_date_day",
        "start_date_month",
        "start_date_year",
        "start_timestamp",
        "system_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "user_external_level"
    ),
    "reaction_type_general_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "image",
        "is_test_data",
        "name",
        "number",
        "reaction_id",
        "reaction_type_id",
        "reaction_type_ml.end_timestamp",
        "reaction_type_ml.is_test_data",
        "reaction_type_ml.lang_code",
        "start_timestamp",
        "text",
        "title",
        "updated_timestamp",
        "updated_user_id",
        "value"
    ),
    "reaction_type_table": (
        "end_timestamp",
        "is_test_data",
        "name",
        "reaction_type_id"
    ),
    "label_storage_table": (
        "is_approved",
        "label_id",
        "label_ml_id",
        "label_storage_id",
        "storage_id",
        "storage_ml_id"
    ),
    "label_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_approved",
        "is_highlight",
        "is_test_data",
        "label_id",
        "location_id",
        "name",
        "name_approved",
        "number",
        "parent_label_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "project_table": (
        "description_old",
        "name",
        "phase",
        "project_id"
    ),
    "profile__project_table": (
        "profile__project_id",
        "profile_id",
        "project_id"
    ),
    "campaign_user_request_table": (
        "campaign_id",
        "campaign_user_request_id",
        "user_request_id"
    ),
    "user_request_table": (
        "request_en",
        "user_request_id",
        "user_request_tablecol"
    ),
    "profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "experience_years_min",
        "identifier",
        "internal_description",
        "is_approved",
        "is_business_profile",
        "is_claimed",
        "is_edu",
        "is_employed",
        "is_looking_for_job",
        "is_main",
        "is_rip",
        "is_system",
        "is_test_data",
        "last_dialog_workflow_state_id_old",
        "main_location_id",
        "main_person_id",
        "main_phone_id",
        "number",
        "preferred_lang_code",
        "profile.gender_id",
        "profile.main_email_address",
        "profile.main_user_id",
        "profile.name",
        "profile_id",
        "profile_type_id",
        "stars",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "message_reaction_table": (
        "created_user_id",
        "message_id",
        "message_reaction_id",
        "profile_id",
        "reaction_id"
    ),
    "profile_reaction_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "from_profile_id",
        "identifier",
        "is_test_data",
        "number",
        "profile_reaction_id",
        "reaction_id",
        "start_timestamp",
        "to_profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_person_table": (
        "birthday_date",
        "birthday_original",
        "birthday_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "day",
        "end_timestamp",
        "father_name",
        "first_name",
        "gender_id",
        "identifier",
        "is_approved",
        "is_first_name_approved",
        "is_identity_confirmed",
        "is_last_name_approved",
        "is_last_name_auto_generated",
        "is_last_name_sure",
        "is_nickname_approved",
        "is_rip",
        "is_system_data",
        "is_test_data",
        "last_coordinate",
        "last_location_id",
        "last_name",
        "last_name_original",
        "location_id",
        "main_full_number_normalized",
        "month",
        "name",
        "nickname",
        "number",
        "person.main_email_address",
        "person.main_email_address_id",
        "person.main_profile_id",
        "person.preferred_lang_code",
        "person_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "year",
        "year_cira"
    ),
    "test_mysql_general_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group",
        "identifier",
        "is_test_data",
        "lang_code",
        "name",
        "number",
        "point",
        "polygon",
        "start_timestamp",
        "test_mysql_id",
        "title",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_mysql_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "point",
        "polygon",
        "start_timestamp",
        "test_mysql.main_email_address",
        "test_mysql_id",
        "unique_int_column",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_with_deleted_and_test_data_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_location_mapping_with_deleted_and_test_data_view": (
        "is_test_data",
        "test_county_id",
        "test_county_location_mapping_id",
        "test_location_id"
    ),
    "test_county_location_mapping_table": (
        "is_test_data",
        "test_county_id",
        "test_county_location_mapping_id",
        "test_location_id"
    ),
    "test_county_table": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "user_external_latest_token_general_view": (
        "access_token",
        "base_credential_token",
        "cloud_project__user_external.description",
        "cloud_project__user_external.name",
        "cloud_project_user_external__access_token.end_timestamp",
        "cloud_project_user_external__access_token.is_test_data",
        "end_timestamp",
        "expiry",
        "is_refresh_token_valid",
        "oauth_state",
        "profile_id",
        "profile_user_external.is_test_data",
        "profile_user_external_id",
        "refresh_token",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "user_external.is_test_data",
        "user_external_id",
        "username"
    ),
    "profile_user_external_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "profile_id",
        "profile_user_external_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id"
    ),
    "person_table": (
        "birthday_date",
        "birthday_original",
        "birthday_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "day",
        "end_timestamp",
        "father_name",
        "first_name",
        "gender_id",
        "identifier",
        "is_approved",
        "is_first_name_approved",
        "is_identity_confirmed",
        "is_last_name_approved",
        "is_last_name_auto_generated",
        "is_last_name_sure",
        "is_nickname_approved",
        "is_rip",
        "is_system_data",
        "is_test_data",
        "last_coordinate",
        "last_location_id",
        "last_name",
        "last_name_original",
        "location_id",
        "main_full_number_normalized",
        "month",
        "name",
        "nickname",
        "number",
        "person.main_email_address",
        "person.main_email_address_id",
        "person.main_profile_id",
        "person.preferred_lang_code",
        "person_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "year",
        "year_cira"
    ),
    "test_gender_view": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "test_gender_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_gender_table": (
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "name",
        "number",
        "start_timestamp",
        "test_gender_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "environment_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "environment_id",
        "github_branch",
        "name",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "user_external_pii_table": (
        "access_token_old",
        "base_credentials_token",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "password_clear_text",
        "password_encrypted",
        "refresh_token",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "user_external_pii_id"
    ),
    "campaign_general_with_deleted_and_test_data_view": (
        "campaign_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "days_of_week",
        "default_sender_profile_id",
        "dialog_workflow_script_id",
        "end_hour",
        "end_timestamp",
        "is_dialog_workflow",
        "is_test_data",
        "max_audience",
        "max_exposure_per_day",
        "message_template_id_old",
        "minimal_days_between_messages_to_the_same_recipient",
        "name",
        "occurrence_id",
        "start_hour",
        "start_state_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_general_view": (
        "campaign_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id_old",
        "days_of_week",
        "default_sender_profile_id",
        "dialog_workflow_script_id",
        "end_hour",
        "end_timestamp",
        "is_dialog_workflow",
        "is_test_data",
        "max_audience",
        "max_exposure_per_day",
        "message_template_id_old",
        "minimal_days_between_messages_to_the_same_recipient",
        "name",
        "occurrence_id",
        "start_hour",
        "start_state_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_criteria_set_criteria_general_view": (
        "campaign.name",
        "campaign_criteria_set_id",
        "campaign_id",
        "child_criteria_set_id",
        "child_name",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "criteria.is_test_data",
        "criteria_id",
        "criteria_id_old",
        "criteria_set.name",
        "criteria_set_id",
        "end_timestamp",
        "entity_type_id",
        "entity_type_name",
        "group_id",
        "group_list_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "location_id",
        "location_list_id",
        "name",
        "organization_name",
        "parent_criteria_set_id",
        "parent_name",
        "profile_list_id",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_timestamp",
        "template_condition",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id"
    ),
    "criteria_set_profile_table": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_set_id",
        "criteria_set_profile_id",
        "identifier",
        "is_test_data",
        "name",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "phone_profile_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_main",
        "is_test_data",
        "is_verified",
        "label_phone_id",
        "name",
        "phone_id",
        "phone_profile_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "criteria_profile_table": (
        "batch_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "criteria_id",
        "criteria_profile_id",
        "description",
        "identifier",
        "is_approved",
        "is_test_data",
        "name",
        "number",
        "profile_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "visibility_id"
    ),
    "campaign_criteria_set_criteria_profile_general_view": (
        "campaign_id",
        "criteria_id",
        "criteria_profile.profile_id",
        "criteria_set_id",
        "criteria_set_profile.profile_id"
    ),
    "profile_list_member_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "profile_id",
        "profile_list_id",
        "profile_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "message_table": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_criteria_set_id",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "channel_id",
        "class_parameters_json",
        "component_id",
        "compound_message",
        "compound_message_json",
        "compound_message_json_version",
        "conversation_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "email_message_id",
        "end_timestamp",
        "from_api_type_id",
        "from_email",
        "from_phone_number",
        "from_profile_id",
        "from_provider_id",
        "from_subsystem_id",
        "from_system_id",
        "from_user_external_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "identifier",
        "inforu_application_id_old",
        "inforu_channel_old",
        "inforu_customer_id_old",
        "inforu_customer_param_old",
        "inforu_incoming_json_old",
        "inforu_keyword_old",
        "inforu_mo_session_id_old",
        "inforu_network_old",
        "inforu_project_id_old",
        "inforu_short_code_old",
        "inforu_type_old",
        "inforu_value_old",
        "is_dialog_workflow",
        "is_gen_ai",
        "is_moderator_approved",
        "is_pin",
        "is_push_notification",
        "is_require_moderator",
        "is_test_data",
        "is_user_request",
        "last_dialog_workflow_state_id",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_id",
        "message_message_relationship_id",
        "message_status_id",
        "message_template_id",
        "message_type_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "next_dialog_workflow_state_id",
        "number",
        "occurrence_id",
        "original_message_id",
        "parent_message_id",
        "process_id",
        "profile_id",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_message",
        "returned_value",
        "scheduled_sent_timestamp",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_phone_number",
        "to_profile_id",
        "to_profile_list_id",
        "to_provider_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "user_request_price_id",
        "visibility_id"
    ),
    "field_alias_table": (
        "field_alias_id",
        "field_id",
        "field_name",
        "profile_id",
        "subsystem_id",
        "system_id",
        "visibility_id"
    ),
    "data_source_type_table": (
        "created_timestamp",
        "created_user_id",
        "data_source_category_id",
        "data_source_type_id",
        "end_timestamp",
        "is_test_data",
        "name",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "folder_table": (
        "folder_id",
        "google_drive_folder_id",
        "name",
        "parent_folder_id"
    ),
    "sql2code_file_table": (
        "filename",
        "folder",
        "folder_id",
        "programming_language_id",
        "repo",
        "sql2code_id",
        "sql2code_repo_id",
        "structure_name"
    ),
    "folder__storage_table": (
        "folder_id",
        "storage_id"
    ),
    "logger_debuging_view": (
        "api_type",
        "component_category",
        "component_name",
        "computer_language",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "developer_email",
        "developer_email_address",
        "function_name",
        "logger_id",
        "message",
        "record",
        "severity_id",
        "timestamp"
    ),
    "data_source_type__field_general_view": (
        "data_source_field_id",
        "data_source_type_id",
        "external_field_name",
        "field_id",
        "name",
        "position"
    ),
    "data_source_type__field_view": (
        "data_source_field_id",
        "data_source_type_id",
        "external_field_name",
        "field_id",
        "position"
    ),
    "test_county_play1_table": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_play1_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "continent_table": (
        "continent_id",
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "group_id",
        "identifier",
        "is_test_data",
        "iso",
        "iso3",
        "name",
        "nicename",
        "numcode",
        "phonecode",
        "updated_timestamp",
        "updated_user_id"
    ),
    "continent_ml_table": (
        "continent_ml_id",
        "country_id",
        "country_id_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_main",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "machine_learning_model_table": (
        "accuracy",
        "batch_size",
        "enabler",
        "epochs",
        "input_shape",
        "is_machine_learning",
        "is_sql",
        "learning_rate",
        "loss",
        "machine_learning_model_id",
        "model_data",
        "model_data_storage_id",
        "model_name",
        "model_path",
        "model_storage_id",
        "name",
        "network",
        "path_to_model",
        "provider",
        "source_code_id",
        "source_code_path_old",
        "sql_statement",
        "test_data",
        "train_data",
        "training_data_storage_id",
        "version",
        "machine_learning_model_id",
        "model_name",
        "name",
        "parameters_b",
        "quantization"
    ),
    "prompt_execution__storage_table": (
        "prompt_execution__storage_id",
        "prompt_id",
        "storage_id"
    ),
    "email_address_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "email_address",
        "email_address_approved",
        "email_address_id",
        "end_timestamp",
        "identifier",
        "internet_domain_id",
        "is_calendar",
        "is_shared_email_address",
        "is_system_data",
        "is_test_data",
        "is_verified",
        "name",
        "number",
        "profile_id_old",
        "start_timestamp",
        "text_block_id",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "username",
        "verification_expiration",
        "verification_pin",
        "visibility_id"
    ),
    "event_table": (
        "access_start_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "description",
        "end_timestamp",
        "event_id",
        "facebook_event_url",
        "genre",
        "identifier",
        "is_beakfast",
        "is_can_be_paid_at_the_entrance_in_cash",
        "is_dinner",
        "is_event_approved",
        "is_event_has_promoters",
        "is_family",
        "is_lunch",
        "is_paid",
        "is_require_end_user_arrival_confirmation",
        "is_require_end_user_registration",
        "is_require_organizer_registration_confirmation",
        "is_show_end_timestamp",
        "is_test_data",
        "is_waitinglist",
        "location_id",
        "location_id_temp",
        "meetup_event_url",
        "name",
        "new_event_id",
        "number",
        "organizers_profile_id",
        "parent_event_id",
        "registration_url",
        "segement",
        "start_timestamp",
        "sub_genre",
        "sub_type",
        "timezone",
        "type",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "website_url"
    ),
    "phone_table": (
        "area_code",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "end_timestamp",
        "extension",
        "full_number_normalized",
        "international_code",
        "is_test_data",
        "is_valid",
        "is_verified_old",
        "local_number_normalized",
        "number",
        "number_original",
        "phone_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "person_phone_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "number",
        "person_id",
        "person_phone_id",
        "phone_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "data_source_type__unique_index_table": (
        "data_source_type_id",
        "data_source_type_unique_index_id",
        "field_id",
        "unique_index_id"
    ),
    "token__user_external_old_view": (
        "access_token",
        "base_credential_token",
        "description",
        "end_timestamp",
        "expiry",
        "is_test_data",
        "name",
        "oauth_state",
        "project",
        "refresh_token_old",
        "scope_json",
        "token__user_external_old_id",
        "updated_timestamp",
        "user_external_id"
    ),
    "token__user_external_old_general_view": (
        "access_token",
        "base_credential_token",
        "description",
        "end_timestamp",
        "expiry",
        "is_test_data",
        "name",
        "oauth_state",
        "project",
        "refresh_token_old",
        "scope_json",
        "token__user_external_old_id",
        "updated_timestamp",
        "user_external_id"
    ),
    "user_external_table": (
        "access_token_old",
        "additional_fields_json",
        "base_credentials_token_old",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "credential_storage_id",
        "customer_id",
        "description",
        "email_address_id",
        "end_timestamp",
        "environment_id_old",
        "expiry_old",
        "handle",
        "identifier",
        "is_refresh_token_valid",
        "is_test_data",
        "main_profile_id",
        "number",
        "oauth_callback_confirmed",
        "oauth_state",
        "oauth_token",
        "oauth_token_secret",
        "person_url_id",
        "profile_url_id",
        "project_id",
        "refresh_token_old",
        "start_timestamp",
        "subsystem_id",
        "system_id",
        "token_2delete",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url_id",
        "user_external_id",
        "username"
    ),
    "test_county_dvlp1_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_dvlp1_table": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_play1_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_play1_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "test_county_dvlp1_view": (
        "coordinate",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "group_id",
        "is_test_data",
        "name",
        "new_test_county_id",
        "number",
        "test_country_id",
        "test_county_id",
        "test_state_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "test_county_dvlp1_ml_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "is_title_approved",
        "lang_code",
        "number",
        "start_timestamp",
        "test_county_id",
        "test_county_ml_id",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "cloud_project_table": (
        "cloud_project_id",
        "cloud_vendor_id",
        "cloud_vendor_name",
        "description",
        "name"
    ),
    "cloud_project__user_external_table": (
        "cloud_project",
        "cloud_project__user_external_id",
        "description",
        "end_timestamp",
        "is_refresh_token_valid",
        "name",
        "refresh_token",
        "scope_json",
        "user_external_id"
    ),
    "cloud_project__user_external_view": (
        "cloud_project",
        "cloud_project__user_external_id",
        "is_refresh_token_valid",
        "refresh_token",
        "scope_json",
        "user_external_id"
    ),
    "cloud_project_user_external__access_token_view": (
        "access_token",
        "base_credential_token",
        "cloud_project",
        "cloud_project__user_external_id",
        "cloud_project_user_external__access_token_id",
        "description",
        "end_timestamp",
        "expiry",
        "is_test_data",
        "name",
        "oauth_state",
        "refresh_token_old",
        "scope_json",
        "updated_timestamp"
    ),
    "cloud_project_user_external__access_token_table": (
        "access_token",
        "base_credential_token",
        "cloud_project",
        "cloud_project__user_external_id",
        "cloud_project_user_external__access_token_id",
        "description",
        "end_timestamp",
        "expiry",
        "is_test_data",
        "name",
        "oauth_state",
        "refresh_token_old",
        "scope_json",
        "updated_timestamp"
    ),
    "user_external_latest_token_general_old_view": (
        "access_token",
        "base_credential_token",
        "description",
        "end_timestamp",
        "expiry",
        "is_refresh_token_valid",
        "name",
        "oauth_state",
        "profile_id",
        "profile_user_external.is_test_data",
        "profile_user_external_id",
        "refresh_token",
        "subsystem_id",
        "system_id",
        "token__user_external.end_timestamp",
        "token__user_external.is_test_data",
        "updated_timestamp",
        "user_external.is_test_data",
        "user_external_id",
        "username"
    ),
    "prompt_prompt_table": (
        "entity_attributes_json",
        "parent_prompt_id",
        "prompt_id",
        "prompt_prompt_id"
    ),
    "json_schema_table": (
        "json_schema",
        "json_schema_id"
    ),
    "token__user_external_old_table": (
        "access_token",
        "base_credential_token",
        "description",
        "end_timestamp",
        "expiry",
        "is_test_data",
        "name",
        "oauth_state",
        "project",
        "refresh_token_old",
        "scope_json",
        "token__user_external_old_id",
        "updated_timestamp",
        "user_external_id"
    ),
    "agreement_version_table": (
        "agreement_id",
        "agreement_version_id",
        "version",
        "version_float"
    ),
    "agreement__user_table": (
        "agreement__user_id"
    ),
    "cond_instances": (
        "NAME",
        "OBJECT_INSTANCE_BEGIN"
    ),
    "error_log": (
        "DATA",
        "ERROR_CODE",
        "LOGGED",
        "PRIO",
        "SUBSYSTEM",
        "THREAD_ID"
    ),
    "events_waits_current": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "FLAGS",
        "INDEX_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_BYTES",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "OPERATION",
        "SOURCE",
        "SPINS",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT"
    ),
    "events_waits_history": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "FLAGS",
        "INDEX_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_BYTES",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "OPERATION",
        "SOURCE",
        "SPINS",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT"
    ),
    "events_waits_history_long": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "FLAGS",
        "INDEX_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_BYTES",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "OPERATION",
        "SOURCE",
        "SPINS",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT"
    ),
    "events_waits_summary_by_host_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT"
    ),
    "events_waits_summary_by_instance": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "OBJECT_INSTANCE_BEGIN",
        "SUM_TIMER_WAIT"
    ),
    "events_waits_summary_by_thread_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "THREAD_ID"
    ),
    "events_waits_summary_by_user_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_waits_summary_by_account_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_waits_summary_global_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT"
    ),
    "file_instances": (
        "EVENT_NAME",
        "FILE_NAME",
        "OPEN_COUNT"
    ),
    "file_summary_by_event_name": (
        "AVG_TIMER_MISC",
        "AVG_TIMER_READ",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_MISC",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_WRITE",
        "EVENT_NAME",
        "MAX_TIMER_MISC",
        "MAX_TIMER_READ",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_MISC",
        "MIN_TIMER_READ",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "SUM_NUMBER_OF_BYTES_READ",
        "SUM_NUMBER_OF_BYTES_WRITE",
        "SUM_TIMER_MISC",
        "SUM_TIMER_READ",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "file_summary_by_instance": (
        "AVG_TIMER_MISC",
        "AVG_TIMER_READ",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_MISC",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_WRITE",
        "EVENT_NAME",
        "FILE_NAME",
        "MAX_TIMER_MISC",
        "MAX_TIMER_READ",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_MISC",
        "MIN_TIMER_READ",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "OBJECT_INSTANCE_BEGIN",
        "SUM_NUMBER_OF_BYTES_READ",
        "SUM_NUMBER_OF_BYTES_WRITE",
        "SUM_TIMER_MISC",
        "SUM_TIMER_READ",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "host_cache": (
        "COUNT_ADDRINFO_PERMANENT_ERRORS",
        "COUNT_ADDRINFO_TRANSIENT_ERRORS",
        "COUNT_AUTH_PLUGIN_ERRORS",
        "COUNT_AUTHENTICATION_ERRORS",
        "COUNT_DEFAULT_DATABASE_ERRORS",
        "COUNT_FCRDNS_ERRORS",
        "COUNT_FORMAT_ERRORS",
        "COUNT_HANDSHAKE_ERRORS",
        "COUNT_HOST_ACL_ERRORS",
        "COUNT_HOST_BLOCKED_ERRORS",
        "COUNT_INIT_CONNECT_ERRORS",
        "COUNT_LOCAL_ERRORS",
        "COUNT_MAX_USER_CONNECTIONS_ERRORS",
        "COUNT_MAX_USER_CONNECTIONS_PER_HOUR_ERRORS",
        "COUNT_NAMEINFO_PERMANENT_ERRORS",
        "COUNT_NAMEINFO_TRANSIENT_ERRORS",
        "COUNT_NO_AUTH_PLUGIN_ERRORS",
        "COUNT_PROXY_USER_ACL_ERRORS",
        "COUNT_PROXY_USER_ERRORS",
        "COUNT_SSL_ERRORS",
        "COUNT_UNKNOWN_ERRORS",
        "FIRST_ERROR_SEEN",
        "FIRST_SEEN",
        "HOST",
        "HOST_VALIDATED",
        "IP",
        "LAST_ERROR_SEEN",
        "LAST_SEEN",
        "SUM_CONNECT_ERRORS"
    ),
    "mutex_instances": (
        "LOCKED_BY_THREAD_ID",
        "NAME",
        "OBJECT_INSTANCE_BEGIN"
    ),
    "objects_summary_global_by_type": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "SUM_TIMER_WAIT"
    ),
    "performance_timers": (
        "TIMER_FREQUENCY",
        "TIMER_NAME",
        "TIMER_OVERHEAD",
        "TIMER_RESOLUTION"
    ),
    "rwlock_instances": (
        "NAME",
        "OBJECT_INSTANCE_BEGIN",
        "READ_LOCKED_BY_COUNT",
        "WRITE_LOCKED_BY_THREAD_ID"
    ),
    "setup_actors": (
        "ENABLED",
        "HISTORY",
        "HOST",
        "ROLE",
        "USER"
    ),
    "setup_consumers": (
        "ENABLED",
        "NAME"
    ),
    "setup_instruments": (
        "DOCUMENTATION",
        "ENABLED",
        "FLAGS",
        "NAME",
        "PROPERTIES",
        "TIMED",
        "VOLATILITY"
    ),
    "setup_objects": (
        "ENABLED",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "TIMED"
    ),
    "setup_threads": (
        "DOCUMENTATION",
        "ENABLED",
        "HISTORY",
        "NAME",
        "PROPERTIES",
        "VOLATILITY"
    ),
    "table_io_waits_summary_by_index_usage": (
        "AVG_TIMER_DELETE",
        "AVG_TIMER_FETCH",
        "AVG_TIMER_INSERT",
        "AVG_TIMER_READ",
        "AVG_TIMER_UPDATE",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_DELETE",
        "COUNT_FETCH",
        "COUNT_INSERT",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_UPDATE",
        "COUNT_WRITE",
        "INDEX_NAME",
        "MAX_TIMER_DELETE",
        "MAX_TIMER_FETCH",
        "MAX_TIMER_INSERT",
        "MAX_TIMER_READ",
        "MAX_TIMER_UPDATE",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_DELETE",
        "MIN_TIMER_FETCH",
        "MIN_TIMER_INSERT",
        "MIN_TIMER_READ",
        "MIN_TIMER_UPDATE",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "SUM_TIMER_DELETE",
        "SUM_TIMER_FETCH",
        "SUM_TIMER_INSERT",
        "SUM_TIMER_READ",
        "SUM_TIMER_UPDATE",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "table_io_waits_summary_by_table": (
        "AVG_TIMER_DELETE",
        "AVG_TIMER_FETCH",
        "AVG_TIMER_INSERT",
        "AVG_TIMER_READ",
        "AVG_TIMER_UPDATE",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_DELETE",
        "COUNT_FETCH",
        "COUNT_INSERT",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_UPDATE",
        "COUNT_WRITE",
        "MAX_TIMER_DELETE",
        "MAX_TIMER_FETCH",
        "MAX_TIMER_INSERT",
        "MAX_TIMER_READ",
        "MAX_TIMER_UPDATE",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_DELETE",
        "MIN_TIMER_FETCH",
        "MIN_TIMER_INSERT",
        "MIN_TIMER_READ",
        "MIN_TIMER_UPDATE",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "SUM_TIMER_DELETE",
        "SUM_TIMER_FETCH",
        "SUM_TIMER_INSERT",
        "SUM_TIMER_READ",
        "SUM_TIMER_UPDATE",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "table_lock_waits_summary_by_table": (
        "AVG_TIMER_READ",
        "AVG_TIMER_READ_EXTERNAL",
        "AVG_TIMER_READ_HIGH_PRIORITY",
        "AVG_TIMER_READ_NO_INSERT",
        "AVG_TIMER_READ_NORMAL",
        "AVG_TIMER_READ_WITH_SHARED_LOCKS",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "AVG_TIMER_WRITE_ALLOW_WRITE",
        "AVG_TIMER_WRITE_CONCURRENT_INSERT",
        "AVG_TIMER_WRITE_EXTERNAL",
        "AVG_TIMER_WRITE_LOW_PRIORITY",
        "AVG_TIMER_WRITE_NORMAL",
        "COUNT_READ",
        "COUNT_READ_EXTERNAL",
        "COUNT_READ_HIGH_PRIORITY",
        "COUNT_READ_NO_INSERT",
        "COUNT_READ_NORMAL",
        "COUNT_READ_WITH_SHARED_LOCKS",
        "COUNT_STAR",
        "COUNT_WRITE",
        "COUNT_WRITE_ALLOW_WRITE",
        "COUNT_WRITE_CONCURRENT_INSERT",
        "COUNT_WRITE_EXTERNAL",
        "COUNT_WRITE_LOW_PRIORITY",
        "COUNT_WRITE_NORMAL",
        "MAX_TIMER_READ",
        "MAX_TIMER_READ_EXTERNAL",
        "MAX_TIMER_READ_HIGH_PRIORITY",
        "MAX_TIMER_READ_NO_INSERT",
        "MAX_TIMER_READ_NORMAL",
        "MAX_TIMER_READ_WITH_SHARED_LOCKS",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MAX_TIMER_WRITE_ALLOW_WRITE",
        "MAX_TIMER_WRITE_CONCURRENT_INSERT",
        "MAX_TIMER_WRITE_EXTERNAL",
        "MAX_TIMER_WRITE_LOW_PRIORITY",
        "MAX_TIMER_WRITE_NORMAL",
        "MIN_TIMER_READ",
        "MIN_TIMER_READ_EXTERNAL",
        "MIN_TIMER_READ_HIGH_PRIORITY",
        "MIN_TIMER_READ_NO_INSERT",
        "MIN_TIMER_READ_NORMAL",
        "MIN_TIMER_READ_WITH_SHARED_LOCKS",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "MIN_TIMER_WRITE_ALLOW_WRITE",
        "MIN_TIMER_WRITE_CONCURRENT_INSERT",
        "MIN_TIMER_WRITE_EXTERNAL",
        "MIN_TIMER_WRITE_LOW_PRIORITY",
        "MIN_TIMER_WRITE_NORMAL",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "SUM_TIMER_READ",
        "SUM_TIMER_READ_EXTERNAL",
        "SUM_TIMER_READ_HIGH_PRIORITY",
        "SUM_TIMER_READ_NO_INSERT",
        "SUM_TIMER_READ_NORMAL",
        "SUM_TIMER_READ_WITH_SHARED_LOCKS",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE",
        "SUM_TIMER_WRITE_ALLOW_WRITE",
        "SUM_TIMER_WRITE_CONCURRENT_INSERT",
        "SUM_TIMER_WRITE_EXTERNAL",
        "SUM_TIMER_WRITE_LOW_PRIORITY",
        "SUM_TIMER_WRITE_NORMAL"
    ),
    "threads": (
        "CONNECTION_TYPE",
        "CONTROLLED_MEMORY",
        "EXECUTION_ENGINE",
        "HISTORY",
        "INSTRUMENTED",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TOTAL_MEMORY",
        "NAME",
        "PARENT_THREAD_ID",
        "PROCESSLIST_COMMAND",
        "PROCESSLIST_DB",
        "PROCESSLIST_HOST",
        "PROCESSLIST_ID",
        "PROCESSLIST_INFO",
        "PROCESSLIST_STATE",
        "PROCESSLIST_TIME",
        "PROCESSLIST_USER",
        "RESOURCE_GROUP",
        "ROLE",
        "TELEMETRY_ACTIVE",
        "THREAD_ID",
        "THREAD_OS_ID",
        "TOTAL_MEMORY",
        "TYPE"
    ),
    "events_stages_current": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "SOURCE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WORK_COMPLETED",
        "WORK_ESTIMATED"
    ),
    "events_stages_history": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "SOURCE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WORK_COMPLETED",
        "WORK_ESTIMATED"
    ),
    "events_stages_history_long": (
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "SOURCE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WORK_COMPLETED",
        "WORK_ESTIMATED"
    ),
    "events_stages_summary_by_thread_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "THREAD_ID"
    ),
    "events_stages_summary_by_account_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_stages_summary_by_user_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_stages_summary_by_host_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT"
    ),
    "events_stages_summary_global_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_WAIT"
    ),
    "events_statements_current": (
        "CPU_TIME",
        "CREATED_TMP_DISK_TABLES",
        "CREATED_TMP_TABLES",
        "CURRENT_SCHEMA",
        "DIGEST",
        "DIGEST_TEXT",
        "END_EVENT_ID",
        "ERRORS",
        "EVENT_ID",
        "EVENT_NAME",
        "EXECUTION_ENGINE",
        "LOCK_TIME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TOTAL_MEMORY",
        "MESSAGE_TEXT",
        "MYSQL_ERRNO",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_LEVEL",
        "NESTING_EVENT_TYPE",
        "NO_GOOD_INDEX_USED",
        "NO_INDEX_USED",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "RETURNED_SQLSTATE",
        "ROWS_AFFECTED",
        "ROWS_EXAMINED",
        "ROWS_SENT",
        "SELECT_FULL_JOIN",
        "SELECT_FULL_RANGE_JOIN",
        "SELECT_RANGE",
        "SELECT_RANGE_CHECK",
        "SELECT_SCAN",
        "SORT_MERGE_PASSES",
        "SORT_RANGE",
        "SORT_ROWS",
        "SORT_SCAN",
        "SOURCE",
        "SQL_TEXT",
        "STATEMENT_ID",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WARNINGS"
    ),
    "events_statements_history": (
        "CPU_TIME",
        "CREATED_TMP_DISK_TABLES",
        "CREATED_TMP_TABLES",
        "CURRENT_SCHEMA",
        "DIGEST",
        "DIGEST_TEXT",
        "END_EVENT_ID",
        "ERRORS",
        "EVENT_ID",
        "EVENT_NAME",
        "EXECUTION_ENGINE",
        "LOCK_TIME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TOTAL_MEMORY",
        "MESSAGE_TEXT",
        "MYSQL_ERRNO",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_LEVEL",
        "NESTING_EVENT_TYPE",
        "NO_GOOD_INDEX_USED",
        "NO_INDEX_USED",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "RETURNED_SQLSTATE",
        "ROWS_AFFECTED",
        "ROWS_EXAMINED",
        "ROWS_SENT",
        "SELECT_FULL_JOIN",
        "SELECT_FULL_RANGE_JOIN",
        "SELECT_RANGE",
        "SELECT_RANGE_CHECK",
        "SELECT_SCAN",
        "SORT_MERGE_PASSES",
        "SORT_RANGE",
        "SORT_ROWS",
        "SORT_SCAN",
        "SOURCE",
        "SQL_TEXT",
        "STATEMENT_ID",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WARNINGS"
    ),
    "events_statements_history_long": (
        "CPU_TIME",
        "CREATED_TMP_DISK_TABLES",
        "CREATED_TMP_TABLES",
        "CURRENT_SCHEMA",
        "DIGEST",
        "DIGEST_TEXT",
        "END_EVENT_ID",
        "ERRORS",
        "EVENT_ID",
        "EVENT_NAME",
        "EXECUTION_ENGINE",
        "LOCK_TIME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TOTAL_MEMORY",
        "MESSAGE_TEXT",
        "MYSQL_ERRNO",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_LEVEL",
        "NESTING_EVENT_TYPE",
        "NO_GOOD_INDEX_USED",
        "NO_INDEX_USED",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "RETURNED_SQLSTATE",
        "ROWS_AFFECTED",
        "ROWS_EXAMINED",
        "ROWS_SENT",
        "SELECT_FULL_JOIN",
        "SELECT_FULL_RANGE_JOIN",
        "SELECT_RANGE",
        "SELECT_RANGE_CHECK",
        "SELECT_SCAN",
        "SORT_MERGE_PASSES",
        "SORT_RANGE",
        "SORT_ROWS",
        "SORT_SCAN",
        "SOURCE",
        "SQL_TEXT",
        "STATEMENT_ID",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "WARNINGS"
    ),
    "events_statements_summary_by_thread_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS",
        "THREAD_ID"
    ),
    "events_statements_summary_by_account_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS",
        "USER"
    ),
    "events_statements_summary_by_user_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS",
        "USER"
    ),
    "events_statements_summary_by_host_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS"
    ),
    "events_statements_summary_global_by_event_name": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS"
    ),
    "events_statements_summary_by_digest": (
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "DIGEST",
        "DIGEST_TEXT",
        "FIRST_SEEN",
        "LAST_SEEN",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_WAIT",
        "QUANTILE_95",
        "QUANTILE_99",
        "QUANTILE_999",
        "QUERY_SAMPLE_SEEN",
        "QUERY_SAMPLE_TEXT",
        "QUERY_SAMPLE_TIMER_WAIT",
        "SCHEMA_NAME",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS"
    ),
    "events_statements_summary_by_program": (
        "AVG_STATEMENTS_WAIT",
        "AVG_TIMER_WAIT",
        "COUNT_SECONDARY",
        "COUNT_STAR",
        "COUNT_STATEMENTS",
        "MAX_CONTROLLED_MEMORY",
        "MAX_STATEMENTS_WAIT",
        "MAX_TIMER_WAIT",
        "MAX_TOTAL_MEMORY",
        "MIN_STATEMENTS_WAIT",
        "MIN_TIMER_WAIT",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_STATEMENTS_WAIT",
        "SUM_TIMER_WAIT",
        "SUM_WARNINGS"
    ),
    "events_statements_histogram_global": (
        "BUCKET_NUMBER",
        "BUCKET_QUANTILE",
        "BUCKET_TIMER_HIGH",
        "BUCKET_TIMER_LOW",
        "COUNT_BUCKET",
        "COUNT_BUCKET_AND_LOWER"
    ),
    "events_statements_histogram_by_digest": (
        "BUCKET_NUMBER",
        "BUCKET_QUANTILE",
        "BUCKET_TIMER_HIGH",
        "BUCKET_TIMER_LOW",
        "COUNT_BUCKET",
        "COUNT_BUCKET_AND_LOWER",
        "DIGEST",
        "SCHEMA_NAME"
    ),
    "events_transactions_current": (
        "ACCESS_MODE",
        "AUTOCOMMIT",
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "GTID",
        "ISOLATION_LEVEL",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_RELEASE_SAVEPOINT",
        "NUMBER_OF_ROLLBACK_TO_SAVEPOINT",
        "NUMBER_OF_SAVEPOINTS",
        "OBJECT_INSTANCE_BEGIN",
        "SOURCE",
        "STATE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "TRX_ID",
        "XA_STATE",
        "XID_BQUAL",
        "XID_FORMAT_ID",
        "XID_GTRID"
    ),
    "events_transactions_history": (
        "ACCESS_MODE",
        "AUTOCOMMIT",
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "GTID",
        "ISOLATION_LEVEL",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_RELEASE_SAVEPOINT",
        "NUMBER_OF_ROLLBACK_TO_SAVEPOINT",
        "NUMBER_OF_SAVEPOINTS",
        "OBJECT_INSTANCE_BEGIN",
        "SOURCE",
        "STATE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "TRX_ID",
        "XA_STATE",
        "XID_BQUAL",
        "XID_FORMAT_ID",
        "XID_GTRID"
    ),
    "events_transactions_history_long": (
        "ACCESS_MODE",
        "AUTOCOMMIT",
        "END_EVENT_ID",
        "EVENT_ID",
        "EVENT_NAME",
        "GTID",
        "ISOLATION_LEVEL",
        "NESTING_EVENT_ID",
        "NESTING_EVENT_TYPE",
        "NUMBER_OF_RELEASE_SAVEPOINT",
        "NUMBER_OF_ROLLBACK_TO_SAVEPOINT",
        "NUMBER_OF_SAVEPOINTS",
        "OBJECT_INSTANCE_BEGIN",
        "SOURCE",
        "STATE",
        "THREAD_ID",
        "TIMER_END",
        "TIMER_START",
        "TIMER_WAIT",
        "TRX_ID",
        "XA_STATE",
        "XID_BQUAL",
        "XID_FORMAT_ID",
        "XID_GTRID"
    ),
    "events_transactions_summary_by_thread_by_event_name": (
        "AVG_TIMER_READ_ONLY",
        "AVG_TIMER_READ_WRITE",
        "AVG_TIMER_WAIT",
        "COUNT_READ_ONLY",
        "COUNT_READ_WRITE",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_READ_ONLY",
        "MAX_TIMER_READ_WRITE",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_READ_ONLY",
        "MIN_TIMER_READ_WRITE",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_READ_ONLY",
        "SUM_TIMER_READ_WRITE",
        "SUM_TIMER_WAIT",
        "THREAD_ID"
    ),
    "events_transactions_summary_by_account_by_event_name": (
        "AVG_TIMER_READ_ONLY",
        "AVG_TIMER_READ_WRITE",
        "AVG_TIMER_WAIT",
        "COUNT_READ_ONLY",
        "COUNT_READ_WRITE",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_READ_ONLY",
        "MAX_TIMER_READ_WRITE",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_READ_ONLY",
        "MIN_TIMER_READ_WRITE",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_READ_ONLY",
        "SUM_TIMER_READ_WRITE",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_transactions_summary_by_user_by_event_name": (
        "AVG_TIMER_READ_ONLY",
        "AVG_TIMER_READ_WRITE",
        "AVG_TIMER_WAIT",
        "COUNT_READ_ONLY",
        "COUNT_READ_WRITE",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_READ_ONLY",
        "MAX_TIMER_READ_WRITE",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_READ_ONLY",
        "MIN_TIMER_READ_WRITE",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_READ_ONLY",
        "SUM_TIMER_READ_WRITE",
        "SUM_TIMER_WAIT",
        "USER"
    ),
    "events_transactions_summary_by_host_by_event_name": (
        "AVG_TIMER_READ_ONLY",
        "AVG_TIMER_READ_WRITE",
        "AVG_TIMER_WAIT",
        "COUNT_READ_ONLY",
        "COUNT_READ_WRITE",
        "COUNT_STAR",
        "EVENT_NAME",
        "HOST",
        "MAX_TIMER_READ_ONLY",
        "MAX_TIMER_READ_WRITE",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_READ_ONLY",
        "MIN_TIMER_READ_WRITE",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_READ_ONLY",
        "SUM_TIMER_READ_WRITE",
        "SUM_TIMER_WAIT"
    ),
    "events_transactions_summary_global_by_event_name": (
        "AVG_TIMER_READ_ONLY",
        "AVG_TIMER_READ_WRITE",
        "AVG_TIMER_WAIT",
        "COUNT_READ_ONLY",
        "COUNT_READ_WRITE",
        "COUNT_STAR",
        "EVENT_NAME",
        "MAX_TIMER_READ_ONLY",
        "MAX_TIMER_READ_WRITE",
        "MAX_TIMER_WAIT",
        "MIN_TIMER_READ_ONLY",
        "MIN_TIMER_READ_WRITE",
        "MIN_TIMER_WAIT",
        "SUM_TIMER_READ_ONLY",
        "SUM_TIMER_READ_WRITE",
        "SUM_TIMER_WAIT"
    ),
    "events_errors_summary_by_user_by_error": (
        "ERROR_NAME",
        "ERROR_NUMBER",
        "FIRST_SEEN",
        "LAST_SEEN",
        "SQL_STATE",
        "SUM_ERROR_HANDLED",
        "SUM_ERROR_RAISED",
        "USER"
    ),
    "events_errors_summary_by_host_by_error": (
        "ERROR_NAME",
        "ERROR_NUMBER",
        "FIRST_SEEN",
        "HOST",
        "LAST_SEEN",
        "SQL_STATE",
        "SUM_ERROR_HANDLED",
        "SUM_ERROR_RAISED"
    ),
    "events_errors_summary_by_account_by_error": (
        "ERROR_NAME",
        "ERROR_NUMBER",
        "FIRST_SEEN",
        "HOST",
        "LAST_SEEN",
        "SQL_STATE",
        "SUM_ERROR_HANDLED",
        "SUM_ERROR_RAISED",
        "USER"
    ),
    "events_errors_summary_by_thread_by_error": (
        "ERROR_NAME",
        "ERROR_NUMBER",
        "FIRST_SEEN",
        "LAST_SEEN",
        "SQL_STATE",
        "SUM_ERROR_HANDLED",
        "SUM_ERROR_RAISED",
        "THREAD_ID"
    ),
    "events_errors_summary_global_by_error": (
        "ERROR_NAME",
        "ERROR_NUMBER",
        "FIRST_SEEN",
        "LAST_SEEN",
        "SQL_STATE",
        "SUM_ERROR_HANDLED",
        "SUM_ERROR_RAISED"
    ),
    "users": (
        "CURRENT_CONNECTIONS",
        "MAX_SESSION_CONTROLLED_MEMORY",
        "MAX_SESSION_TOTAL_MEMORY",
        "TOTAL_CONNECTIONS",
        "USER"
    ),
    "accounts": (
        "CURRENT_CONNECTIONS",
        "HOST",
        "MAX_SESSION_CONTROLLED_MEMORY",
        "MAX_SESSION_TOTAL_MEMORY",
        "TOTAL_CONNECTIONS",
        "USER"
    ),
    "hosts": (
        "CURRENT_CONNECTIONS",
        "HOST",
        "MAX_SESSION_CONTROLLED_MEMORY",
        "MAX_SESSION_TOTAL_MEMORY",
        "TOTAL_CONNECTIONS"
    ),
    "socket_instances": (
        "EVENT_NAME",
        "IP",
        "OBJECT_INSTANCE_BEGIN",
        "PORT",
        "SOCKET_ID",
        "STATE",
        "THREAD_ID"
    ),
    "socket_summary_by_instance": (
        "AVG_TIMER_MISC",
        "AVG_TIMER_READ",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_MISC",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_WRITE",
        "EVENT_NAME",
        "MAX_TIMER_MISC",
        "MAX_TIMER_READ",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_MISC",
        "MIN_TIMER_READ",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "OBJECT_INSTANCE_BEGIN",
        "SUM_NUMBER_OF_BYTES_READ",
        "SUM_NUMBER_OF_BYTES_WRITE",
        "SUM_TIMER_MISC",
        "SUM_TIMER_READ",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "socket_summary_by_event_name": (
        "AVG_TIMER_MISC",
        "AVG_TIMER_READ",
        "AVG_TIMER_WAIT",
        "AVG_TIMER_WRITE",
        "COUNT_MISC",
        "COUNT_READ",
        "COUNT_STAR",
        "COUNT_WRITE",
        "EVENT_NAME",
        "MAX_TIMER_MISC",
        "MAX_TIMER_READ",
        "MAX_TIMER_WAIT",
        "MAX_TIMER_WRITE",
        "MIN_TIMER_MISC",
        "MIN_TIMER_READ",
        "MIN_TIMER_WAIT",
        "MIN_TIMER_WRITE",
        "SUM_NUMBER_OF_BYTES_READ",
        "SUM_NUMBER_OF_BYTES_WRITE",
        "SUM_TIMER_MISC",
        "SUM_TIMER_READ",
        "SUM_TIMER_WAIT",
        "SUM_TIMER_WRITE"
    ),
    "session_connect_attrs": (
        "ATTR_NAME",
        "ATTR_VALUE",
        "ORDINAL_POSITION",
        "PROCESSLIST_ID"
    ),
    "session_account_connect_attrs": (
        "ATTR_NAME",
        "ATTR_VALUE",
        "ORDINAL_POSITION",
        "PROCESSLIST_ID"
    ),
    "keyring_keys": (
        "BACKEND_KEY_ID",
        "KEY_ID",
        "KEY_OWNER"
    ),
    "memory_summary_global_by_event_name": (
        "COUNT_ALLOC",
        "COUNT_FREE",
        "CURRENT_COUNT_USED",
        "CURRENT_NUMBER_OF_BYTES_USED",
        "EVENT_NAME",
        "HIGH_COUNT_USED",
        "HIGH_NUMBER_OF_BYTES_USED",
        "LOW_COUNT_USED",
        "LOW_NUMBER_OF_BYTES_USED",
        "SUM_NUMBER_OF_BYTES_ALLOC",
        "SUM_NUMBER_OF_BYTES_FREE"
    ),
    "memory_summary_by_account_by_event_name": (
        "COUNT_ALLOC",
        "COUNT_FREE",
        "CURRENT_COUNT_USED",
        "CURRENT_NUMBER_OF_BYTES_USED",
        "EVENT_NAME",
        "HIGH_COUNT_USED",
        "HIGH_NUMBER_OF_BYTES_USED",
        "HOST",
        "LOW_COUNT_USED",
        "LOW_NUMBER_OF_BYTES_USED",
        "SUM_NUMBER_OF_BYTES_ALLOC",
        "SUM_NUMBER_OF_BYTES_FREE",
        "USER"
    ),
    "memory_summary_by_host_by_event_name": (
        "COUNT_ALLOC",
        "COUNT_FREE",
        "CURRENT_COUNT_USED",
        "CURRENT_NUMBER_OF_BYTES_USED",
        "EVENT_NAME",
        "HIGH_COUNT_USED",
        "HIGH_NUMBER_OF_BYTES_USED",
        "HOST",
        "LOW_COUNT_USED",
        "LOW_NUMBER_OF_BYTES_USED",
        "SUM_NUMBER_OF_BYTES_ALLOC",
        "SUM_NUMBER_OF_BYTES_FREE"
    ),
    "memory_summary_by_thread_by_event_name": (
        "COUNT_ALLOC",
        "COUNT_FREE",
        "CURRENT_COUNT_USED",
        "CURRENT_NUMBER_OF_BYTES_USED",
        "EVENT_NAME",
        "HIGH_COUNT_USED",
        "HIGH_NUMBER_OF_BYTES_USED",
        "LOW_COUNT_USED",
        "LOW_NUMBER_OF_BYTES_USED",
        "SUM_NUMBER_OF_BYTES_ALLOC",
        "SUM_NUMBER_OF_BYTES_FREE",
        "THREAD_ID"
    ),
    "memory_summary_by_user_by_event_name": (
        "COUNT_ALLOC",
        "COUNT_FREE",
        "CURRENT_COUNT_USED",
        "CURRENT_NUMBER_OF_BYTES_USED",
        "EVENT_NAME",
        "HIGH_COUNT_USED",
        "HIGH_NUMBER_OF_BYTES_USED",
        "LOW_COUNT_USED",
        "LOW_NUMBER_OF_BYTES_USED",
        "SUM_NUMBER_OF_BYTES_ALLOC",
        "SUM_NUMBER_OF_BYTES_FREE",
        "USER"
    ),
    "table_handles": (
        "EXTERNAL_LOCK",
        "INTERNAL_LOCK",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "OWNER_EVENT_ID",
        "OWNER_THREAD_ID"
    ),
    "metadata_locks": (
        "COLUMN_NAME",
        "LOCK_DURATION",
        "LOCK_STATUS",
        "LOCK_TYPE",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "OBJECT_TYPE",
        "OWNER_EVENT_ID",
        "OWNER_THREAD_ID",
        "SOURCE"
    ),
    "data_locks": (
        "ENGINE",
        "ENGINE_LOCK_ID",
        "ENGINE_TRANSACTION_ID",
        "EVENT_ID",
        "INDEX_NAME",
        "LOCK_DATA",
        "LOCK_MODE",
        "LOCK_STATUS",
        "LOCK_TYPE",
        "OBJECT_INSTANCE_BEGIN",
        "OBJECT_NAME",
        "OBJECT_SCHEMA",
        "PARTITION_NAME",
        "SUBPARTITION_NAME",
        "THREAD_ID"
    ),
    "data_lock_waits": (
        "BLOCKING_ENGINE_LOCK_ID",
        "BLOCKING_ENGINE_TRANSACTION_ID",
        "BLOCKING_EVENT_ID",
        "BLOCKING_OBJECT_INSTANCE_BEGIN",
        "BLOCKING_THREAD_ID",
        "ENGINE",
        "REQUESTING_ENGINE_LOCK_ID",
        "REQUESTING_ENGINE_TRANSACTION_ID",
        "REQUESTING_EVENT_ID",
        "REQUESTING_OBJECT_INSTANCE_BEGIN",
        "REQUESTING_THREAD_ID"
    ),
    "replication_connection_configuration": (
        "AUTO_POSITION",
        "CHANNEL_NAME",
        "COMPRESSION_ALGORITHM",
        "CONNECTION_RETRY_COUNT",
        "CONNECTION_RETRY_INTERVAL",
        "GET_PUBLIC_KEY",
        "GTID_ONLY",
        "HEARTBEAT_INTERVAL",
        "HOST",
        "NETWORK_INTERFACE",
        "NETWORK_NAMESPACE",
        "PORT",
        "PUBLIC_KEY_PATH",
        "SOURCE_CONNECTION_AUTO_FAILOVER",
        "SSL_ALLOWED",
        "SSL_CA_FILE",
        "SSL_CA_PATH",
        "SSL_CERTIFICATE",
        "SSL_CIPHER",
        "SSL_CRL_FILE",
        "SSL_CRL_PATH",
        "SSL_KEY",
        "SSL_VERIFY_SERVER_CERTIFICATE",
        "TLS_CIPHERSUITES",
        "TLS_VERSION",
        "USER",
        "ZSTD_COMPRESSION_LEVEL"
    ),
    "replication_group_members": (
        "CHANNEL_NAME",
        "MEMBER_COMMUNICATION_STACK",
        "MEMBER_HOST",
        "MEMBER_ID",
        "MEMBER_PORT",
        "MEMBER_ROLE",
        "MEMBER_STATE",
        "MEMBER_VERSION"
    ),
    "replication_connection_status": (
        "CHANNEL_NAME",
        "COUNT_RECEIVED_HEARTBEATS",
        "GROUP_NAME",
        "LAST_ERROR_MESSAGE",
        "LAST_ERROR_NUMBER",
        "LAST_ERROR_TIMESTAMP",
        "LAST_HEARTBEAT_TIMESTAMP",
        "LAST_QUEUED_TRANSACTION",
        "LAST_QUEUED_TRANSACTION_END_QUEUE_TIMESTAMP",
        "LAST_QUEUED_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "LAST_QUEUED_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "LAST_QUEUED_TRANSACTION_START_QUEUE_TIMESTAMP",
        "QUEUEING_TRANSACTION",
        "QUEUEING_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "QUEUEING_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "QUEUEING_TRANSACTION_START_QUEUE_TIMESTAMP",
        "RECEIVED_TRANSACTION_SET",
        "SERVICE_STATE",
        "SOURCE_UUID",
        "THREAD_ID"
    ),
    "replication_applier_configuration": (
        "ASSIGN_GTIDS_TO_ANONYMOUS_TRANSACTIONS_TYPE",
        "ASSIGN_GTIDS_TO_ANONYMOUS_TRANSACTIONS_VALUE",
        "CHANNEL_NAME",
        "DESIRED_DELAY",
        "PRIVILEGE_CHECKS_USER",
        "REQUIRE_ROW_FORMAT",
        "REQUIRE_TABLE_PRIMARY_KEY_CHECK"
    ),
    "replication_applier_status": (
        "CHANNEL_NAME",
        "COUNT_TRANSACTIONS_RETRIES",
        "REMAINING_DELAY",
        "SERVICE_STATE"
    ),
    "replication_applier_status_by_coordinator": (
        "CHANNEL_NAME",
        "LAST_ERROR_MESSAGE",
        "LAST_ERROR_NUMBER",
        "LAST_ERROR_TIMESTAMP",
        "LAST_PROCESSED_TRANSACTION",
        "LAST_PROCESSED_TRANSACTION_END_BUFFER_TIMESTAMP",
        "LAST_PROCESSED_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "LAST_PROCESSED_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "LAST_PROCESSED_TRANSACTION_START_BUFFER_TIMESTAMP",
        "PROCESSING_TRANSACTION",
        "PROCESSING_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "PROCESSING_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "PROCESSING_TRANSACTION_START_BUFFER_TIMESTAMP",
        "SERVICE_STATE",
        "THREAD_ID"
    ),
    "replication_applier_status_by_worker": (
        "APPLYING_TRANSACTION",
        "APPLYING_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "APPLYING_TRANSACTION_LAST_TRANSIENT_ERROR_MESSAGE",
        "APPLYING_TRANSACTION_LAST_TRANSIENT_ERROR_NUMBER",
        "APPLYING_TRANSACTION_LAST_TRANSIENT_ERROR_TIMESTAMP",
        "APPLYING_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "APPLYING_TRANSACTION_RETRIES_COUNT",
        "APPLYING_TRANSACTION_START_APPLY_TIMESTAMP",
        "CHANNEL_NAME",
        "LAST_APPLIED_TRANSACTION",
        "LAST_APPLIED_TRANSACTION_END_APPLY_TIMESTAMP",
        "LAST_APPLIED_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP",
        "LAST_APPLIED_TRANSACTION_LAST_TRANSIENT_ERROR_MESSAGE",
        "LAST_APPLIED_TRANSACTION_LAST_TRANSIENT_ERROR_NUMBER",
        "LAST_APPLIED_TRANSACTION_LAST_TRANSIENT_ERROR_TIMESTAMP",
        "LAST_APPLIED_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP",
        "LAST_APPLIED_TRANSACTION_RETRIES_COUNT",
        "LAST_APPLIED_TRANSACTION_START_APPLY_TIMESTAMP",
        "LAST_ERROR_MESSAGE",
        "LAST_ERROR_NUMBER",
        "LAST_ERROR_TIMESTAMP",
        "SERVICE_STATE",
        "THREAD_ID",
        "WORKER_ID"
    ),
    "replication_group_member_stats": (
        "CHANNEL_NAME",
        "COUNT_CONFLICTS_DETECTED",
        "COUNT_TRANSACTIONS_CHECKED",
        "COUNT_TRANSACTIONS_IN_QUEUE",
        "COUNT_TRANSACTIONS_LOCAL_PROPOSED",
        "COUNT_TRANSACTIONS_LOCAL_ROLLBACK",
        "COUNT_TRANSACTIONS_REMOTE_APPLIED",
        "COUNT_TRANSACTIONS_REMOTE_IN_APPLIER_QUEUE",
        "COUNT_TRANSACTIONS_ROWS_VALIDATING",
        "LAST_CONFLICT_FREE_TRANSACTION",
        "MEMBER_ID",
        "TRANSACTIONS_COMMITTED_ALL_MEMBERS",
        "VIEW_ID"
    ),
    "replication_applier_filters": (
        "ACTIVE_SINCE",
        "CHANNEL_NAME",
        "CONFIGURED_BY",
        "COUNTER",
        "FILTER_NAME",
        "FILTER_RULE"
    ),
    "replication_applier_global_filters": (
        "ACTIVE_SINCE",
        "CONFIGURED_BY",
        "FILTER_NAME",
        "FILTER_RULE"
    ),
    "replication_asynchronous_connection_failover": (
        "CHANNEL_NAME",
        "HOST",
        "MANAGED_NAME",
        "NETWORK_NAMESPACE",
        "PORT",
        "WEIGHT",
        "Channel_name",
        "Host",
        "Managed_name",
        "Network_namespace",
        "Port",
        "Weight"
    ),
    "replication_asynchronous_connection_failover_managed": (
        "CHANNEL_NAME",
        "CONFIGURATION",
        "MANAGED_NAME",
        "MANAGED_TYPE",
        "Channel_name",
        "Configuration",
        "Managed_name",
        "Managed_type"
    ),
    "log_status": (
        "LOCAL",
        "REPLICATION",
        "SERVER_UUID",
        "STORAGE_ENGINES"
    ),
    "prepared_statements_instances": (
        "AVG_TIMER_EXECUTE",
        "COUNT_EXECUTE",
        "COUNT_REPREPARE",
        "COUNT_SECONDARY",
        "EXECUTION_ENGINE",
        "MAX_CONTROLLED_MEMORY",
        "MAX_TIMER_EXECUTE",
        "MAX_TOTAL_MEMORY",
        "MIN_TIMER_EXECUTE",
        "OBJECT_INSTANCE_BEGIN",
        "OWNER_EVENT_ID",
        "OWNER_OBJECT_NAME",
        "OWNER_OBJECT_SCHEMA",
        "OWNER_OBJECT_TYPE",
        "OWNER_THREAD_ID",
        "SQL_TEXT",
        "STATEMENT_ID",
        "STATEMENT_NAME",
        "SUM_CPU_TIME",
        "SUM_CREATED_TMP_DISK_TABLES",
        "SUM_CREATED_TMP_TABLES",
        "SUM_ERRORS",
        "SUM_LOCK_TIME",
        "SUM_NO_GOOD_INDEX_USED",
        "SUM_NO_INDEX_USED",
        "SUM_ROWS_AFFECTED",
        "SUM_ROWS_EXAMINED",
        "SUM_ROWS_SENT",
        "SUM_SELECT_FULL_JOIN",
        "SUM_SELECT_FULL_RANGE_JOIN",
        "SUM_SELECT_RANGE",
        "SUM_SELECT_RANGE_CHECK",
        "SUM_SELECT_SCAN",
        "SUM_SORT_MERGE_PASSES",
        "SUM_SORT_RANGE",
        "SUM_SORT_ROWS",
        "SUM_SORT_SCAN",
        "SUM_TIMER_EXECUTE",
        "SUM_WARNINGS",
        "TIMER_PREPARE"
    ),
    "user_variables_by_thread": (
        "THREAD_ID",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "status_by_account": (
        "HOST",
        "USER",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "status_by_host": (
        "HOST",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "status_by_thread": (
        "THREAD_ID",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "status_by_user": (
        "USER",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "global_status": (
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "session_status": (
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "variables_by_thread": (
        "THREAD_ID",
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "global_variables": (
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "session_variables": (
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "variables_info": (
        "MAX_VALUE",
        "MIN_VALUE",
        "SET_HOST",
        "SET_TIME",
        "SET_USER",
        "VARIABLE_NAME",
        "VARIABLE_PATH",
        "VARIABLE_SOURCE"
    ),
    "persisted_variables": (
        "VARIABLE_NAME",
        "VARIABLE_VALUE"
    ),
    "user_defined_functions": (
        "UDF_LIBRARY",
        "UDF_NAME",
        "UDF_RETURN_TYPE",
        "UDF_TYPE",
        "UDF_USAGE_COUNT"
    ),
    "binary_log_transaction_compression_stats": (
        "COMPRESSED_BYTES_COUNTER",
        "COMPRESSION_PERCENTAGE",
        "COMPRESSION_TYPE",
        "FIRST_TRANSACTION_COMPRESSED_BYTES",
        "FIRST_TRANSACTION_ID",
        "FIRST_TRANSACTION_TIMESTAMP",
        "FIRST_TRANSACTION_UNCOMPRESSED_BYTES",
        "LAST_TRANSACTION_COMPRESSED_BYTES",
        "LAST_TRANSACTION_ID",
        "LAST_TRANSACTION_TIMESTAMP",
        "LAST_TRANSACTION_UNCOMPRESSED_BYTES",
        "LOG_TYPE",
        "TRANSACTION_COUNTER",
        "UNCOMPRESSED_BYTES_COUNTER"
    ),
    "tls_channel_status": (
        "CHANNEL",
        "PROPERTY",
        "VALUE"
    ),
    "rds_processlist": (
        "COMMAND",
        "DB",
        "HOST",
        "ID",
        "INFO",
        "QUERY_DIGEST_HASH",
        "QUERY_DIGEST_TEXT",
        "STATE",
        "TIME",
        "USER"
    ),
    "rds_events_threads_waits_current": (
        "COMMAND",
        "DB",
        "EVENT_NAME",
        "EVENT_TIMER_END",
        "EVENT_TIMER_START",
        "EVENT_TIMER_WAIT",
        "HOST",
        "INFO",
        "NAME",
        "PARENT_THREAD_ID",
        "PROCESSLIST_ID",
        "QUERY_DIGEST_HASH",
        "QUERY_DIGEST_TEXT",
        "STATE",
        "THREAD_ID",
        "TIME",
        "TYPE",
        "USER"
    ),
    "rds_events_threads_waits_lock_current": (
        "BLOCKER_HOST",
        "BLOCKER_QUERY",
        "BLOCKER_STATE",
        "BLOCKER_THREAD_ID",
        "BLOCKER_TRANS_ID",
        "BLOCKER_USER",
        "WAIT_HOST",
        "WAIT_LOCK_ID",
        "WAIT_LOCK_MODE",
        "WAIT_LOCK_TYPE",
        "WAIT_QUERY",
        "WAIT_TABLE_LOCK_ID",
        "WAIT_THREAD_ID",
        "WAIT_TRANS_ID",
        "WAIT_USER"
    ),
    "keyring_component_status": (
        "STATUS_KEY",
        "STATUS_VALUE"
    ),
    "malloc_stats_totals": (
        "ACTIVE",
        "ALLOCATED",
        "MAPPED",
        "METADATA",
        "RESIDENT",
        "RETAINED"
    ),
    "malloc_stats": (
        "ALLOCATED",
        "NDALLOC",
        "NMALLOC",
        "NREQUESTS",
        "TYPE"
    ),
    "component": (
        "component_group_id",
        "component_id",
        "component_urn"
    ),
    "columns_priv": (
        "Column_name",
        "Column_priv",
        "Db",
        "Host",
        "Table_name",
        "Timestamp",
        "User"
    ),
    "db": (
        "Alter_priv",
        "Alter_routine_priv",
        "Create_priv",
        "Create_routine_priv",
        "Create_tmp_table_priv",
        "Create_view_priv",
        "Db",
        "Delete_priv",
        "Drop_priv",
        "Event_priv",
        "Execute_priv",
        "Grant_priv",
        "Host",
        "Index_priv",
        "Insert_priv",
        "Lock_tables_priv",
        "References_priv",
        "Select_priv",
        "Show_view_priv",
        "Trigger_priv",
        "Update_priv",
        "User"
    ),
    "default_roles": (
        "DEFAULT_ROLE_HOST",
        "DEFAULT_ROLE_USER",
        "HOST",
        "USER"
    ),
    "engine_cost": (
        "comment",
        "cost_name",
        "cost_value",
        "default_value",
        "device_type",
        "engine_name",
        "last_update"
    ),
    "global_grants": (
        "HOST",
        "PRIV",
        "USER",
        "WITH_GRANT_OPTION"
    ),
    "help_category": (
        "help_category_id",
        "name",
        "parent_category_id",
        "url"
    ),
    "help_keyword": (
        "help_keyword_id",
        "name"
    ),
    "help_relation": (
        "help_keyword_id",
        "help_topic_id"
    ),
    "help_topic": (
        "description",
        "example",
        "help_category_id",
        "help_topic_id",
        "name",
        "url"
    ),
    "plugin": (
        "dl",
        "name"
    ),
    "password_history": (
        "Host",
        "Password",
        "Password_timestamp",
        "User"
    ),
    "proxies_priv": (
        "Grantor",
        "Host",
        "Proxied_host",
        "Proxied_user",
        "Timestamp",
        "User",
        "With_grant"
    ),
    "role_edges": (
        "FROM_HOST",
        "FROM_USER",
        "TO_HOST",
        "TO_USER",
        "WITH_ADMIN_OPTION"
    ),
    "servers": (
        "Db",
        "Host",
        "Owner",
        "Password",
        "Port",
        "Server_name",
        "Socket",
        "Username",
        "Wrapper"
    ),
    "server_cost": (
        "comment",
        "cost_name",
        "cost_value",
        "default_value",
        "last_update"
    ),
    "replication_group_member_actions": (
        "enabled",
        "error_handling",
        "event",
        "name",
        "priority",
        "type"
    ),
    "tables_priv": (
        "Column_priv",
        "Db",
        "Grantor",
        "Host",
        "Table_name",
        "Table_priv",
        "Timestamp",
        "User"
    ),
    "time_zone_name": (
        "Name",
        "Time_zone_id"
    ),
    "time_zone_leap_second": (
        "Correction",
        "Transition_time"
    ),
    "time_zone_transition": (
        "Time_zone_id",
        "Transition_time",
        "Transition_type_id"
    ),
    "time_zone_transition_type": (
        "Abbreviation",
        "Is_DST",
        "Offset",
        "Time_zone_id",
        "Transition_type_id"
    ),
    "func": (
        "dl",
        "name",
        "ret",
        "type"
    ),
    "gtid_executed": (
        "interval_end",
        "interval_start",
        "source_uuid"
    ),
    "slave_master_info": (
        "Bind",
        "Channel_name",
        "Connect_retry",
        "Enabled_auto_position",
        "Enabled_ssl",
        "Get_public_key",
        "Gtid_only",
        "Heartbeat",
        "Host",
        "Ignored_server_ids",
        "Master_compression_algorithm",
        "Master_log_name",
        "Master_log_pos",
        "Master_zstd_compression_level",
        "Network_namespace",
        "Number_of_lines",
        "Port",
        "Public_key_path",
        "Retry_count",
        "Source_connection_auto_failover",
        "Ssl_ca",
        "Ssl_capath",
        "Ssl_cert",
        "Ssl_cipher",
        "Ssl_crl",
        "Ssl_crlpath",
        "Ssl_key",
        "Ssl_verify_server_cert",
        "Tls_ciphersuites",
        "Tls_version",
        "User_name",
        "User_password",
        "Uuid"
    ),
    "slave_worker_info": (
        "Channel_name",
        "Checkpoint_group_bitmap",
        "Checkpoint_group_size",
        "Checkpoint_master_log_name",
        "Checkpoint_master_log_pos",
        "Checkpoint_relay_log_name",
        "Checkpoint_relay_log_pos",
        "Checkpoint_seqno",
        "Id",
        "Master_log_name",
        "Master_log_pos",
        "Relay_log_name",
        "Relay_log_pos"
    ),
    "slave_relay_log_info": (
        "Assign_gtids_to_anonymous_transactions_type",
        "Assign_gtids_to_anonymous_transactions_value",
        "Channel_name",
        "Id",
        "Master_log_name",
        "Master_log_pos",
        "Number_of_lines",
        "Number_of_workers",
        "Privilege_checks_hostname",
        "Privilege_checks_username",
        "Relay_log_name",
        "Relay_log_pos",
        "Require_row_format",
        "Require_table_primary_key_check",
        "Sql_delay"
    ),
    "procs_priv": (
        "Db",
        "Grantor",
        "Host",
        "Proc_priv",
        "Routine_name",
        "Routine_type",
        "Timestamp",
        "User"
    ),
    "user": (
        "account_locked",
        "Alter_priv",
        "Alter_routine_priv",
        "authentication_string",
        "Create_priv",
        "Create_role_priv",
        "Create_routine_priv",
        "Create_tablespace_priv",
        "Create_tmp_table_priv",
        "Create_user_priv",
        "Create_view_priv",
        "Delete_priv",
        "Drop_priv",
        "Drop_role_priv",
        "Event_priv",
        "Execute_priv",
        "File_priv",
        "Grant_priv",
        "Host",
        "Index_priv",
        "Insert_priv",
        "Lock_tables_priv",
        "max_connections",
        "max_questions",
        "max_updates",
        "max_user_connections",
        "password_expired",
        "password_last_changed",
        "password_lifetime",
        "Password_require_current",
        "Password_reuse_history",
        "Password_reuse_time",
        "plugin",
        "Process_priv",
        "References_priv",
        "Reload_priv",
        "Repl_client_priv",
        "Repl_slave_priv",
        "Select_priv",
        "Show_db_priv",
        "Show_view_priv",
        "Shutdown_priv",
        "ssl_cipher",
        "ssl_type",
        "Super_priv",
        "Trigger_priv",
        "Update_priv",
        "User",
        "User_attributes",
        "x509_issuer",
        "x509_subject"
    ),
    "time_zone": (
        "Time_zone_id",
        "Use_leap_seconds"
    ),
    "sys_config": (
        "set_by",
        "set_time",
        "value",
        "variable"
    ),
    "rds_reserved_users": (
        "description",
        "host",
        "user"
    ),
    "rds_configuration": (
        "action_timestamp",
        "description",
        "name",
        "value"
    ),
    "rds_history": (
        "action",
        "action_timestamp",
        "auto_position",
        "called_by_user",
        "master_gtid",
        "master_host",
        "master_log_file",
        "master_log_pos",
        "master_port",
        "master_ssl",
        "master_user",
        "mysql_version"
    ),
    "rds_replication_status": (
        "action",
        "action_timestamp",
        "called_by_user",
        "master_host",
        "master_port",
        "mysql_version",
        "replication_gtid",
        "replication_log_file",
        "replication_stop_point"
    ),
    "repo_history_table": (
        "created_timestamp",
        "is_graphql",
        "is_local",
        "is_restapi",
        "is_server",
        "is_serverless_com",
        "repo_id",
        "serverless_framework_version"
    ),
    "mobile_application__user_external_table": (
        "fcm_token",
        "mobile_application__user_external_id",
        "mobile_application_id",
        "user_external_id"
    ),
    "release_bundle_table": (
        "release_bundle_id",
        "release_bundle_version"
    ),
    "microservice_release_bundle_table": (
        "microservice_id",
        "microservice_release_bundle_id",
        "release_bundle_id"
    ),
    "microservice_history_table": (
        "microservice_id",
        "microservice_version_id",
        "version"
    ),
    "microservice_table": (
        "microservice_id",
        "name",
        "repo_id"
    ),
    "mobile_application_table": (
        "brand_id",
        "mobile_application_id",
        "name"
    ),
    "test_suite_table": (
        "name",
        "test_suite_id"
    ),
    "test_result_table": (
        "created_timestamp",
        "number_of_errors",
        "number_of_warnings",
        "repo_id",
        "test_result_id",
        "test_suite_id"
    ),
    "calendar_item_table": (
        "action_item_id",
        "activity_id",
        "attendees",
        "available_for_activity_id",
        "billable_profile",
        "busy_with_action_item_id",
        "busy_with_activity_id",
        "calendar_id",
        "calendar_item_id",
        "calendar_item_type",
        "created_effective_prodile_id",
        "created_effective_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "event_id",
        "extendedProperties",
        "google_id",
        "htmlLink",
        "is_billable",
        "location",
        "location_id",
        "name",
        "occurrence_id",
        "recurrnece",
        "remiders",
        "source_user_external_id",
        "start_timestamp",
        "status",
        "updated_timestamp",
        "updated_user_id"
    ),
    "machine_learning_model_instance_table": (
        "consolidation_engine",
        "dns_ip",
        "machine_learning_model_id",
        "machine_learning_model_instance_id",
        "name",
        "port"
    ),
    "technical_release_notes_table": (
        "affected_release_bundle_version_id",
        "description",
        "fixed_component_version_id",
        "jira_work_item_id",
        "release_bundle_id",
        "summary",
        "technical_release_notes_id"
    ),
    "release_bundle_content_table": (
        "component_id",
        "component_version_id",
        "parent_release_bundle_content_id",
        "release_bundle_content_id"
    ),
    "component_table": (
        "api_type",
        "component_category",
        "component_id",
        "component_type",
        "created_timestamp",
        "created_user_id",
        "developer_email",
        "developer_profile_id",
        "end_timestamp",
        "github_repo",
        "layer_id",
        "name",
        "parent_component_id",
        "programming_language",
        "programming_language_id",
        "repo_name",
        "start_timestamp",
        "testing_framework",
        "updated_timestamp",
        "updated_user_id"
    ),
    "layer_table": (
        "layer_id"
    ),
    "field_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "database_field_name",
        "database_sub_field_name",
        "database_sub_field_value",
        "description",
        "end_timestamp",
        "enity_type_id",
        "field_id",
        "identifier",
        "internal_description",
        "is_action_item",
        "is_ml_table",
        "is_test_data",
        "name",
        "number",
        "processing_database_field_name",
        "processing_id",
        "start_timestamp",
        "table_id",
        "type",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "variable_id",
        "visibility_id"
    ),
    "message_template_text_block_ml_types_table": (
        "length",
        "message_template_text_block_ml_types",
        "name",
        "system_id"
    ),
    "message_template_text_block_ml_table": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "default_body_template",
        "default_body_template_length",
        "default_subject_template",
        "email_body_html_template",
        "email_subject_template",
        "end_timestamp",
        "gender_id",
        "identifier",
        "is_facebook_approved",
        "is_test_data",
        "lang_code",
        "message_template_category_id",
        "message_template_text_block_id",
        "message_template_text_block_ml_id",
        "number",
        "smartlink_type_id",
        "sms_body_template",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "whatsapp_body_template"
    ),
    "storage_table": (
        "activity_id",
        "activity_type_id",
        "created_timestamp",
        "created_user_id",
        "document_type_id",
        "end_timestamp",
        "entity_type_id",
        "file_extension_id",
        "file_type_id",
        "filename",
        "folder_id",
        "google_drive_document_id",
        "group_id",
        "group_list_id",
        "height",
        "html_map_tag",
        "is_ai_approved",
        "is_blur_not_direct_connect",
        "is_contacts_sheet",
        "is_google_sheet",
        "is_inbound_sync",
        "is_manual_approved",
        "is_outbound_sync",
        "is_test_data",
        "label_id",
        "label_list_id",
        "path",
        "ratio_x",
        "ration_y",
        "region",
        "sector_id",
        "smart_datastore_json",
        "start_timestamp",
        "storage_id",
        "storage_type_id",
        "table_definition_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external_id",
        "visibility_id",
        "width"
    ),
    "repo_table": (
        "category",
        "is_package",
        "is_separete_publish_gpr_job",
        "is_serverless_com",
        "is_using_our_own_serverless_graphql_server",
        "lastest_serverless_framework_version",
        "main_developer_profile_id",
        "name",
        "number_of_modules",
        "programming_langage_id",
        "programming_language_name",
        "repo_id",
        "serverless_domain_manager_status",
        "type",
        "url"
    ),
    "cicd_workflow_table": (
        "cicd_engine",
        "cicd_workflow_id",
        "component_id",
        "filename",
        "path",
        "repo_id"
    ),
    "cicd_workflow_history_table": (
        "both_play1_dvlp1_in_same_yml",
        "cicd_workflow_history_id",
        "cicd_workflow_id",
        "component_version",
        "dependencies_json",
        "devDependencies_json",
        "environment_name",
        "eslint_version",
        "name",
        "node_version",
        "npm_version",
        "old_long_approach",
        "serverless_version",
        "status",
        "typescript_version",
        "using_reusable_workflows",
        "yarn_version"
    ),
    "cicd_workflow_history_general_view": (
        "dependencies_json",
        "devdependencies_json",
        "environment_name",
        "eslint_version",
        "node_version",
        "npm_version",
        "programming_language_name",
        "repo.name",
        "serverless_version",
        "status",
        "typescript_version",
        "yarn_version"
    ),
    "api_limit_table": (
        "api_limit_id",
        "api_type_id",
        "api_type_level_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "external_limit_code",
        "external_limit_name",
        "hard_limit_unit",
        "hard_limit_value",
        "is_test_data",
        "parent_api_limit_id",
        "profile_id",
        "region",
        "soft_limit_unit",
        "soft_limit_value",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url_to_see_the limit",
        "user_external_id"
    ),
    "api_custom_limit_table": (
        "api_custom_limit_id",
        "api_type_level_id",
        "is_enabled",
        "name",
        "value_varchar",
        "var_int"
    ),
    "api_type_level_table": (
        "api_type_id",
        "api_type_level_table",
        "cost_usd",
        "is_credit_card_required",
        "name"
    ),
    "secret_table": (
        "brand_id",
        "secret_id",
        "user_external_id"
    ),
    "api_type_parent_table": (
        "api_type_parent_id",
        "name"
    ),
    "api_type_parent__api_type_table": (
        "api_type_id",
        "api_type_parent__api_type_id",
        "api_type_parent_id",
        "probability"
    ),
    "api_type_version_table": (
        "api_type_id",
        "api_type_version_id",
        "endpoint_version",
        "is_approved",
        "openapi_3_0",
        "version"
    ),
    "file_keyword_table": (
        "entity_type_id",
        "file_keyword_id",
        "filename_keyword",
        "name"
    ),
    "expand_text_table": (
        "expand_text_id",
        "original_text",
        "target_text"
    ),
    "prompt_table": (
        "entity_attributes_json",
        "is_system_prompt",
        "is_user_prompt",
        "machine_learning_model_id",
        "output_json_schema",
        "output_json_schema_id",
        "prompt",
        "prompt_id",
        "prompt_llm",
        "prompt_type"
    ),
    "prompt_execution_table": (
        "conversation_id",
        "prompt_execution_id",
        "prompt_id",
        "system_prompt_id",
        "timestamp",
        "user_prompt_id"
    ),
    "agent_table": (
        "agent_id",
        "allow",
        "ans",
        "name",
        "not_allowed"
    ),
    "internet_domain_table": (
        "common_email_format_type_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "domain",
        "end_timestamp",
        "identifier",
        "internet_domain_id",
        "is_commercial_webmail",
        "is_email_assign_profile_organization_to_people",
        "is_test_data",
        "is_url_assign_profile_organization_to_people",
        "number",
        "organization_id",
        "profile_id",
        "profile_list_id",
        "start_timestamp",
        "top_level_domain",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "contact_phone_duplicates": (
        "clean_phone",
        "contact_ids",
        "duplicate_count"
    ),
    "contact_email_duplicates_per_profile": (
        "clean_email",
        "contact_ids",
        "duplicate_count",
        "main_profile_id"
    ),
    "contact_email_duplicates_per_owner": (
        "clean_email",
        "contact_ids",
        "duplicate_count",
        "owner_profile_id"
    ),
    "contact_email_duplicates_per_profile_v2": (
        "clean_email",
        "contact_ids",
        "duplicate_count",
        "owner_profile_id"
    ),
    "contact_email_duplicates_all": (
        "clean_email",
        "contact_ids",
        "duplicate_count",
        "owner_profile_id",
        "severity"
    ),
    "business_process_table": (
        "business_process_id",
        "business_process_type_id",
        "name"
    ),
    "business_process_type_table": (
        "business_process_type_id",
        "name"
    ),
    "logger_table": (
        "action_id",
        "activity",
        "activity_id",
        "api_type",
        "api_type_id",
        "business_process_id",
        "class_name",
        "client_ip_v4",
        "client_ip_v6",
        "component_category",
        "component_id",
        "component_name",
        "component_type",
        "compound_message",
        "computer_language",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "destination_emails",
        "developer_email",
        "developer_email_address",
        "end_datetime",
        "end_timestamp",
        "environment_name",
        "error_stack",
        "field_id",
        "field_value_new",
        "field_value_old",
        "filename",
        "form_id",
        "function_name",
        "group_id",
        "group_id1",
        "group_id2",
        "identifier",
        "is_assertion_error",
        "is_test_data",
        "json_version",
        "line_number",
        "locals_before_exception",
        "locals_before_exception_text",
        "location_id",
        "logger_id",
        "message",
        "message_id",
        "message_template_id",
        "message_template_text_block_id",
        "number",
        "package_json",
        "path",
        "payload",
        "process_id",
        "profile_id",
        "question_id",
        "reaction_id",
        "real_name",
        "recipient",
        "record",
        "relationship_type_id",
        "result",
        "return_code",
        "return_message",
        "returned_value",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "severity_id",
        "smartlink_id",
        "smartlink_identifier",
        "source_email",
        "sql_formatted",
        "sql_parameters",
        "sql_statement",
        "start_datetime",
        "start_timestamp",
        "state_id",
        "status_id",
        "stderr",
        "stdout",
        "testing_framework",
        "text_block_id",
        "thread_id",
        "timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_identifier",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "schema_table": (
        "microservice_id",
        "name",
        "schema_id",
        "version"
    ),
    "ai_tool_table": (
        "ai_tool_id",
        "name"
    ),
    "ai_agent__profile_table": (
        "ai_agent__profile_id",
        "ai_agent_id",
        "profile_id"
    ),
    "ai_tool__profile_table": (
        "ai_tool__profile_id",
        "ai_tool_id",
        "profile_id"
    ),
    "file_type__storage_table": (
        "file_type__storage_id",
        "file_type_id",
        "is_created_by_ai",
        "is_created_by_human",
        "probability",
        "storage_id"
    ),
    "brand__prompt_table": (
        "brand__prompt_id",
        "brand__prompt_type_id",
        "brand_id",
        "prompt_id"
    ),
    "brand__prompt_type_table": (
        "brand__prompt_type_id",
        "name"
    ),
    "brand__domain_table": (
        "brand__domain_id",
        "brand_id",
        "domain_id"
    ),
    "event__prompt_table": (
        "event__prompt_id",
        "event_id",
        "prompt_id"
    ),
    "contact__prompt_table": (
        "contact__prompt_id",
        "contact_id",
        "prompt_id"
    ),
    "logger_with_deleted_and_test_data_view": (
        "action_id",
        "activity",
        "activity_id",
        "api_type",
        "api_type_id",
        "business_process_id",
        "class_name",
        "client_ip_v4",
        "client_ip_v6",
        "component_category",
        "component_id",
        "component_name",
        "component_type",
        "compound_message",
        "computer_language",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "destination_emails",
        "developer_email",
        "developer_email_address",
        "end_datetime",
        "end_timestamp",
        "environment_name",
        "error_stack",
        "field_id",
        "field_value_new",
        "field_value_old",
        "filename",
        "form_id",
        "function_name",
        "group_id",
        "group_id1",
        "group_id2",
        "identifier",
        "is_assertion_error",
        "is_test_data",
        "json_version",
        "line_number",
        "locals_before_exception",
        "locals_before_exception_text",
        "location_id",
        "logger_id",
        "message",
        "message_id",
        "message_template_id",
        "message_template_text_block_id",
        "number",
        "package_json",
        "path",
        "payload",
        "process_id",
        "profile_id",
        "question_id",
        "reaction_id",
        "real_name",
        "recipient",
        "record",
        "relationship_type_id",
        "result",
        "return_code",
        "return_message",
        "returned_value",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "severity_id",
        "smartlink_id",
        "smartlink_identifier",
        "source_email",
        "sql_formatted",
        "sql_parameters",
        "sql_statement",
        "start_datetime",
        "start_timestamp",
        "state_id",
        "status_id",
        "stderr",
        "stdout",
        "testing_framework",
        "text_block_id",
        "thread_id",
        "timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "user_id",
        "user_identifier",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "rfp_referral_with_deleted_and_test_data_view": (
        "buyer_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "identifier",
        "is_test_data",
        "message_id",
        "number",
        "press_link_in_content_message_timestamp",
        "rfp_id",
        "rfp_referral_id",
        "seller_approach_buyer_using_message_to_email_timestamp",
        "seller_approach_buyer_using_message_to_phone_timestamp",
        "seller_exposed_buyer_address_timestmap",
        "seller_exposed_buyer_email_address_timestmap",
        "seller_exposed_buyer_phone_timestmap",
        "seller_exposed_rfp_details_timestmap",
        "seller_id",
        "seller_interested_in_rfp_timestamp",
        "seller_interested_that_buyer_will_approach_seller_timestamp",
        "seller_interested_to_approach_buyer_timestamp",
        "smartlink_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id"
    ),
    "campaign_traction_with_deleted_and_test_data_view": (
        "campaign_id",
        "campaign_traction_id",
        "contact_id",
        "is_exposed_deal_details",
        "is_test_data",
        "message_id",
        "person_id",
        "smartlink_id"
    ),
    "advertisement_group_with_deleted_and_test_data_view": (
        "advertisement_id",
        "group_id",
        "id",
        "is_test_data"
    ),
    "contact_with_deleted_and_test_data_view": (
        "account_name",
        "additional_name",
        "address1_city",
        "address1_country",
        "address1_postal_code",
        "address1_state",
        "address1_street",
        "address2_city",
        "address2_country",
        "address2_postal_code",
        "address2_state",
        "address2_street",
        "anniversary",
        "birthday",
        "cira",
        "contact_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "day",
        "department",
        "display_as",
        "email1",
        "email2",
        "email3",
        "end_timestamp",
        "first_name",
        "full_name",
        "identifier",
        "is_test_data",
        "job_title",
        "job_title_id",
        "job_title_ml_id",
        "last_name",
        "last_sync_timestamp",
        "main_profile_id",
        "month",
        "name_prefix",
        "name_suffix",
        "nickname",
        "notes",
        "number",
        "organization",
        "organization_profile_id",
        "original_first_name",
        "original_last_name",
        "owner_profile_id",
        "person_id",
        "phone1",
        "phone2",
        "phone3",
        "photo_file_name",
        "photo_url",
        "source",
        "start_timestamp",
        "title",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "uploaded_timestamp",
        "visibility_id",
        "website1",
        "website2",
        "website3",
        "year"
    ),
    "importer_with_deleted_and_test_data_view": (
        "created_timestamp",
        "created_user_id",
        "data_imported_json",
        "data_source_external_user_id_old",
        "data_source_instance_id",
        "data_source_type_id_old",
        "end_timestamp",
        "entity_id",
        "entity_type_id",
        "google_calendar_id",
        "google_people_api_resource_name",
        "importer_id",
        "is_test_data",
        "number",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "user_external_country_id_old",
        "user_external_id"
    ),
    "campaign_traction_table": (
        "campaign_id",
        "campaign_traction_id",
        "contact_id",
        "is_exposed_deal_details",
        "is_test_data",
        "message_id",
        "person_id",
        "smartlink_id"
    ),
    "computer_vision_storage_with_deleted_and_test_data_view": (
        "angry",
        "black",
        "computer_vision_storage_id",
        "coordinate",
        "created_timestamp",
        "disgust",
        "dominant_emotion",
        "dominant_gender",
        "end_timestamp",
        "fear",
        "gender_id",
        "gender_man",
        "gender_women",
        "happy",
        "is_test_data",
        "json",
        "latino_hispanic",
        "location_id",
        "max_age",
        "middle eastern",
        "min_age",
        "profile_id",
        "region",
        "sad",
        "start_timestamp",
        "storage_id",
        "surprise",
        "time_end",
        "time_start",
        "updated_timestamp",
        "white",
        "x1",
        "x2",
        "y1",
        "y2"
    ),
    "advertisement_group_table": (
        "advertisement_id",
        "group_id",
        "id",
        "is_test_data"
    ),
    "computer_vision_storage_table": (
        "angry",
        "black",
        "computer_vision_storage_id",
        "coordinate",
        "created_timestamp",
        "disgust",
        "dominant_emotion",
        "dominant_gender",
        "end_timestamp",
        "fear",
        "gender_id",
        "gender_man",
        "gender_women",
        "happy",
        "is_test_data",
        "json",
        "latino_hispanic",
        "location_id",
        "max_age",
        "middle eastern",
        "min_age",
        "profile_id",
        "region",
        "sad",
        "start_timestamp",
        "storage_id",
        "surprise",
        "time_end",
        "time_start",
        "updated_timestamp",
        "white",
        "x1",
        "x2",
        "y1",
        "y2"
    ),
    "message_inforu_with_deleted_and_test_data_view": (
        "inforu_application_id",
        "inforu_channel",
        "inforu_customer_id",
        "inforu_customer_param",
        "inforu_incoming_json",
        "inforu_keyword",
        "inforu_mo_session_id",
        "inforu_network",
        "inforu_project_id",
        "inforu_short_code",
        "inforu_type",
        "inforu_value",
        "is_test_data",
        "message_id"
    ),
    "criteria_with_deleted_and_test_data_view": (
        "coordinate",
        "country_id_old",
        "county_id_old",
        "created_timestamp",
        "created_user_id",
        "criteria_id",
        "end_timestamp",
        "entity_type_id",
        "gender_list_id_old",
        "group_id",
        "group_list_id",
        "international_code",
        "internet_domain_id",
        "internet_domain_list_id",
        "is_test_data",
        "job_group_list_id",
        "job_location_list_id",
        "job_types_id",
        "label_id",
        "location_id",
        "location_list_id",
        "location_type_id",
        "max_age_old",
        "max_height_old",
        "max_kids_age_old",
        "max_number_of_kids_old",
        "min_age_old",
        "min_height_old",
        "min_kids_age_old",
        "min_number_of_kids_old",
        "name",
        "number_of_partners_old",
        "organization_name",
        "partner_experience_level_old",
        "profile_list_id",
        "radius",
        "radius_km",
        "radius_measure",
        "start_date_type_id",
        "start_timestamp",
        "template_condition",
        "updated_timestamp",
        "updated_user_id",
        "user_external_id",
        "vacancy_list_id",
        "visibility_id",
        "where_sql",
        "workplace_profile_list_id"
    ),
    "event_with_deleted_and_test_data_view": (
        "access_start_timestamp",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "data_source_instance_id",
        "description",
        "end_timestamp",
        "event_id",
        "facebook_event_url",
        "genre",
        "identifier",
        "is_beakfast",
        "is_can_be_paid_at_the_entrance_in_cash",
        "is_dinner",
        "is_event_approved",
        "is_event_has_promoters",
        "is_family",
        "is_lunch",
        "is_paid",
        "is_require_end_user_arrival_confirmation",
        "is_require_end_user_registration",
        "is_require_organizer_registration_confirmation",
        "is_show_end_timestamp",
        "is_test_data",
        "is_waitinglist",
        "location_id",
        "location_id_temp",
        "meetup_event_url",
        "name",
        "new_event_id",
        "number",
        "organizers_profile_id",
        "parent_event_id",
        "registration_url",
        "segement",
        "start_timestamp",
        "sub_genre",
        "sub_type",
        "timezone",
        "type",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id",
        "website_url"
    ),
    "form_transaction_with_deleted_and_test_data_view": (
        "action_id",
        "actual_channel_id",
        "actual_sent_timestamp",
        "bcc_email",
        "bcc_profile_id",
        "bcc_profile_list_id",
        "body",
        "campaign_id",
        "cc_email",
        "cc_profile_id",
        "cc_profile_list_id",
        "class_parameters_json",
        "compound_form",
        "compound_form_json",
        "compound_form_json_version",
        "created_timestamp",
        "created_user_id",
        "email_form_id",
        "end_timestamp",
        "form_id",
        "form_transaction_id",
        "form_type_id",
        "function_parameters_json",
        "group_id",
        "group_list_id",
        "is_gen_ai",
        "is_moderator_approved",
        "is_require_moderator",
        "is_test_data",
        "location_id",
        "machine_learning_model_id",
        "mailbox_id",
        "message_channel_id",
        "message_status_id",
        "moderator_feedback_text",
        "moderator_feedback_type",
        "moderator_profile_id",
        "number",
        "occurrence_id",
        "pin",
        "previous_form_id",
        "process_id",
        "profile_id",
        "provider_id",
        "push_notification",
        "queue_status_id",
        "request_to_send_timestamp",
        "requested_channel_id",
        "return_code",
        "return_form",
        "returned_value",
        "scheduled_sent_timestamp",
        "sender_email",
        "sender_profile_id",
        "server_ip_v4",
        "server_ip_v6",
        "session",
        "session_id",
        "start_timestamp",
        "stderr",
        "stdout",
        "subject",
        "thread_id",
        "to_email",
        "to_profile_id",
        "to_profile_list_id",
        "updated_timestamp",
        "updated_user_id",
        "user_jwt",
        "visibility_id"
    ),
    "label_with_deleted_and_test_data_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_approved",
        "is_highlight",
        "is_test_data",
        "label_id",
        "location_id",
        "name",
        "name_approved",
        "number",
        "parent_label_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "label_location_with_deleted_and_test_data_view": (
        "is_approved",
        "is_test_data",
        "label_id",
        "label_location_id",
        "label_ml_id",
        "location_id",
        "location_ml_id"
    ),
    "label_location_table": (
        "is_approved",
        "is_test_data",
        "label_id",
        "label_location_id",
        "label_ml_id",
        "location_id",
        "location_ml_id"
    ),
    "location_list_member_with_deleted_and_test_data_view": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "location_id",
        "location_list_id",
        "location_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_list_member_table": (
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "location_id",
        "location_list_id",
        "location_list_member_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "location_profile_with_deleted_and_test_data_view": (
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "end_timestamp",
        "is_test_data",
        "location_id",
        "location_profile_id",
        "location_type_id",
        "profile_id",
        "start_timestamp",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "visibility_id"
    ),
    "resource_with_deleted_and_test_data_view": (
        "buy_price_set",
        "buy_price_usd",
        "is_test_data",
        "location_id",
        "resource_id",
        "resource_type_id",
        "sell_price_set",
        "sell_price_usd"
    ),
    "resource_table": (
        "buy_price_set",
        "buy_price_usd",
        "is_test_data",
        "location_id",
        "resource_id",
        "resource_type_id",
        "sell_price_set",
        "sell_price_usd"
    ),
    "service_order_table": (
        "created_timestamp",
        "created_user_id",
        "currency_id",
        "end_timestamp",
        "group_supplier_id",
        "is_test_data",
        "location_id",
        "order_time",
        "price",
        "service_offer_id",
        "service_order_id",
        "service_provider_id",
        "service_receiver_id",
        "start_timestamp",
        "updated_timestamp",
        "updated_user_id"
    ),
    "action_item__text_block_table": (
        "action_item__text_block_id",
        "action_item_id",
        "text_block_id"
    ),
    "action_item__code_table": (
        "action_item__code_id",
        "action_item_id",
        "filename",
        "line_number_from",
        "line_number_until",
        "organization",
        "path",
        "source_control",
        "text"
    ),
    "ai_model_evaluation_table": (
        "ai_model_evaluation_id",
        "ai_model_instance_id",
        "result",
        "test_dataset_id"
    ),
    "ai_model_instance_table": (
        "ai_model_instance_id",
        "chunk_size",
        "chunk_unit",
        "data_type_id",
        "embedding_model_id",
        "generative_model_id",
        "vector_database_id"
    ),
    "ai_model_table": (
        "ai_model_id",
        "name"
    ),
    "text_block_table": (
        "ai_model_id",
        "created_effective_profile_id",
        "created_effective_user_id",
        "created_real_user_id",
        "created_timestamp",
        "created_user_id",
        "effective_timestamp",
        "end_timestamp",
        "fields_extracted_json",
        "identifier",
        "is_test_data",
        "number",
        "profile_id",
        "prompt_id",
        "seq",
        "start_timestamp",
        "system_entity_id",
        "system_id",
        "text",
        "text_block_id",
        "text_block_type_id",
        "text_expanded",
        "text_hash",
        "text_without_empty_lines",
        "text_without_empty_lines_hash",
        "updated_effective_profile_id",
        "updated_effective_user_id",
        "updated_real_user_id",
        "updated_timestamp",
        "updated_user_id",
        "url",
        "visibility_id"
    ),
    "action_item__variable_table": (
        "action_item__variable_id",
        "action_item_id",
        "variable_id",
        "variable_value_new",
        "variable_value_old"
    ),
    "action_item__field_table": (
        "action_item__field_id",
        "action_item_id",
        "field_id",
        "field_value_new",
        "field_value_old"
    ),
    "country_table": (
        "continent_id",
        "coordinate",
        "country_id",
        "created_timestamp",
        "created_user_id",
        "group_id",
        "identifier",
        "is_test_data",
        "iso",
        "iso3",
        "name",
        "nicename",
        "numcode",
        "phonecode",
        "updated_timestamp",
        "updated_user_id"
    ),
    "database_schema__repo_table": (
        "database_schema__repo_id",
        "database_schema_id",
        "database_schema_name_temp",
        "is_master_data",
        "repo_id",
        "repo_name_temp"
    ),
    "database_schema__repo_view": (
        "database_schema__repo_id",
        "database_schema_id",
        "database_schema_name_temp",
        "is_master_data",
        "repo_id",
        "repo_name_temp"
    ),
    "database_schema__repo_general_view": (
        "database_schema_name_temp",
        "repo_name",
        "repo_name_temp",
        "schema_name"
    )
}
