# Generated by SQL2Code (circles-zone/sql2code-local-python-packge/sql_to_code/src/sqltocode.py running from GHA .github/workflows using sql2code-command-line parameter and calling github-workflows/.github/workflows)


# args=Namespace(schema='database', table='table_definition_table', columns=None, groupby='table_name', orderby='table_name', include_nulls='true', language='Python', format='dictionary', output_path='/repos/circlez/database-mysql-local-python-package/database-mysql-local-python-package/database_mysql_local/src/')
# 01/03/2026, 00:19:41
# Python code as a dictionary

table_definition = {
  "accommodation_deal_table": {
    "table_definition_id": 157,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "accommodation_table": {
    "table_definition_id": 156,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "account_profile_table": {
    "table_definition_id": 85,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_category_ml_table": {
    "table_definition_id": 537,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_category_table": {
    "table_definition_id": 536,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_contact_table": {
    "table_definition_id": 455,
    "entity_type_id1": 29,
    "entity_type_id2": 7,
    "schema": "action_item_contact",
    "view_name": "action_item_contact_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_ml_table": {
    "table_definition_id": 491,
    "entity_type_id1": 29,
    "entity_type_id2": None,
    "schema": "action_item",
    "view_name": "action_item_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": 1,
    "is_main_column": 1,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_priority_table": {
    "table_definition_id": 493,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item",
    "view_name": "action_item_priority_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_profile_table": {
    "table_definition_id": 454,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_table": {
    "table_definition_id": 453,
    "entity_type_id1": 29,
    "entity_type_id2": None,
    "schema": "action_item",
    "view_name": "action_item_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_type_ml_table": {
    "table_definition_id": 534,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_item_type_table": {
    "table_definition_id": 535,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_item_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_star_subscription_table": {
    "table_definition_id": 318,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action_star_subscription",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "action_table": {
    "table_definition_id": 410,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "action",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": 0,
    "is_user_data_whole_table": 0,
    "end_timestamp": None
  },
  "activity_activity_table": {
    "table_definition_id": 421,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_group_location_occurrence_profile_ml": {
    "table_definition_id": 86,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity_group_location_occurrence_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_group_location_occurrence_profile_ta": {
    "table_definition_id": 79,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity_group_location_occurrence_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_list_member_table": {
    "table_definition_id": 73,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_list_table": {
    "table_definition_id": 261,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_ml_table": {
    "table_definition_id": 492,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_ml_table_old": {
    "table_definition_id": 260,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_star_subscription_table": {
    "table_definition_id": 262,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity_star_subscription",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_table": {
    "table_definition_id": 504,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_table_old": {
    "table_definition_id": 503,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_trigger_event": {
    "table_definition_id": 62,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "trigger_event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_type_ml_table": {
    "table_definition_id": 374,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "activity_type_table": {
    "table_definition_id": 373,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "activity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_campaign_table": {
    "table_definition_id": 162,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement_campaign",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_costs_table": {
    "table_definition_id": 159,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_group_list_table": {
    "table_definition_id": 113,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement_group_list",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_group_table": {
    "table_definition_id": 112,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_message_template_table": {
    "table_definition_id": 216,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_profile": {
    "table_definition_id": 111,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_storage_table": {
    "table_definition_id": 217,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_table": {
    "table_definition_id": 158,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_type_ml_table": {
    "table_definition_id": 160,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "advertisement_type_table": {
    "table_definition_id": 161,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "advertisement",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "age_group_group_table": {
    "table_definition_id": 145,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "age_group_table": {
    "table_definition_id": 144,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "age_media": {
    "table_definition_id": 54,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_budget_table": {
    "table_definition_id": 274,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_budget",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_call_table": {
    "table_definition_id": 35,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_call",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": 1,
    "sensative_columns": "outgoing_body",
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_limit_table": {
    "table_definition_id": 285,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_limit",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_parameter_table": {
    "table_definition_id": 268,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_price_table": {
    "table_definition_id": 270,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_price",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_type_field_table": {
    "table_definition_id": 176,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_type_table": {
    "table_definition_id": 415,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "api_type_user_external_table": {
    "table_definition_id": 308,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "api_type_user_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "avatar_table": {
    "table_definition_id": 409,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "avatar",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "base_entity_table": {
    "table_definition_id": 213,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_group_table": {
    "table_definition_id": 435,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_location_table": {
    "table_definition_id": 406,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_ml_table": {
    "table_definition_id": 80,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_table": {
    "table_definition_id": 412,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_type_table": {
    "table_definition_id": 413,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "brand_url_table": {
    "table_definition_id": 338,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "brand_url",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "buyer_ml_table": {
    "table_definition_id": 198,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "buyer_table": {
    "table_definition_id": 376,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_ml_table": {
    "table_definition_id": 495,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_profile_person_table": {
    "table_definition_id": 502,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar_item_profile_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_resource_table": {
    "table_definition_id": 500,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar_item_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_table": {
    "table_definition_id": 515,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_type_ml_table": {
    "table_definition_id": 513,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar_item_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_item_type_table": {
    "table_definition_id": 514,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar_item_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_ml_table": {
    "table_definition_id": 498,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_profile_table": {
    "table_definition_id": 501,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "calendar_table": {
    "table_definition_id": 505,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "calendar",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_criteria_table": {
    "table_definition_id": 307,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_criteria",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_group_list_table": {
    "table_definition_id": 116,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_group_list",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_group_table": {
    "table_definition_id": 115,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_message_template_table": {
    "table_definition_id": 372,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_ml_table": {
    "table_definition_id": 289,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_smartlink_table": {
    "table_definition_id": 297,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_smartlink",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_status_ml_table": {
    "table_definition_id": 290,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_status_table": {
    "table_definition_id": 288,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_table": {
    "table_definition_id": 314,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "campaign_traction_table": {
    "table_definition_id": 382,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "campaign_traction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "channel_provider_table": {
    "table_definition_id": 184,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "checklist_group_table": {
    "table_definition_id": 106,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "checklist",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "checklist_table": {
    "table_definition_id": 61,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "checklist",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "checklists_templates_table": {
    "table_definition_id": 68,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "checklist",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "cicd_workflow_history_table": {
    "table_definition_id": 460,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "cicd",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "cicd_workflow_table": {
    "table_definition_id": 459,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "cicd",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "circlez_team_members": {
    "table_definition_id": 48,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "city_ml_table": {
    "table_definition_id": 275,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "city_table": {
    "table_definition_id": 23,
    "entity_type_id1": 38,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "cloud_region_table": {
    "table_definition_id": 371,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "master",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "cloud_vendor_table": {
    "table_definition_id": 370,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "master",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "community_waiting_list_table": {
    "table_definition_id": 143,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "community",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "component_table": {
    "table_definition_id": 316,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "component",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "computer_vision_storage_table": {
    "table_definition_id": 197,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "computer_vision_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "consumer_category_table": {
    "table_definition_id": 173,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_email_address_table": {
    "table_definition_id": 37,
    "entity_type_id1": 7,
    "entity_type_id2": 23,
    "schema": "contact_email_address",
    "view_name": "contact_email_address_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_group_table": {
    "table_definition_id": 485,
    "entity_type_id1": 7,
    "entity_type_id2": 11,
    "schema": "contact_group",
    "view_name": "contact_group_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_group_type_ml_table": {
    "table_definition_id": 194,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_group_type_table": {
    "table_definition_id": 172,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_importer_table": {
    "table_definition_id": 362,
    "entity_type_id1": 7,
    "entity_type_id2": 35,
    "schema": "contact_importer",
    "view_name": "contact_importer_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_internet_domain_table": {
    "table_definition_id": 471,
    "entity_type_id1": 7,
    "entity_type_id2": 31,
    "schema": "contact_internet_domain",
    "view_name": "contact_internet_domain_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_location_table": {
    "table_definition_id": 518,
    "entity_type_id1": 7,
    "entity_type_id2": 28,
    "schema": "contact_location",
    "view_name": "contact_location_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_location_type_ml_table": {
    "table_definition_id": 517,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_location_type_table": {
    "table_definition_id": 516,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact_location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_ml_table": {
    "table_definition_id": 452,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_note_contact_table_old": {
    "table_definition_id": 392,
    "entity_type_id1": 32,
    "entity_type_id2": 7,
    "schema": "contact_note",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_note_group_table": {
    "table_definition_id": 473,
    "entity_type_id1": 32,
    "entity_type_id2": 11,
    "schema": "contact_note",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_note_table": {
    "table_definition_id": 393,
    "entity_type_id1": 32,
    "entity_type_id2": None,
    "schema": "contact_note",
    "view_name": "contact_note_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_note_text_block_table": {
    "table_definition_id": 472,
    "entity_type_id1": 32,
    "entity_type_id2": 30,
    "schema": "contact_note",
    "view_name": "contact_note_text_block_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_person_table": {
    "table_definition_id": 31,
    "entity_type_id1": 7,
    "entity_type_id2": 17,
    "schema": "contact_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": 0,
    "is_history": 0,
    "is_one_table_per_all_tenants": 0,
    "is_tree": 0,
    "is_graph": 0,
    "is_metrics": 0,
    "is_visibility": None,
    "is_pii": 0,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_person_type_table": {
    "table_definition_id": 532,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_phone_table": {
    "table_definition_id": 474,
    "entity_type_id1": 7,
    "entity_type_id2": 26,
    "schema": "contact_phone",
    "view_name": "contact_phone_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": 1,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_profile_table": {
    "table_definition_id": 32,
    "entity_type_id1": 7,
    "entity_type_id2": 21,
    "schema": "contact_profile",
    "view_name": "contact_profile_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_reaction_table": {
    "table_definition_id": 400,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "contact_reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_table": {
    "table_definition_id": 30,
    "entity_type_id1": 7,
    "entity_type_id2": None,
    "schema": "contact",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": 1,
    "is_history": 0,
    "is_one_table_per_all_tenants": 0,
    "is_tree": 0,
    "is_graph": 0,
    "is_metrics": 0,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_text_block_table": {
    "table_definition_id": 509,
    "entity_type_id1": 7,
    "entity_type_id2": 30,
    "schema": "contact_text_block",
    "view_name": "contact_text_block_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "contact_user_external_table": {
    "table_definition_id": 475,
    "entity_type_id1": 7,
    "entity_type_id2": 33,
    "schema": "contact_user_external",
    "view_name": "contact_user_external_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "country_ml_table": {
    "table_definition_id": 326,
    "entity_type_id1": 36,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "country_table": {
    "table_definition_id": 317,
    "entity_type_id1": 36,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "county_ml_table": {
    "table_definition_id": 327,
    "entity_type_id1": 37,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "county_table": {
    "table_definition_id": 323,
    "entity_type_id1": 37,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_group_table": {
    "table_definition_id": 339,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "criteria_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_promotional_message_to_person_table": {
    "table_definition_id": 196,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "criteria_promotional_message_to_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_realestate": {
    "table_definition_id": 133,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods_realestate",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_set_segment_table": {
    "table_definition_id": 450,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "criteria",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_table": {
    "table_definition_id": 451,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "criteria",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "criteria_vacancy_table": {
    "table_definition_id": 204,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "currency_ml_table": {
    "table_definition_id": 273,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "currency",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "currency_table": {
    "table_definition_id": 272,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "currency",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_external_user_ml_table": {
    "table_definition_id": 428,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source_external_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_external_user_table": {
    "table_definition_id": 432,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source_external_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_field_table": {
    "table_definition_id": 137,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source_field",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_ml_table": {
    "table_definition_id": 431,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": "I think we renamed it to data_source_type_table",
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_table": {
    "table_definition_id": 430,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": "I think we renamed it to data_source_type_table",
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_type_ml_table": {
    "table_definition_id": 429,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "data_source_type_table": {
    "table_definition_id": 427,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "data_source",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": 0,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "database_engine_table": {
    "table_definition_id": 351,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database_engine",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "deployment": {
    "table_definition_id": 51,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "deployment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "deployment_status": {
    "table_definition_id": 50,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "deployment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "device_table": {
    "table_definition_id": 84,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "device_user_external_table": {
    "table_definition_id": 363,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "device_user_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "dialog_workflow_script_ml_table": {
    "table_definition_id": 480,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "dialog_workflow_script_table": {
    "table_definition_id": 340,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "dialog_workflow_state_history_table_old": {
    "table_definition_id": 53,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": 0,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "dialog_workflow_state_ml_table": {
    "table_definition_id": 481,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "dialog_workflow_state_table": {
    "table_definition_id": 367,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_account_mailbox_table": {
    "table_definition_id": 528,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_account_mailbox",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_account_table": {
    "table_definition_id": 527,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_account",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_account_user_external_table": {
    "table_definition_id": 530,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_account_user_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_label_table": {
    "table_definition_id": 468,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_ml_table": {
    "table_definition_id": 36,
    "entity_type_id1": 23,
    "entity_type_id2": None,
    "schema": "email_address",
    "view_name": "email_address_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_person_table": {
    "table_definition_id": 436,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_address_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_profile_table": {
    "table_definition_id": 3,
    "entity_type_id1": 23,
    "entity_type_id2": 21,
    "schema": "email_address_profile",
    "view_name": "email_address_profile_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": 1,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_profile_verification_table": {
    "table_definition_id": 200,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_address_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_table": {
    "table_definition_id": 4,
    "entity_type_id1": 23,
    "entity_type_id2": None,
    "schema": "email_address",
    "view_name": "email_address_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": 37,
    "profile_mapping_table_id": 3,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": 1,
    "sensative_columns": "email_address",
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": 1,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "email_address_text_block_table": {
    "table_definition_id": 365,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "email_address_text_block",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "employer_profile_profile_table": {
    "table_definition_id": 191,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "employer",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "employer_profile_table": {
    "table_definition_id": 192,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "employer",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "employer_profile_vacancy_table": {
    "table_definition_id": 193,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "employer",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "entity_moderation_history_table": {
    "table_definition_id": 303,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "entity_moderation",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "entity_moderation_table": {
    "table_definition_id": 304,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "entity_moderation",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "entity_type_ml_table": {
    "table_definition_id": 433,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "entity_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "entity_type_table": {
    "table_definition_id": 306,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "entity_type",
    "view_name": "entity_type_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "environment": {
    "table_definition_id": 49,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "deployment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "environment_ml_table": {
    "table_definition_id": 186,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "environment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "environment_repo_table": {
    "table_definition_id": 352,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "environment_repo",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "environment_table": {
    "table_definition_id": 185,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "environment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "environnment_sql_report_table": {
    "table_definition_id": 425,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sql_report",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "error_table": {
    "table_definition_id": 407,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "error",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_event_external_table": {
    "table_definition_id": 296,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_event_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_external_table": {
    "table_definition_id": 357,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_group_table": {
    "table_definition_id": 334,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_hashag_table": {
    "table_definition_id": 496,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_location_table": {
    "table_definition_id": 286,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_ml_table": {
    "table_definition_id": 341,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_news_table": {
    "table_definition_id": 403,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "news",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_participant_table": {
    "table_definition_id": 494,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_price_ml_table": {
    "table_definition_id": 164,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_price_table": {
    "table_definition_id": 163,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_profile_email_table": {
    "table_definition_id": 405,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_profile_table": {
    "table_definition_id": 442,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_profile_type_ml_table": {
    "table_definition_id": 441,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_profile_type_table": {
    "table_definition_id": 440,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_status_ml_table": {
    "table_definition_id": 279,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_status_table": {
    "table_definition_id": 278,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_storage_table": {
    "table_definition_id": 282,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "event_table": {
    "table_definition_id": 19,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "field_alias_table": {
    "table_definition_id": 292,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "field_profile_text_block_table": {
    "table_definition_id": 523,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field_profile_text_block",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "field_table": {
    "table_definition_id": 520,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "field_text_block_type_table": {
    "table_definition_id": 521,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field_text_block_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "file_extension_table": {
    "table_definition_id": 138,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "file_type_ml_table": {
    "table_definition_id": 300,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "file_type_table": {
    "table_definition_id": 302,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "first_name_alias_table": {
    "table_definition_id": 322,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "name",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "first_name_ml_table": {
    "table_definition_id": 344,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "name",
    "view_name": "first_name_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": 1,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "first_name_table": {
    "table_definition_id": 321,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "name",
    "view_name": "first_name_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "form_message_template_table": {
    "table_definition_id": 458,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "form",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "form_table": {
    "table_definition_id": 457,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "form",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "form_transaction_table": {
    "table_definition_id": 478,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "form_transaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "form_type_table": {
    "table_definition_id": 477,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "form",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "gender_list_member_table": {
    "table_definition_id": 364,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "gender",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 0,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 1,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": 0,
    "is_sensative": 0,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "gender_list_table": {
    "table_definition_id": 506,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "gender",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 0,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 1,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": 0,
    "is_sensative": 0,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "gender_ml_table": {
    "table_definition_id": 179,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "gender",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 0,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 1,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": 0,
    "is_sensative": 0,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "gender_table": {
    "table_definition_id": 507,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "gender",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 0,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 1,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": 0,
    "is_sensative": 0,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "goods_ml_table": {
    "table_definition_id": 384,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "goods_table": {
    "table_definition_id": 383,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_category_ml_table": {
    "table_definition_id": 201,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_category_table": {
    "table_definition_id": 476,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_group_category_table": {
    "table_definition_id": 439,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_group_table": {
    "table_definition_id": 165,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_group_type_table": {
    "table_definition_id": 438,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_job_title_table": {
    "table_definition_id": 208,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_job_title",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_label_table": {
    "table_definition_id": 9,
    "entity_type_id1": 11,
    "entity_type_id2": 27,
    "schema": "group_label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_list_member_table": {
    "table_definition_id": 350,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_list_table": {
    "table_definition_id": 349,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_ml_table": {
    "table_definition_id": 524,
    "entity_type_id1": 11,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": 1,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_news_table": {
    "table_definition_id": 404,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_news",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_profile_table": {
    "table_definition_id": 7,
    "entity_type_id1": 11,
    "entity_type_id2": 21,
    "schema": "group_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_profile_type_ml_table": {
    "table_definition_id": 202,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_profile_type_table": {
    "table_definition_id": 203,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_recruitment_agency": {
    "table_definition_id": 168,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recruitment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_storage_blob_table": {
    "table_definition_id": 233,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_storage_table": {
    "table_definition_id": 234,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_table": {
    "table_definition_id": 8,
    "entity_type_id1": 11,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 7,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_text_block_table": {
    "table_definition_id": 294,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group_text_block",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "group_type_ml_table": {
    "table_definition_id": 447,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": 0,
    "is_user_data_whole_table": 0,
    "end_timestamp": None
  },
  "group_type_table": {
    "table_definition_id": 448,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "group",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": 0,
    "is_user_data_whole_table": 0,
    "end_timestamp": None
  },
  "identifier_table": {
    "table_definition_id": 319,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "identifier",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "identity_table": {
    "table_definition_id": 72,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "identity",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "import_contact_table": {
    "table_definition_id": 354,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "import_contact",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "importer_table": {
    "table_definition_id": 479,
    "entity_type_id1": 35,
    "entity_type_id2": None,
    "schema": "importer",
    "view_name": "importer_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "intent_table": {
    "table_definition_id": 269,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "intent",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "internet_domain_table": {
    "table_definition_id": 456,
    "entity_type_id1": 31,
    "entity_type_id2": None,
    "schema": "internet_domain",
    "view_name": "internet_domain_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "invoice_table": {
    "table_definition_id": 150,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "finance_accounts_receivable",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "item_profile_table": {
    "table_definition_id": 391,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "item_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "item_rendered_table": {
    "table_definition_id": 345,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "timeline",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "item_type_table": {
    "table_definition_id": 346,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "item_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "job_title_ml_table": {
    "table_definition_id": 487,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "job_title_table": {
    "table_definition_id": 486,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "job_title_vacancy_table": {
    "table_definition_id": 209,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "json_version_source_code_table": {
    "table_definition_id": 401,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "json_version",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "json_version_table": {
    "table_definition_id": 402,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "json_version",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_email_table_old": {
    "table_definition_id": 103,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_location_table": {
    "table_definition_id": 531,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label_location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_message_table": {
    "table_definition_id": 335,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label_message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_ml_table": {
    "table_definition_id": 470,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_phone_table": {
    "table_definition_id": 469,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_profile_table": {
    "table_definition_id": 132,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_question_table": {
    "table_definition_id": 129,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "label_table": {
    "table_definition_id": 228,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "label",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "lang_code_table": {
    "table_definition_id": 336,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "lang_code",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "language_user_table": {
    "table_definition_id": 263,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "language_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "last_name_ml_table": {
    "table_definition_id": 510,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "name",
    "view_name": "last_name_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": 1,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "last_name_table": {
    "table_definition_id": 525,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "name",
    "view_name": "last_name_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "lesson_table": {
    "table_definition_id": 142,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service_education",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "listings": {
    "table_definition_id": 108,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods_realestate",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_field_ml_table": {
    "table_definition_id": 328,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_list_member_table": {
    "table_definition_id": 74,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_list_table": {
    "table_definition_id": 69,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_ml_table": {
    "table_definition_id": 329,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_news_table": {
    "table_definition_id": 218,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "news",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_person_table": {
    "table_definition_id": 434,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_profile_ml_table": {
    "table_definition_id": 309,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location_profile",
    "view_name": "location_profile_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_profile_table": {
    "table_definition_id": 14,
    "entity_type_id1": 28,
    "entity_type_id2": 21,
    "schema": "location_profile",
    "view_name": "location_profile_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_table": {
    "table_definition_id": 13,
    "entity_type_id1": 28,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": "location_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 14,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_type_ml_table": {
    "table_definition_id": 311,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "location_type_table": {
    "table_definition_id": 310,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "logger_severity_level_table": {
    "table_definition_id": 114,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "logger_table": {
    "table_definition_id": 20,
    "entity_type_id1": 40,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": "logger_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 0,
    "is_ml_child_table": 0,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": 0,
    "is_history": 1,
    "is_one_table_per_all_tenants": 0,
    "is_tree": 1,
    "is_graph": 0,
    "is_metrics": None,
    "is_visibility": 0,
    "is_pii": 1,
    "is_sensative": 1,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "logger_table_old": {
    "table_definition_id": 71,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "login_history_table": {
    "table_definition_id": 25,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "login",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": 1,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "mailbox_profile_table": {
    "table_definition_id": 526,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "mailbox_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "mailbox_table": {
    "table_definition_id": 529,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "mailbox",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "media_type_ml_table": {
    "table_definition_id": 131,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "media_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "media_type_table": {
    "table_definition_id": 130,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "media_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_channel_ml_table": {
    "table_definition_id": 490,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_channel_provider_ml_table": {
    "table_definition_id": 284,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_channel_provider_table": {
    "table_definition_id": 283,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_channel_table": {
    "table_definition_id": 488,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_content_ml_table": {
    "table_definition_id": 247,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_content_table": {
    "table_definition_id": 248,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_outgoing_table_shall_we_delete_it": {
    "table_definition_id": 258,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_profile_table": {
    "table_definition_id": 347,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_read_table": {
    "table_definition_id": 246,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_sent_table": {
    "table_definition_id": 214,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_status_ml_table": {
    "table_definition_id": 245,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_status_table": {
    "table_definition_id": 242,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_table": {
    "table_definition_id": 21,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": 1,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_function_table": {
    "table_definition_id": 443,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_message_template_text_block_": {
    "table_definition_id": 348,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_rfp_table": {
    "table_definition_id": 375,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_rfp",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_table": {
    "table_definition_id": 215,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_text_block_ml_table": {
    "table_definition_id": 360,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_text_block_table": {
    "table_definition_id": 467,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_text_block_type_ml_table": {
    "table_definition_id": 361,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_template_text_block_type_table": {
    "table_definition_id": 463,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message_template",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_type_ml_table": {
    "table_definition_id": 243,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "message_type_table": {
    "table_definition_id": 244,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "message",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "mobile_account": {
    "table_definition_id": 77,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "mobile_account_profile": {
    "table_definition_id": 78,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "mobile_device": {
    "table_definition_id": 76,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "neighborhood_ml_table": {
    "table_definition_id": 330,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "neighborhood_table": {
    "table_definition_id": 22,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "news_ml_table": {
    "table_definition_id": 226,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "news",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "news_table": {
    "table_definition_id": 225,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "news",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "node_table": {
    "table_definition_id": 368,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "node",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "occurrence_list_occurrence_table": {
    "table_definition_id": 170,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "occurrence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "occurrence_list_table": {
    "table_definition_id": 169,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "occurrence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "occurrence_table": {
    "table_definition_id": 171,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "occurrence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "operational_hours_table": {
    "table_definition_id": 395,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "operational_hours",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "organization_ml_table": {
    "table_definition_id": 511,
    "entity_type_id1": 34,
    "entity_type_id2": None,
    "schema": "organization",
    "view_name": "organization_ml_not_deleted_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "organization_profile_table": {
    "table_definition_id": 298,
    "entity_type_id1": 34,
    "entity_type_id2": 21,
    "schema": "organization_profile",
    "view_name": "organization_profile_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "organization_table": {
    "table_definition_id": 512,
    "entity_type_id1": 34,
    "entity_type_id2": None,
    "schema": "organization",
    "view_name": "organization_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "os": {
    "table_definition_id": 75,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "os_table": {
    "table_definition_id": 83,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_bank_table": {
    "table_definition_id": 148,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_ml_table": {
    "table_definition_id": 155,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_table": {
    "table_definition_id": 152,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_template_bank_table": {
    "table_definition_id": 151,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_template_ml_table": {
    "table_definition_id": 154,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "package_template_table": {
    "table_definition_id": 153,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "package",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "participant_category_table": {
    "table_definition_id": 175,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "participant_table": {
    "table_definition_id": 497,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "permission_repo_table": {
    "table_definition_id": 420,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "repo",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_kid_table": {
    "table_definition_id": 287,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_ml_table": {
    "table_definition_id": 34,
    "entity_type_id1": 17,
    "entity_type_id2": None,
    "schema": "person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_phone_table": {
    "table_definition_id": 437,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "person_phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_profile_table": {
    "table_definition_id": 39,
    "entity_type_id1": 17,
    "entity_type_id2": 21,
    "schema": "person_profile",
    "view_name": "person_profile_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_table": {
    "table_definition_id": 17,
    "entity_type_id1": 17,
    "entity_type_id2": None,
    "schema": "person",
    "view_name": "person_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 39,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": 1,
    "name_of_single_value_column": "main_email_address",
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": 1,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "person_text_block_table": {
    "table_definition_id": 508,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "person_text_block",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "phone_ml_table": {
    "table_definition_id": 235,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "phone_profile_table": {
    "table_definition_id": 2,
    "entity_type_id1": 26,
    "entity_type_id2": 21,
    "schema": "phone_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": 1,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "phone_profile_verification_table": {
    "table_definition_id": 295,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "phone_table": {
    "table_definition_id": 1,
    "entity_type_id1": 26,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": "phone_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 2,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": 1,
    "sensative_columns": "number_original,local_number_normalized,full_",
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": 0,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": 1,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "phone_table_old": {
    "table_definition_id": 104,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "phone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "planned_database_schema_change_table": {
    "table_definition_id": 449,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "platform_invitation_table": {
    "table_definition_id": 466,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "platform_invitation",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "platform_table": {
    "table_definition_id": 320,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service_education",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "preference_table": {
    "table_definition_id": 265,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "presence": {
    "table_definition_id": 55,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "presence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "presence_history": {
    "table_definition_id": 56,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "presence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "presence_ml": {
    "table_definition_id": 57,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "presence",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "price_set_currency_table": {
    "table_definition_id": 271,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_price",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "price_set_range_table": {
    "table_definition_id": 276,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_price",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "price_set_table": {
    "table_definition_id": 238,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_price",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_deal_ml_table": {
    "table_definition_id": 107,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_deal_table": {
    "table_definition_id": 277,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_metrics_table": {
    "table_definition_id": 24,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "product_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": 1,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_metrics_type_ml_table": {
    "table_definition_id": 423,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "product_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_metrics_type_table": {
    "table_definition_id": 422,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "product_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_table": {
    "table_definition_id": 280,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "product_url_table": {
    "table_definition_id": 281,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_account_table": {
    "table_definition_id": 149,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "finance_accounts_receivable",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_category_ml_table": {
    "table_definition_id": 229,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_category_table": {
    "table_definition_id": 230,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_history_table": {
    "table_definition_id": 418,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_level_ml_table": {
    "table_definition_id": 127,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_level",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_level_table": {
    "table_definition_id": 538,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_level",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_list_member_table": {
    "table_definition_id": 419,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_list_profile_table": {
    "table_definition_id": 254,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_list_table": {
    "table_definition_id": 253,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_media_ml_table": {
    "table_definition_id": 257,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_media_table": {
    "table_definition_id": 256,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_metics_table": {
    "table_definition_id": 10,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_metics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_metrics_table": {
    "table_definition_id": 26,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": 1,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_metrics_type_ml_table": {
    "table_definition_id": 353,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_metrics_type_table": {
    "table_definition_id": 178,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_metrics",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_milestone_table": {
    "table_definition_id": 27,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_milestone",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": 1,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_ml_table": {
    "table_definition_id": 33,
    "entity_type_id1": 21,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_page_ml_table": {
    "table_definition_id": 356,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_page",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_page_table": {
    "table_definition_id": 355,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_page",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_preference_group_group_table": {
    "table_definition_id": 220,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_preference_group_ml_table": {
    "table_definition_id": 221,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_preference_group_table": {
    "table_definition_id": 219,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_preference_ml_table": {
    "table_definition_id": 222,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_preference_table": {
    "table_definition_id": 266,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_profile_level_table": {
    "table_definition_id": 134,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_profile_level",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_profile_role_ml_table": {
    "table_definition_id": 417,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_profile_role_table": {
    "table_definition_id": 416,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_profile_table": {
    "table_definition_id": 15,
    "entity_type_id1": 21,
    "entity_type_id2": 21,
    "schema": "profile_profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_question_answr_table": {
    "table_definition_id": 128,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_question",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_reaction_table": {
    "table_definition_id": 399,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_recruitment_agency": {
    "table_definition_id": 167,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recruitment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_storage_table": {
    "table_definition_id": 126,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_storage_type_ml_table": {
    "table_definition_id": 299,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_storage_type_table": {
    "table_definition_id": 301,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_table": {
    "table_definition_id": 16,
    "entity_type_id1": 21,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_text_block_table": {
    "table_definition_id": 539,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_text_block",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_type_ml_table": {
    "table_definition_id": 252,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_type_table": {
    "table_definition_id": 414,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_url_table": {
    "table_definition_id": 11,
    "entity_type_id1": 21,
    "entity_type_id2": 24,
    "schema": "profile_url",
    "view_name": "profile_url_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_user_external_table": {
    "table_definition_id": 259,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile_user_external",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "profile_user_table": {
    "table_definition_id": 6,
    "entity_type_id1": 21,
    "entity_type_id2": 25,
    "schema": "profile_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 0,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "programming_language_table": {
    "table_definition_id": 315,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "component",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "promotional_message_to_person_ml_table": {
    "table_definition_id": 195,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "promotional_message_to_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "promotional_message_to_person_table": {
    "table_definition_id": 312,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "promotional_message_to_person",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "proxy_group_list_table": {
    "table_definition_id": 105,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "proxy_profile": {
    "table_definition_id": 102,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "question_ml_table": {
    "table_definition_id": 482,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "question",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "question_possible_answers_ml_table": {
    "table_definition_id": 359,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "question",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "question_table": {
    "table_definition_id": 464,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "question",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "question_type_table": {
    "table_definition_id": 462,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "question",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "queue_item_table": {
    "table_definition_id": 333,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "queue",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "queue_status_ml_table": {
    "table_definition_id": 267,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "queue",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_category_ml_table": {
    "table_definition_id": 140,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_category_table": {
    "table_definition_id": 87,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_ml_table": {
    "table_definition_id": 398,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_reaction_category_table": {
    "table_definition_id": 88,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_table": {
    "table_definition_id": 397,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_type_ml_table": {
    "table_definition_id": 396,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "reaction_type_table": {
    "table_definition_id": 139,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "reaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "recommendation_model_table": {
    "table_definition_id": 390,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recommendation_model",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "recruitment_agency_table": {
    "table_definition_id": 166,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recruitment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "recruitment_rfp_table": {
    "table_definition_id": 446,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recruitment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "region_ml_table": {
    "table_definition_id": 331,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "region_table": {
    "table_definition_id": 324,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "relationship_classification_table_old": {
    "table_definition_id": 177,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "profile",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "relationship_type_classification_ml_table": {
    "table_definition_id": 251,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "relationship_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "relationship_type_classification_table": {
    "table_definition_id": 250,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "relationship_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "relationship_type_ml_table": {
    "table_definition_id": 249,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "relationship_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "relationship_type_table": {
    "table_definition_id": 484,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "relationship_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "repo_dependency_table": {
    "table_definition_id": 388,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "repo",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "repo_table": {
    "table_definition_id": 408,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "repo",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_instance_ml_table": {
    "table_definition_id": 121,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_instance_table": {
    "table_definition_id": 120,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_ml_table": {
    "table_definition_id": 117,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_table": {
    "table_definition_id": 499,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_type_ml_table": {
    "table_definition_id": 118,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "resource_type_table": {
    "table_definition_id": 119,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_resource",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "rfp_ml_table": {
    "table_definition_id": 380,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_rfp",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "rfp_referral_table": {
    "table_definition_id": 387,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_rfp",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "rfp_requirements_table": {
    "table_definition_id": 381,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_rfp",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "rfp_table": {
    "table_definition_id": 385,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_rfp",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "rfp_type_table": {
    "table_definition_id": 445,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "recruitment",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "role_ml_table": {
    "table_definition_id": 305,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "role",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "role_table": {
    "table_definition_id": 533,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "role",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "role_user_table": {
    "table_definition_id": 291,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "role_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "search_request_table": {
    "table_definition_id": 237,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "search",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "search_result_table": {
    "table_definition_id": 236,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "search",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "seller_ml_table": {
    "table_definition_id": 379,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "seller_table": {
    "table_definition_id": 369,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_ml_table": {
    "table_definition_id": 89,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_offer_location_table": {
    "table_definition_id": 96,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_offer_ml_table": {
    "table_definition_id": 95,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_offer_table": {
    "table_definition_id": 94,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_order_table": {
    "table_definition_id": 97,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_provider_ml_table": {
    "table_definition_id": 91,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_provider_table": {
    "table_definition_id": 90,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_receiver_ml_table": {
    "table_definition_id": 93,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_receiver_table": {
    "table_definition_id": 92,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_review_ml_table": {
    "table_definition_id": 99,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_review_table": {
    "table_definition_id": 98,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_table": {
    "table_definition_id": 377,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "service_type_table": {
    "table_definition_id": 378,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "severity_ml_table": {
    "table_definition_id": 189,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "severity_table": {
    "table_definition_id": 190,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "severity_table_m_old": {
    "table_definition_id": 82,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "severity_table_old": {
    "table_definition_id": 81,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "shop_ml_table": {
    "table_definition_id": 122,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "shop_seller_table": {
    "table_definition_id": 123,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "shop_table": {
    "table_definition_id": 211,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_goods",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "signature_table": {
    "table_definition_id": 136,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "signature",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "smartlink_table": {
    "table_definition_id": 386,
    "entity_type_id1": 18,
    "entity_type_id2": None,
    "schema": "smartlink",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "smartlink_type_table": {
    "table_definition_id": 411,
    "entity_type_id1": 39,
    "entity_type_id2": None,
    "schema": "smartlink",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": 0,
    "is_user_data_whole_table": 0,
    "end_timestamp": None
  },
  "source_code_table": {
    "table_definition_id": 389,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "repo",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sport_event_ml_table": {
    "table_definition_id": 223,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sport_event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sport_event_table": {
    "table_definition_id": 224,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sport_event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sql_report_table": {
    "table_definition_id": 424,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sql_report",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sql_structure2code_table": {
    "table_definition_id": 343,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sql2code",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sql2code_file_table": {
    "table_definition_id": 342,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sql2code",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sql2code_table": {
    "table_definition_id": 444,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sql2code",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "star_transaction_table": {
    "table_definition_id": 29,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "star_transaction",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": 1,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": 1,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "state_ml_table": {
    "table_definition_id": 332,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "state_table": {
    "table_definition_id": 325,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "location",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "status_table": {
    "table_definition_id": 109,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "status_table_ml": {
    "table_definition_id": 110,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "logger",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "storage_table": {
    "table_definition_id": 394,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "storage",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "student_table": {
    "table_definition_id": 183,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service_education",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subscription_action_limit_template_table": {
    "table_definition_id": 232,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "subscription",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subscription_ml_table": {
    "table_definition_id": 187,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "subscription",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subscription_table": {
    "table_definition_id": 293,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "subscription",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subscription_user_profile_action_limit_table": {
    "table_definition_id": 231,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "subscription_user_profile_action_limit",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subscription_user_table": {
    "table_definition_id": 255,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "subscription_user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subsystem_ml_table": {
    "table_definition_id": 239,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "system",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "subsystem_table": {
    "table_definition_id": 240,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "system",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "supplier_category_table": {
    "table_definition_id": 174,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "sys_config": {
    "table_definition_id": 181,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "sys",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "system_entity_table": {
    "table_definition_id": 135,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "system",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "system_ml_table": {
    "table_definition_id": 188,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "system",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "system_table": {
    "table_definition_id": 241,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "system",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_definition_table": {
    "table_definition_id": 540,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_deployment": {
    "table_definition_id": 66,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_deployments": {
    "table_definition_id": 59,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_history_table": {
    "table_definition_id": 227,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_list_member": {
    "table_definition_id": 67,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_list_old": {
    "table_definition_id": 426,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_old": {
    "table_definition_id": 70,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_size_history_table": {
    "table_definition_id": 42,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database_history",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": 1,
    "is_one_table_per_all_tenants": 1,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "table_type_table": {
    "table_definition_id": 313,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "database",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "teacher_table": {
    "table_definition_id": 182,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "marketplace_service_education",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_location_profile_ml_table": {
    "table_definition_id": 207,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": 1,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_location_profile_table": {
    "table_definition_id": 206,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": 1,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_mysql_ml_table": {
    "table_definition_id": 38,
    "entity_type_id1": 1,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": "test_mysql_ml_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": 1,
    "is_main_column": 1,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_mysql_table": {
    "table_definition_id": 18,
    "entity_type_id1": 1,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": "test_mysql_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": 1,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_orm_table": {
    "table_definition_id": 141,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_typeorm_ml_table": {
    "table_definition_id": 212,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "test_typeorm_table": {
    "table_definition_id": 199,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "test",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_block_table": {
    "table_definition_id": 519,
    "entity_type_id1": 30,
    "entity_type_id2": None,
    "schema": "text_block",
    "view_name": "text_block_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_block_type_ml_table": {
    "table_definition_id": 180,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "text_block_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_block_type_regex_table": {
    "table_definition_id": 489,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "text_block_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_block_type_table": {
    "table_definition_id": 522,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "text_block_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_category_table": {
    "table_definition_id": 100,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "text",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "text_table": {
    "table_definition_id": 101,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "text",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "trigger_event": {
    "table_definition_id": 60,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "trigger_event",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "url_table": {
    "table_definition_id": 41,
    "entity_type_id1": 24,
    "entity_type_id2": None,
    "schema": "url",
    "view_name": "url_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 11,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "url_type_ml_table": {
    "table_definition_id": 337,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "url_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "url_type_table": {
    "table_definition_id": 366,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "url_type",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_age_range": {
    "table_definition_id": 124,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_external_table": {
    "table_definition_id": 12,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user_external",
    "view_name": "user_external_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": 1,
    "sensative_columns": "access_token, refresh_token",
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_invitation": {
    "table_definition_id": 47,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user_invitation",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_language": {
    "table_definition_id": 125,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_list_member_table": {
    "table_definition_id": 65,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_list_old": {
    "table_definition_id": 64,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_list_table": {
    "table_definition_id": 63,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_pii_table": {
    "table_definition_id": 28,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user_pii",
    "view_name": "user_pii_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": 1,
    "sensative_columns": "password_clear_text, password_encrypted",
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_preference_table": {
    "table_definition_id": 264,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user_preference",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 1,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_quota": {
    "table_definition_id": 52,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user_quota",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_status_history_table": {
    "table_definition_id": 46,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_status_ml_table": {
    "table_definition_id": 45,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_status_table": {
    "table_definition_id": 44,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "user_table": {
    "table_definition_id": 5,
    "entity_type_id1": 25,
    "entity_type_id2": None,
    "schema": "user",
    "view_name": "user_view",
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": 6,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": 1,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": 0,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 0,
    "is_user_data": 1,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "vacancy_ml_table": {
    "table_definition_id": 210,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "vacancy_table": {
    "table_definition_id": 205,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "vacancy",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "variable_ml_table": {
    "table_definition_id": 483,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "variable_table": {
    "table_definition_id": 461,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "field",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": 1,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": None,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "visibility_ml_table": {
    "table_definition_id": 147,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "visibility",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "visibility_table": {
    "table_definition_id": 146,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "visibility",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": 1,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "workflow_action_ml_table": {
    "table_definition_id": 358,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": 1,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "workflow_action_table": {
    "table_definition_id": 465,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": 1,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 1,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  },
  "workflow_state_old": {
    "table_definition_id": 43,
    "entity_type_id1": None,
    "entity_type_id2": None,
    "schema": "dialog_workflow",
    "view_name": None,
    "seq": 0,
    "table_list_id": None,
    "is_created_updated_fields": 1,
    "is_logical_delete_fields": 1,
    "contact_mapping_table_id": None,
    "profile_mapping_table_id": None,
    "is_ml_parent_table": None,
    "is_ml_child_table": None,
    "table_type": None,
    "is_number_column": None,
    "is_test_data_column": None,
    "is_main_column": None,
    "is_single_and_multi_value": None,
    "name_of_single_value_column": None,
    "is_mapping_table": None,
    "is_entity": None,
    "is_history": None,
    "is_one_table_per_all_tenants": 0,
    "is_tree": None,
    "is_graph": None,
    "is_metrics": None,
    "is_visibility": None,
    "is_pii": None,
    "is_sensative": None,
    "sensative_columns": None,
    "is_dynamic": None,
    "description": None,
    "insert_is_undelete": None,
    "is_metadata": 0,
    "is_metadata_whole_table": None,
    "is_user_data": None,
    "is_user_data_whole_table": None,
    "end_timestamp": None
  }
}
